Package: linf
Type: Package
Title: L-Infinity Normalization and Dominant Community State Types
Version: 0.1.0
Authors@R: 
    person("Pawel", "Gajer", , "pgajer@gmail.com", role = c("aut", "cre"))
Author: Pawel Gajer [aut, cre]
Maintainer: Pawel Gajer <pgajer@gmail.com>
Description: Implements L-infinity normalization for compositional matrices,
    assigns samples to dominant features, constructs truncated and
    hierarchically refined dominant community state types, and computes
    representative landmark profiles. The methods are described in the
    accompanying publication <doi:10.48550/arXiv.2503.21543>. Bundled vaginal
    and gut microbiome data support reproducible demonstrations of the package
    interface; phenotype fields in the stratified gut subset are illustrative
    and are not suitable for population-level inference.
License: MIT + file LICENSE
Encoding: UTF-8
Language: en-US
LazyData: true
Depends: R (>= 4.0)
Imports: Matrix, methods
Suggests: testthat (>= 3.2.0), htmltools, knitr, plotly, rmarkdown
VignetteBuilder: knitr
Config/testthat/edition: 3
URL: https://github.com/pgajer/linf, https://arxiv.org/abs/2503.21543
BugReports: https://github.com/pgajer/linf/issues
Config/roxygen2/version: 8.0.0
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-07-26 21:51:49 UTC; pgajer
Repository: CRAN
Date/Publication: 2026-08-05 08:40:02 UTC
Built: R 4.6.1; ; 2026-08-05 09:16:43 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,linf.csts)
S3method(summary,linf.csts)
export(asv.to.linf.csts)
export(collapse.rare)
export(expand.rare)
export(filter.asv)
export(latex.linf.csts)
export(linf.cells)
export(linf.csts)
export(linf.dcst.landmark.pipeline)
export(linf.feature.labels)
export(linf.hypercube.embedding)
export(linf.landmarks)
export(normalize.linf)
export(refine.linf.csts)
export(refine.linf.csts.iter)
export(transfer.dcsts)
importFrom(methods,as)

<!-- README.md is generated from README.Rmd. Please edit that file -->

<!-- badges: start -->

[![R-CMD-check](https://github.com/pgajer/linf/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/pgajer/linf/actions/workflows/R-CMD-check.yaml)
[![DOI](https://zenodo.org/badge/DOI/10.48550/arXiv.2503.21543.svg)](https://doi.org/10.48550/arXiv.2503.21543)
<!-- badges: end -->

# linf — L∞ Normalization and Dominant Community State Types for Compositional Data

**linf** is a lightweight R package for analysing compositional data
through L-infinity (L∞) normalization and Dominant Community State Types
(dCSTs).

Classical centered log-ratio (CLR) and isometric log-ratio (ILR)
coordinates are defined only for strictly positive compositions.
Zero-containing observations lie on the boundary of the simplex and must
first undergo pseudocount addition or zero replacement, which represents
them as interior compositions. This is questionable when zeros represent
true absence, especially when every sample contains structural zeros, as
is common in microbiome feature tables: replacement then represents
every sample as containing every feature in the analysed feature set.
For vaginal 16S rRNA or metagenomic data, this would imply that every
vaginal microbial community contains every phylotype included in the
analysis, a biologically implausible assumption. L∞ normalization
retains zeros: dividing each sample by its maximum places the
observation on the boundary of the unit L∞ ball. The dominant feature —
the one that achieves the maximum — defines a natural, parameter-free
partition of samples into **dominance sample sets**.

dCST construction is rank based. Depth-1 dCSTs partition samples by the
rank-1 (most abundant) feature. Deeper dCSTs iteratively refine each
retained dominance-lineage using the next-ranked feature. The support
threshold *n*₀ sets the minimum sample count required to retain a
dominance sample set.

## Installation

``` r
# From GitHub (development version)
# install.packages("devtools")
devtools::install_github("pgajer/linf", build_vignettes = TRUE)
```

or

``` r
# From CRAN
install.packages("linf")
```

## Features

- **L∞ normalization** — `normalize.linf()`: row-wise division by
  maximum, mapping each sample to the L∞ unit-ball boundary.
- **Dominant-feature assignment** — `linf.dominant.features()`: rank-1
  assignment per sample, returning indices, labels, and level sets.
- **Truncated dCSTs** — `linf.csts()`: apply the support threshold *n*₀
  and either group low-support samples together or absorb them into
  retained states.
- **Iterative refinement** — `refine.linf.csts()`: depth-2+ dCSTs via
  successive rank decomposition.
- **Landmark profiles** — `linf.landmarks()`: representative
  compositional profiles (endpoint max/min, mean) for each dCST.
- **ASV filtering** — `filter.asv()`: library-size and prevalence
  filtering for amplicon count matrices.
- **Pipeline wrapper** — `asv.to.linf.csts()`: counts → filter →
  normalise → truncated dCSTs in one call.

## Quick Start

``` r
library(linf)

set.seed(1)

# toy counts (samples × features)
S.counts <- matrix(rpois(10 * 3, 5), nrow = 10, ncol = 3,
                   dimnames = list(paste0("s", 1:10), c("A", "B", "C")))

# L∞ relatives (nonzero rows have max 1; zeros remain zero)
Z <- normalize.linf(S.counts)
apply(Z, 1, max)
#> returns 1 for nonzero rows, 0 for all-zero rows

# Dominant-feature assignments: indices + labels
dominant.features <- linf.dominant.features(Z)
table(dominant.features$label, useNA = "ifany")

# Absorb-policy dCSTs: reassign low-support samples among retained states
res <- linf.csts(Z, n0 = 4, low.freq.policy = "absorb")
table(res$lineage.label, useNA = "ifany")
```

## Gut Microbiome Demonstration

The figure below illustrates depth-1 dCSTs fitted under the absorb
policy to a bundled set of 766 gut microbiome samples from the American
Gut Project (AGP). After filtering, 763 samples and 307 taxa remain.
With *n*₀ = 30, samples from provisional dominance sample sets below the
support threshold are reassigned to the retained state for which they
have the largest normalized abundance; no composite rare category is
shown.

The subset was deliberately stratified to include every sample assigned
to four selected uncommon dCSTs; the remaining slots are a seed-42
simple random sample from the eligible background. Phenotypes do not
influence selection. The object is suitable for demonstrating the
package workflow, but its phenotype frequencies, effect sizes, and
p-values must not be interpreted as population estimates because
inclusion probabilities differ by dCST.

### dCST Size Distribution

<img src="man/figures/readme-dcst-barplot.png" alt="Barplot of absorb-policy depth-1 dCST sizes in 763 filtered AGP gut samples" width="700" />

The 763 filtered samples are assigned among seven retained dCSTs. The
largest are *Bacteroides* (239 samples), *Escherichia-Shigella* (134),
and *Staphylococcus* (102). In total, 159 samples from low-support
provisional dominance sample sets are absorbed into retained states.

See `vignette("linf-intro")` for the package-safe demonstration. The
[full-analysis
source](https://github.com/pgajer/linf/blob/main/vignettes/articles/gut-dcst-disease-analysis.Rmd)
uses an independently selected 5,000-sample analysis cohort and is
maintained as a companion repository article rather than a package
vignette.

## Vignettes

The package ships with two vignettes:

- **Dominant Community State Types: From Normalization to a Gut
  Demonstration** — a step-by-step tutorial covering L∞ normalization,
  dominant-feature assignment, truncated dCSTs, and descriptive
  exploration of the bundled AGP gut data.
- **dCSTs for Vaginal Microbiome Data** — applying the pipeline to the
  Valencia 2k vaginal dataset, demonstrating how dCSTs recover the
  classical community state types (CST I–V) without supervised training.

``` r
browseVignettes("linf")
```

## Citation

If you use this package, please cite:

> Gajer, P. & Ravel, J. (2025). A New Approach to Compositional Data
> Analysis using L∞-normalization with Applications to Vaginal
> Microbiome. *arXiv preprint arXiv:2503.21543* \[stat.CO\]. doi:
> [10.48550/arXiv.2503.21543](https://doi.org/10.48550/arXiv.2503.21543)

**BibTeX**

``` bibtex
@article{gajer2025linf,
  title   = {A New Approach to Compositional Data Analysis using
             {$L^{\infty}$}-normalization with Applications to
             Vaginal Microbiome},
  author  = {Gajer, Pawel and Ravel, Jacques},
  year    = {2025},
  eprint  = {2503.21543},
  archivePrefix = {arXiv},
  primaryClass  = {stat.CO},
  journal = {arXiv preprint arXiv:2503.21543},
  doi     = {10.48550/arXiv.2503.21543},
  url     = {https://arxiv.org/abs/2503.21543}
}
```

## License

MIT © 2025 Pawel Gajer. See `LICENSE` / `LICENSE.md`. Bundled-data
sources and upstream terms are recorded in `inst/DATA_PROVENANCE.md`.

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  fig.width = 7,
  fig.height = 5
)
library(linf)

## ----toy-normalize------------------------------------------------------------
set.seed(1)
S.counts <- matrix(
  rpois(30, lambda = 5),
  nrow = 10, ncol = 3,
  dimnames = list(paste0("s", 1:10), c("Taxon_A", "Taxon_B", "Taxon_C"))
)

Z <- normalize.linf(S.counts)
apply(Z, 1, max)   # all 1

## ----toy-cells----------------------------------------------------------------
cells <- linf.cells(Z)
table(cells$label)

## ----toy-csts-----------------------------------------------------------------
A <- matrix(c(5, 1, 0), nrow = 6, ncol = 3, byrow = TRUE)
B <- matrix(c(1, 5, 0), nrow = 2, ncol = 3, byrow = TRUE)
C <- matrix(c(1, 0, 5), nrow = 2, ncol = 3, byrow = TRUE)
S <- rbind(A, B, C)
S <- sweep(S, 1, rowSums(S), "/")
colnames(S) <- c("Dom1", "Dom2", "Dom3")

res <- linf.csts(S, n0 = 5)
table(res$cell.label, useNA = "ifany")

## ----load-data----------------------------------------------------------------
data(agp_gut)
str(agp_gut, max.level = 1)

dim(agp_gut$counts)   # samples x taxa

## ----meta-summary-------------------------------------------------------------
disease_cols <- c("IBS", "IBD", "Obesity", "Cardiovascular_disease",
                  "Autoimmune", "Acid_reflux")
sapply(disease_cols, function(col) sum(agp_gut$meta[[col]], na.rm = TRUE))

## ----filter-------------------------------------------------------------------
filt <- filter.asv(agp_gut$counts, min.lib = 1000, prev.prop = 0.05,
                   min.count = 2)
dim(filt$counts)
dim(filt$rel)

## ----normalize----------------------------------------------------------------
M <- normalize.linf(filt$counts)

# Verify: every row max is 1
stopifnot(all(abs(apply(M, 1, max) - 1) < 1e-10))

## ----dcst-depth1--------------------------------------------------------------
csts <- linf.csts(M, n0 = 30)

# dCST size distribution
sort(table(csts$cell.label), decreasing = TRUE)

## ----dcst-depth2--------------------------------------------------------------
csts2 <- refine.linf.csts(M, csts, n0 = 30)

# Show depth-2 dCSTs with >= 20 samples
tab2 <- sort(table(csts2$cell.label), decreasing = TRUE)
tab2[tab2 >= 20]

## ----dominance-dist, fig.cap = "Distribution of dominance strength across gut samples. Most samples show moderate dominance, with a tail of strongly dominated communities."----
rel <- filt$rel
dom_strength <- apply(rel, 1, max)
hist(dom_strength, breaks = 50, col = "#2ecc71", border = "white",
     main = "Dominance Strength in Gut Microbiome",
     xlab = "Relative abundance of dominant species",
     ylab = "Number of samples")
abline(v = 0.5, col = "red", lty = 2, lwd = 2)
legend("topright", "50% dominance", col = "red", lty = 2, lwd = 2, bty = "n")

## ----dcst-barplot, fig.cap = "dCST size distribution at depth 1. Bacteroides and Escherichia-Shigella dominate; rare dCSTs in the tail are collapsed into RARE_DOMINANT."----
tab1 <- sort(table(csts$cell.label), decreasing = TRUE)
par(mar = c(10, 4, 3, 1))
bp <- barplot(tab1,
        col = ifelse(names(tab1) == "RARE_DOMINANT", "#e74c3c", "#3498db"),
        las = 2, cex.names = 0.7,
        ylab = "Number of samples",
        main = "Gut dCST Size Distribution (depth 1)")
text(bp, tab1 + 5, labels = tab1, cex = 0.7, pos = 3)

## ----phenotype-summary--------------------------------------------------------
meta <- agp_gut$meta[match(rownames(M), agp_gut$meta$Run), ]
stopifnot(all(meta$Run == rownames(M)))

conditions <- c("IBS", "IBD", "Obesity", "Cardiovascular_disease",
                "Autoimmune", "Acid_reflux", "Lung_disease")
phenotype_summary <- data.frame(
  Condition = conditions,
  Recorded_cases = vapply(
    conditions,
    function(x) sum(meta[[x]] == 1, na.rm = TRUE),
    integer(1)
  ),
  Non_missing = vapply(
    conditions,
    function(x) sum(!is.na(meta[[x]])),
    integer(1)
  )
)
knitr::kable(
  phenotype_summary,
  caption = "Recorded phenotypes in the selected demonstration subset"
)

#' Landmark-aware dCST pipeline
#'
#' @description
#' Small wrapper that normalizes a nonnegative matrix with
#' \code{\link{normalize.linf}}, computes depth-1 dCSTs with
#' \code{\link{linf.csts}}, refines once to depth 2 with
#' \code{\link{refine.linf.csts}}, and then computes landmark points for both
#' depths with \code{\link{linf.landmarks}}.
#'
#' The function is intentionally explicit rather than highly abstracted. It
#' keeps the intermediate objects visible and returns them together in one list
#' so downstream workflows can inspect the normalized matrix, the depth-1 dCSTs,
#' the depth-2 dCSTs, and the landmark tables without re-running the pipeline.
#'
#' @param X Numeric matrix (samples x features). Must be finite and nonnegative.
#' @param feature.ids Optional character vector of stable feature identifiers,
#'   length \code{ncol(X)}.
#' @param feature.labels Optional character vector of display labels, length
#'   \code{ncol(X)}.
#' @param n0.depth1 Integer >= 1. Minimum support for a depth-1 dominance
#'   sample set.
#' @param n0.depth2 Integer >= 1. Minimum support for a depth-2 child lineage.
#' @param refinement.factor Numeric > 0. Auto-refinement threshold multiplier
#'   passed to \code{\link{refine.linf.csts}}.
#' @param sep Character scalar used to join depth-refined dCST path tokens.
#' @param low.freq.policy Character. One of \code{"pure"} or \code{"absorb"}.
#'   Controls the active dCST view while still preserving parallel rare/absorb
#'   views inside the returned dCST objects. The legacy value \code{"rare"} is
#'   still accepted as a deprecated alias for \code{"pure"}.
#' @param rare.label Character scalar for rare buckets.
#' @param depth1.landmark.types Character vector of landmark types for depth 1.
#' @param depth2.landmark.types Character vector of landmark types for depth 2.
#' @param landmark.view Character. One of \code{"absorb"}, \code{"active"}, or
#'   \code{"rare"}. Defaults to \code{"absorb"} because landmark reporting is
#'   typically requested on the absorb view.
#' @param tie.method Character. Tie handling passed through to
#'   \code{\link{linf.csts}} and \code{\link{linf.landmarks}}.
#' @param verbose Logical. Passed to \code{\link{refine.linf.csts}}.
#' @param backend Character. Matrix backend to use: \code{"auto"},
#'   \code{"dense"}, or \code{"sparse"}.
#'
#' @return A named list with components:
#' \itemize{
#'   \item \code{linf.rel}: L-infinity-normalized matrix.
#'   \item \code{dcst.depth1}: depth-1 \code{"linf.csts"} object.
#'   \item \code{dcst.depth2}: depth-2 \code{"linf.csts"} object.
#'   \item \code{landmarks.depth1}: \code{"linf.landmarks"} object at depth 1.
#'   \item \code{landmarks.depth2}: \code{"linf.landmarks"} object at depth 2.
#'   \item \code{params}: compact record of the pipeline arguments used.
#' }
#'
#' @examples
#' X <- rbind(
#'   s1 = c(10, 8, 1),
#'   s2 = c(9, 7, 2),
#'   s3 = c(8, 2, 7),
#'   s4 = c(7, 1, 8),
#'   s5 = c(1, 10, 2),
#'   s6 = c(2, 9, 1)
#' )
#' ids <- c("asv_1", "asv_2", "asv_3")
#' labels <- c("L. iners 1", "Gard. vaginalis 2", "BVAB1 3")
#'
#' out <- linf.dcst.landmark.pipeline(
#'   X,
#'   feature.ids = ids,
#'   feature.labels = labels,
#'   n0.depth1 = 2,
#'   n0.depth2 = 2,
#'   refinement.factor = 2,
#'   low.freq.policy = "pure",
#'   landmark.view = "absorb",
#'   verbose = FALSE
#' )
#'
#' names(out)
#' out$dcst.depth2$cst.depth
#' out$landmarks.depth2$view
#' @export
linf.dcst.landmark.pipeline <- function(
  X,
  feature.ids = NULL,
  feature.labels = NULL,
  n0.depth1 = 50,
  n0.depth2 = 25,
  refinement.factor = 2,
  sep = "__",
  low.freq.policy = c("pure", "absorb"),
  rare.label = "RARE_DOMINANT",
  depth1.landmark.types = c("endpoint.max", "endpoint.min", "mean.rep", "median.rep"),
  depth2.landmark.types = c("endpoint.max", "endpoint.min", "mean.rep", "median.rep"),
  landmark.view = c("absorb", "active", "rare"),
  tie.method = c("first", "random", "error"),
  verbose = FALSE,
  backend = c("auto", "dense", "sparse")
) {
  low.freq.policy <- linf.normalize.low.freq.policy(
    low.freq.policy,
    "linf.dcst.landmark.pipeline"
  )
  landmark.view <- match.arg(landmark.view)
  tie.method <- match.arg(tie.method)
  prep <- linf.prepare.matrix(X, backend = backend, fun.name = "linf.dcst.landmark.pipeline")
  M <- prep$X
  backend <- prep$backend
  linf.validate.matrix(M, backend = backend, fun.name = "linf.dcst.landmark.pipeline")
  if (!is.numeric(n0.depth1) || length(n0.depth1) != 1L || n0.depth1 < 1 || n0.depth1 %% 1 != 0) {
    stop("linf.dcst.landmark.pipeline: n0.depth1 must be integer >= 1")
  }
  if (!is.numeric(n0.depth2) || length(n0.depth2) != 1L || n0.depth2 < 1 || n0.depth2 %% 1 != 0) {
    stop("linf.dcst.landmark.pipeline: n0.depth2 must be integer >= 1")
  }
  if (!is.numeric(refinement.factor) || length(refinement.factor) != 1L ||
      !is.finite(refinement.factor) || refinement.factor <= 0) {
    stop("linf.dcst.landmark.pipeline: refinement.factor must be a finite numeric > 0")
  }
  if (!is.character(sep) || length(sep) != 1L || !nzchar(sep)) {
    stop("linf.dcst.landmark.pipeline: sep must be a non-empty character scalar")
  }
  if (!is.character(rare.label) || length(rare.label) != 1L || !nzchar(rare.label)) {
    stop("linf.dcst.landmark.pipeline: rare.label must be a non-empty character scalar")
  }

  linf.rel <- normalize.linf(M, backend = backend)

  dcst.depth1 <- linf.csts(
    linf.rel,
    feature.ids = feature.ids,
    feature.labels = feature.labels,
    n0 = n0.depth1,
    low.freq.policy = low.freq.policy,
    rare.label = rare.label,
    tie.method = tie.method,
    backend = backend
  )

  dcst.depth2 <- refine.linf.csts(
    linf.rel,
    dcst.depth1,
    n0 = n0.depth2,
    refinement.factor = refinement.factor,
    sep = sep,
    low.freq.policy = low.freq.policy,
    rare.label = rare.label,
    verbose = verbose,
    backend = backend
  )

  landmarks.depth1 <- linf.landmarks(
    linf.rel,
    dcst.depth1,
    depth = 1L,
    view = landmark.view,
    landmark.types = depth1.landmark.types,
    tie.method = tie.method,
    backend = backend
  )

  landmarks.depth2 <- linf.landmarks(
    linf.rel,
    dcst.depth2,
    depth = 2L,
    view = landmark.view,
    landmark.types = depth2.landmark.types,
    tie.method = tie.method,
    backend = backend
  )

  list(
    linf.rel = linf.rel,
    dcst.depth1 = dcst.depth1,
    dcst.depth2 = dcst.depth2,
    landmarks.depth1 = landmarks.depth1,
    landmarks.depth2 = landmarks.depth2,
    params = list(
      n0.depth1 = as.integer(n0.depth1),
      n0.depth2 = as.integer(n0.depth2),
      refinement.factor = refinement.factor,
      sep = sep,
      low.freq.policy = low.freq.policy,
      rare.label = rare.label,
      landmark.view = landmark.view,
      depth1.landmark.types = depth1.landmark.types,
      depth2.landmark.types = depth2.landmark.types,
      tie.method = tie.method,
      backend = backend
    )
  )
}

test_that("linf.dcst.landmark.pipeline returns the expected object bundle", {
  X <- rbind(
    s1 = c(10, 8, 1),
    s2 = c(9, 7, 2),
    s3 = c(8, 2, 7),
    s4 = c(7, 1, 8),
    s5 = c(1, 10, 2),
    s6 = c(2, 9, 1)
  )
  ids <- c("asv_1", "asv_2", "asv_3")
  labels <- c("L. iners 1", "Gard. vaginalis 2", "BVAB1 3")

  out <- linf.dcst.landmark.pipeline(
    X,
    feature.ids = ids,
    feature.labels = labels,
    n0.depth1 = 2,
    n0.depth2 = 2,
    refinement.factor = 2,
    low.freq.policy = "pure",
    landmark.view = "absorb",
    verbose = FALSE
  )

  expect_named(
    out,
    c("linf.rel", "dcst.depth1", "dcst.depth2", "landmarks.depth1", "landmarks.depth2", "params")
  )
  expect_equal(dim(out$linf.rel), dim(X))
  expect_s3_class(out$dcst.depth1, "linf.csts")
  expect_s3_class(out$dcst.depth2, "linf.csts")
  expect_s3_class(out$landmarks.depth1, "linf.landmarks")
  expect_s3_class(out$landmarks.depth2, "linf.landmarks")
  expect_equal(out$dcst.depth1$cst.depth, 1L)
  expect_equal(out$dcst.depth2$cst.depth, 2L)
})

test_that("linf.dcst.landmark.pipeline computes landmark tables for both depths", {
  X <- rbind(
    r1 = c(1.00, 0.80, 0.20),
    r2 = c(0.95, 0.70, 0.30),
    r3 = c(0.90, 0.20, 0.70),
    r4 = c(0.85, 0.10, 0.80),
    r5 = c(0.20, 1.00, 0.10),
    r6 = c(0.30, 0.95, 0.10)
  )
  ids <- c("asv_1", "asv_4", "asv_6")
  labels <- c("L. iners 1", "L. iners 4", "BVAB1 6")

  out <- linf.dcst.landmark.pipeline(
    X,
    feature.ids = ids,
    feature.labels = labels,
    n0.depth1 = 2,
    n0.depth2 = 2,
    refinement.factor = 2,
    low.freq.policy = "pure",
    landmark.view = "absorb",
    depth1.landmark.types = c("endpoint.max", "mean.rep"),
    depth2.landmark.types = c("endpoint.max", "median.rep"),
    verbose = FALSE
  )

  expect_true(all(c("cells", "landmarks") %in% names(out$landmarks.depth1)))
  expect_true(all(c("cells", "landmarks") %in% names(out$landmarks.depth2)))
  expect_gt(nrow(out$landmarks.depth1$cells), 0L)
  expect_gt(nrow(out$landmarks.depth1$landmarks), 0L)
  expect_gt(nrow(out$landmarks.depth2$cells), 0L)
  expect_gt(nrow(out$landmarks.depth2$landmarks), 0L)
  expect_equal(sort(unique(out$landmarks.depth1$landmarks$landmark.type)), c("endpoint.max", "mean.rep"))
  expect_equal(sort(unique(out$landmarks.depth2$landmarks$landmark.type)), c("endpoint.max", "median.rep"))
})

test_that("linf.dcst.landmark.pipeline respects absorb landmark views under pure policy", {
  A <- matrix(c(5, 1, 0), nrow = 6, ncol = 3, byrow = TRUE)
  B <- matrix(c(1, 5, 0), nrow = 2, ncol = 3, byrow = TRUE)
  C <- matrix(c(1, 0, 5), nrow = 2, ncol = 3, byrow = TRUE)
  X <- rbind(A, B, C)
  rownames(X) <- paste0("s", seq_len(nrow(X)))
  colnames(X) <- c("Dom1", "Dom2", "Dom3")

  out <- linf.dcst.landmark.pipeline(
    X,
    n0.depth1 = 5,
    n0.depth2 = 3,
    refinement.factor = 2,
    low.freq.policy = "pure",
    landmark.view = "absorb",
    depth1.landmark.types = "endpoint.max",
    depth2.landmark.types = "endpoint.max",
    verbose = FALSE
  )

  expect_identical(out$dcst.depth1$low.freq.policy, "pure")
  expect_identical(out$landmarks.depth1$view, "absorb")
  expect_identical(out$landmarks.depth2$view, "absorb")
  expect_false(any(out$landmarks.depth1$cells$cell.id == out$dcst.depth1$rare.label))
  expect_false(any(out$landmarks.depth2$cells$cell.id == out$dcst.depth2$rare.label))
})

test_that("linf.dcst.landmark.pipeline carries feature ids and labels into landmark output", {
  X <- rbind(
    s1 = c(0.95, 0.80, 0.05),
    s2 = c(0.90, 0.75, 0.10),
    s3 = c(0.85, 0.20, 0.75),
    s4 = c(0.84, 0.15, 0.80),
    s5 = c(0.10, 0.95, 0.05),
    s6 = c(0.05, 0.94, 0.08)
  )
  ids <- c("asv_1", "asv_4", "asv_5")
  labels <- c("L. iners 1", "L. iners 4", "Mega. lornae")

  out <- linf.dcst.landmark.pipeline(
    X,
    feature.ids = ids,
    feature.labels = labels,
    n0.depth1 = 2,
    n0.depth2 = 2,
    refinement.factor = 2,
    low.freq.policy = "pure",
    landmark.view = "absorb",
    depth1.landmark.types = "endpoint.max",
    depth2.landmark.types = "endpoint.max",
    verbose = FALSE
  )

  expect_true(all(out$dcst.depth1$feature.ids == ids))
  expect_true(all(out$dcst.depth1$feature.labels == labels))
  expect_true(all(out$dcst.depth2$feature.ids == ids))
  expect_true(all(out$dcst.depth2$feature.labels == labels))
  expect_true(all(out$landmarks.depth1$cells$target.feature.id %in% c(ids, NA_character_)))
  expect_true(all(out$landmarks.depth1$cells$target.feature.label %in% c(labels, NA_character_)))
  expect_true(any(grepl("^asv_1(__|$)", out$landmarks.depth2$cells$cell.id)))
  expect_true(any(grepl("^L\\. iners 1(__|$)", out$landmarks.depth2$cells$cell.label)))
})

#' Transfer samples into a frozen dCST hierarchy
#'
#' @description
#' Assigns rows of a new sample-by-feature matrix into the realized lineage sets
#' of a fitted \code{"linf.csts"} hierarchy. The hierarchy is not refit: each
#' sample is walked through the frozen tree by choosing, at each depth, the
#' realized child lineage whose newly added feature has the largest abundance in
#' that sample.
#'
#' @param X Nonnegative sample-by-feature matrix to transfer. Columns may be in
#'   any order and may include features not present in \code{csts}; missing
#'   retained features are treated as zero.
#' @param csts A fitted \code{"linf.csts"} object produced by
#'   \code{\link{linf.csts}} and optionally refined by
#'   \code{\link{refine.linf.csts}} or \code{\link{refine.linf.csts.iter}}.
#' @param depth Integer vector of requested depths. Defaults to all fitted
#'   depths in \code{csts}.
#' @param view Which fitted hierarchy view to transfer into. \code{"absorb"}
#'   uses \code{cst.levels.absorb}; \code{"rare"} uses
#'   \code{cst.levels.rare}; \code{"active"} uses \code{cst.levels}.
#' @param match.by Whether columns in \code{X} are aligned to
#'   \code{csts$feature.ids} or \code{csts$feature.labels}.
#' @param feature.ids Optional feature identifiers for columns of \code{X}.
#'   Defaults to \code{colnames(X)}.
#' @param feature.labels Optional feature labels for columns of \code{X}.
#'   Defaults to \code{feature.ids}.
#' @param tie.method How to resolve ties among frozen child lineages with equal
#'   sample abundance. \code{"support"} chooses the tied lineage with largest
#'   reference support and then lexical order; \code{"first"} uses frozen child
#'   order; \code{"random"} samples one tied lineage; \code{"error"} stops.
#' @param carry.forward.terminal.depths Logical. If \code{TRUE}, an unmatched
#'   depth-1 sample may fall back to its dominant retained feature. Terminal
#'   lineages at later depths are assigned only when they appear as realized
#'   stable lineage sets in the fitted hierarchy.
#' @param sep Separator used in lineage labels. Defaults to \code{csts$sep} or
#'   \code{"__"}.
#' @param backend Matrix backend; passed to the internal matrix preparer.
#'
#' @return A list of class \code{"linf.dcst.transfer"} with components:
#'   \itemize{
#'     \item \code{assignment}: character matrix of transferred labels for the
#'       requested depths.
#'     \item \code{all.depths}: character matrix for all fitted depths.
#'     \item \code{depth}: requested depth vector.
#'     \item \code{view}, \code{match.by}, \code{tie.method}: settings used.
#'   }
#'
#' @examples
#' X <- rbind(
#'   s1 = c(A = 10, B = 2, C = 1),
#'   s2 = c(A = 9, B = 3, C = 1),
#'   s3 = c(A = 1, B = 10, C = 2),
#'   s4 = c(A = 1, B = 9, C = 3)
#' )
#' M <- normalize.linf(X)
#' fit <- linf.csts(M, n0 = 2, low.freq.policy = "absorb")
#' transfer.dcsts(X, fit)$assignment
#'
#' @export
transfer.dcsts <- function(X,
                           csts,
                           depth = NULL,
                           view = c("absorb", "active", "rare"),
                           match.by = c("feature.ids", "feature.labels"),
                           feature.ids = NULL,
                           feature.labels = NULL,
                           tie.method = c("support", "first", "random", "error"),
                           carry.forward.terminal.depths = TRUE,
                           sep = NULL,
                           backend = c("auto", "dense", "sparse")) {
  validate.linf.csts(csts)
  view <- match.arg(view)
  match.by <- match.arg(match.by)
  tie.method <- match.arg(tie.method)
  if (is.null(sep)) sep <- csts$sep %||% "__"
  if (!is.character(sep) || length(sep) != 1L || !nzchar(sep)) {
    stop("transfer.dcsts: sep must be a non-empty character scalar")
  }
  if (!is.logical(carry.forward.terminal.depths) ||
      length(carry.forward.terminal.depths) != 1L ||
      is.na(carry.forward.terminal.depths)) {
    stop("transfer.dcsts: carry.forward.terminal.depths must be TRUE or FALSE")
  }

  prep <- linf.prepare.matrix(X, backend = backend, fun.name = "transfer.dcsts")
  X <- prep$X
  backend <- prep$backend
  linf.validate.matrix(X, backend = backend, fun.name = "transfer.dcsts")

  feature.ids <- feature.ids %||% colnames(X)
  if (is.null(feature.ids)) feature.ids <- paste0("feature", seq_len(ncol(X)))
  feature.labels <- feature.labels %||% feature.ids
  if (length(feature.ids) != ncol(X)) {
    stop("transfer.dcsts: feature.ids must have length ncol(X)")
  }
  if (length(feature.labels) != ncol(X)) {
    stop("transfer.dcsts: feature.labels must have length ncol(X)")
  }

  levels <- resolve.linf.cst.levels(csts, kind = "label", view = view)
  max.depth <- length(levels)
  if (!max.depth) stop("transfer.dcsts: csts does not contain fitted levels")
  if (is.null(depth)) depth <- seq_len(max.depth)
  if (!is.numeric(depth) || any(!is.finite(depth)) || any(depth < 1) ||
      any(depth %% 1 != 0) || any(depth > max.depth)) {
    stop("transfer.dcsts: depth must contain fitted depths between 1 and csts$cst.depth")
  }
  depth <- as.integer(depth)

  ref.features <- switch(
    match.by,
    feature.ids = csts$feature.ids,
    feature.labels = csts$feature.labels
  )
  if (is.null(ref.features)) {
    ref.features <- switch(
      match.by,
      feature.ids = csts$feature.labels,
      feature.labels = csts$feature.ids
    )
  }
  if (is.null(ref.features)) {
    stop("transfer.dcsts: csts does not contain feature metadata for matching")
  }
  ref.features <- as.character(ref.features)

  query.features <- switch(
    match.by,
    feature.ids = as.character(feature.ids),
    feature.labels = as.character(feature.labels)
  )
  X.aligned <- linf.align.transfer.matrix(X, query.features, ref.features, backend = backend)
  colnames(X.aligned) <- ref.features

  tree <- linf.dcst.transfer.tree(levels, max.depth = max.depth, sep = sep)
  assigned.raw <- vapply(
    seq_len(nrow(X.aligned)),
    function(i) {
      vals <- X.aligned[i, , drop = TRUE]
      names(vals) <- ref.features
      linf.transfer.one.sample(
        vals,
        tree = tree,
        max.depth = max.depth,
        tie.method = tie.method,
        carry.forward.terminal.depths = carry.forward.terminal.depths,
        sep = sep
      )
    },
    character(max.depth)
  )
  assignments <- if (is.null(dim(assigned.raw))) {
    matrix(assigned.raw, ncol = 1L)
  } else {
    t(assigned.raw)
  }
  rownames(assignments) <- rownames(X)
  colnames(assignments) <- paste0("depth", seq_len(max.depth))

  out <- list(
    assignment = assignments[, depth, drop = FALSE],
    all.depths = assignments,
    depth = depth,
    view = view,
    match.by = match.by,
    tie.method = tie.method,
    carry.forward.terminal.depths = carry.forward.terminal.depths,
    sep = sep
  )
  class(out) <- "linf.dcst.transfer"
  out
}

linf.align.transfer.matrix <- function(X, query.features, ref.features, backend) {
  idx <- match(ref.features, query.features)
  if (backend == "sparse") {
    out <- Matrix::Matrix(0, nrow = nrow(X), ncol = length(ref.features), sparse = TRUE)
    hit <- !is.na(idx)
    if (any(hit)) out[, hit] <- X[, idx[hit], drop = FALSE]
    rownames(out) <- rownames(X)
    return(out)
  }
  out <- matrix(0, nrow = nrow(X), ncol = length(ref.features))
  hit <- !is.na(idx)
  if (any(hit)) out[, hit] <- X[, idx[hit], drop = FALSE]
  rownames(out) <- rownames(X)
  out
}

linf.dcst.transfer.tree <- function(levels, max.depth, sep) {
  tree <- vector("list", max.depth)
  support <- vector("list", max.depth)

  level1 <- as.character(levels[[1L]])
  level1 <- level1[!is.na(level1) & nzchar(level1)]
  root.counts <- sort(table(level1), decreasing = TRUE)
  tree[[1L]] <- list("__ROOT__" = names(root.counts))
  support[[1L]] <- list("__ROOT__" = as.numeric(root.counts))
  names(support[[1L]][["__ROOT__"]]) <- names(root.counts)

  if (max.depth >= 2L) {
    for (d in 2:max.depth) {
      parent <- as.character(levels[[d - 1L]])
      child <- as.character(levels[[d]])
      keep <- !is.na(parent) & nzchar(parent) & !is.na(child) & nzchar(child)
      parent <- parent[keep]
      child <- child[keep]
      pairs <- unique(data.frame(parent = parent, child = child, stringsAsFactors = FALSE))
      counts <- as.data.frame(table(parent, child), stringsAsFactors = FALSE)
      counts <- counts[counts$Freq > 0, , drop = FALSE]

      tree[[d]] <- list()
      support[[d]] <- list()
      if (!nrow(pairs)) next
      parent.groups <- split(pairs$child, pairs$parent)
      support.groups <- split(counts, counts$parent)
      for (key in sort(names(parent.groups))) {
        children <- unique(parent.groups[[key]])
        tree[[d]][[key]] <- children
        counts.df <- support.groups[[key]]
        counts.vec <- stats::setNames(counts.df$Freq, as.character(counts.df$child))
        support[[d]][[key]] <- counts.vec
      }
    }
  }

  list(children = tree, support = support, sep = sep)
}

linf.transfer.one.sample <- function(sample.values,
                                     tree,
                                     max.depth,
                                     tie.method,
                                     carry.forward.terminal.depths,
                                     sep) {
  out <- rep(NA_character_, max.depth)
  parent.key <- "__ROOT__"

  fallback.to.dominant <- function() {
    vals <- sample.values
    vals[is.na(vals)] <- 0
    best <- max(vals)
    if (!is.finite(best) || best <= 0) return(NA_character_)
    top <- names(vals)[vals == best]
    sort(top)[[1L]]
  }

  child.score <- function(label) {
    component <- linf.lineage.last.component(label, sep = sep)
    if (!nzchar(component)) return(0)
    val <- unname(sample.values[component])
    if (!length(val) || is.na(val)) return(0)
    as.numeric(val[[1L]])
  }

  for (d in seq_len(max.depth)) {
    candidates <- tree$children[[d]][[parent.key]]
    if (is.null(candidates) || !length(candidates)) {
      if (d == 1L && carry.forward.terminal.depths) {
        fallback <- fallback.to.dominant()
        if (!is.na(fallback)) out[[d]] <- fallback
      }
      break
    }

    vals <- vapply(candidates, child.score, numeric(1L))
    best <- max(vals)
    if (!is.finite(best) || best <= 0) {
      if (d == 1L && carry.forward.terminal.depths) {
        fallback <- fallback.to.dominant()
        if (!is.na(fallback)) out[[d]] <- fallback
      }
      break
    }

    top <- candidates[vals == best]
    if (length(top) > 1L) {
      if (tie.method == "support") {
        supp <- tree$support[[d]][[parent.key]][top]
        supp[is.na(supp)] <- 0
        top <- top[supp == max(supp)]
        if (length(top) > 1L) top <- sort(top)
      } else if (tie.method == "first") {
        top <- top[[1L]]
      } else if (tie.method == "random") {
        top <- sample(top, 1L)
      } else {
        stop("transfer.dcsts: tie during frozen transfer and tie.method = 'error'")
      }
    }

    chosen <- top[[1L]]
    out[[d]] <- chosen
    parent.key <- chosen
  }

  out
}

linf.lineage.last.component <- function(label, sep) {
  if (is.na(label) || !nzchar(label)) return("")
  parts <- strsplit(label, sep, fixed = TRUE)[[1L]]
  utils::tail(parts, 1L)[[1L]]
}

#' Format feature labels for dominant-feature assignments and dCSTs
#'
#' @description
#' Builds unique display labels from stable feature IDs and taxonomy strings.
#' This is useful when dCST computation should operate on stable internal feature
#' identifiers (for example `asv_4`) while reports and figures should use
#' human-readable labels such as `L. iners 4`.
#'
#' Underscores in taxonomy strings are converted to spaces before abbreviation
#' and aliasing. If multiple features share the same display taxon, an index can
#' be appended either from the global feature ID (for example `asv_4 -> 4`) or
#' by within-taxon order.
#'
#' @param feature.ids Character vector of stable feature identifiers.
#' @param taxonomy Character vector of taxonomy strings, same length as
#'   \code{feature.ids}.
#' @param abbreviations Optional named character vector mapping genus tokens to
#'   display abbreviations, for example \code{c(Lactobacillus = "L.")}.
#' @param aliases Optional named character vector mapping full taxonomy strings
#'   to alternate labels, for example \code{c("Ca. Lachnocurva vaginae" = "BVAB1")}.
#' @param duplicate.index One of \code{"global"}, \code{"within_taxon"}, or
#'   \code{"none"}. Controls how duplicate display taxa are disambiguated.
#' @param fallback.to.id Logical. If \code{TRUE}, missing taxonomy values fall
#'   back to the feature ID.
#'
#' @return Character vector of unique display labels.
#' @examples
#' ids <- c("asv_1", "asv_4", "asv_5", "asv_6")
#' tax <- c(
#'   "Lactobacillus iners",
#'   "Lactobacillus iners",
#'   "Megasphaera lornae",
#'   "Ca_Lachnocurva_vaginae"
#' )
#' abbr <- c(
#'   Lactobacillus = "L.",
#'   Gardnerella = "Gard.",
#'   Megasphaera = "Mega."
#' )
#' aliases <- c(
#'   "Ca. Lachnocurva vaginae" = "BVAB1",
#'   "Ca_Lachnocurva_vaginae" = "BVAB1"
#' )
#' linf.feature.labels(ids, tax, abbreviations = abbr, aliases = aliases)
#' @export
linf.feature.labels <- function(feature.ids,
                                taxonomy,
                                abbreviations = NULL,
                                aliases = NULL,
                                duplicate.index = c("global", "within_taxon", "none"),
                                fallback.to.id = TRUE) {
  duplicate.index <- match.arg(duplicate.index)

  if (length(feature.ids) != length(taxonomy)) {
    stop("linf.feature.labels: feature.ids and taxonomy must have the same length")
  }

  ids <- as.character(feature.ids)
  tax <- as.character(taxonomy)

  base <- trimws(gsub("_+", " ", tax))
  missing_tax <- is.na(base) | !nzchar(base)
  if (isTRUE(fallback.to.id)) {
    base[missing_tax] <- ids[missing_tax]
  }

  if (!is.null(aliases)) {
    if (is.null(names(aliases)) || any(!nzchar(names(aliases)))) {
      stop("linf.feature.labels: aliases must be a named character vector")
    }
    alias_lookup <- aliases
    original_tax <- as.character(taxonomy)
    for (ii in seq_along(base)) {
      key_orig <- original_tax[[ii]]
      key_norm <- base[[ii]]
      if (!is.na(key_orig) && key_orig %in% names(alias_lookup)) {
        base[[ii]] <- alias_lookup[[key_orig]]
      } else if (!is.na(key_norm) && key_norm %in% names(alias_lookup)) {
        base[[ii]] <- alias_lookup[[key_norm]]
      }
    }
  }

  if (!is.null(abbreviations)) {
    if (is.null(names(abbreviations)) || any(!nzchar(names(abbreviations)))) {
      stop("linf.feature.labels: abbreviations must be a named character vector")
    }
    for (ii in seq_along(base)) {
      parts <- strsplit(base[[ii]], "\\s+", perl = TRUE)[[1L]]
      if (!length(parts)) next
      genus <- parts[[1L]]
      if (genus %in% names(abbreviations)) {
        parts[[1L]] <- abbreviations[[genus]]
        base[[ii]] <- paste(parts, collapse = " ")
      }
    }
  }

  counts <- table(base)
  display <- base
  duplicated_taxa <- names(counts[counts > 1L])

  if (length(duplicated_taxa)) {
    if (duplicate.index == "global") {
      extracted <- sub("^.*?(\\d+)$", "\\1", ids)
      has_numeric_suffix <- grepl("\\d+$", ids)
      extracted[!has_numeric_suffix] <- as.character(seq_along(ids))[!has_numeric_suffix]
      for (taxon in duplicated_taxa) {
        idx <- which(base == taxon)
        display[idx] <- paste(base[idx], extracted[idx])
      }
    } else if (duplicate.index == "within_taxon") {
      for (taxon in duplicated_taxa) {
        idx <- which(base == taxon)
        display[idx] <- paste(base[idx], seq_along(idx))
      }
    }
  }

  make.unique(display, sep = "_")
}

resolve.linf.feature.meta <- function(X,
                                      feature.ids = NULL,
                                      feature.labels = NULL) {
  ids <- feature.ids
  if (is.null(ids)) {
    ids <- colnames(X)
  }
  if (is.null(ids)) {
    ids <- paste0("V", seq_len(ncol(X)))
  }
  ids <- make.unique(as.character(ids), sep = "_")

  labels <- feature.labels
  if (is.null(labels)) {
    labels <- ids
  }
  if (length(labels) != length(ids)) {
    stop("resolve.linf.feature.meta: feature.labels must have the same length as feature.ids")
  }
  labels <- make.unique(as.character(labels), sep = "_")

  list(feature.ids = ids, feature.labels = labels)
}

linf.resolve.backend <- function(X, backend = c("auto", "dense", "sparse")) {
  backend <- match.arg(backend)
  if (backend == "auto") {
    if (inherits(X, "sparseMatrix")) {
      return("sparse")
    }
    return("dense")
  }
  backend
}

linf.prepare.matrix <- function(X, backend = c("auto", "dense", "sparse"), fun.name = "linf") {
  backend <- linf.resolve.backend(X, backend)

  if (backend == "dense") {
    X <- as.matrix(X)
    storage.mode(X) <- "numeric"
    return(list(X = X, backend = backend))
  }

  if (!requireNamespace("Matrix", quietly = TRUE)) {
    stop(fun.name, ": sparse backend requires the 'Matrix' package")
  }

  if (inherits(X, "sparseMatrix")) {
    X <- as(X, "dgCMatrix")
  } else {
    X <- Matrix::Matrix(X, sparse = TRUE)
    X <- as(X, "dgCMatrix")
  }

  X <- Matrix::drop0(X)
  X@x <- as.numeric(X@x)
  list(X = X, backend = backend)
}

linf.validate.matrix <- function(X, backend, fun.name) {
  if (backend == "dense") {
    if (any(!is.finite(X))) {
      stop(fun.name, ": non-finite entries found")
    }
    if (any(X < 0, na.rm = TRUE)) {
      stop(fun.name, ": negative entries found")
    }
  } else {
    if (any(!is.finite(X@x))) {
      stop(fun.name, ": non-finite entries found")
    }
    if (any(X@x < 0, na.rm = TRUE)) {
      stop(fun.name, ": negative entries found")
    }
  }

  if (nrow(X) == 0L || ncol(X) == 0L) {
    stop(fun.name, ": matrix has zero rows or columns")
  }
}

linf.row.max.sparse <- function(X) {
  Xr <- as(X, "RsparseMatrix")
  p <- Xr@p
  x <- Xr@x
  out <- numeric(nrow(Xr))

  for (ii in seq_len(nrow(Xr))) {
    start <- p[[ii]] + 1L
    end <- p[[ii + 1L]]
    if (start <= end) {
      out[[ii]] <- max(x[start:end])
    }
  }

  out
}

linf.normalize.sparse <- function(X, tol = 0) {
  Xr <- as(X, "RsparseMatrix")
  p <- Xr@p
  row.max <- linf.row.max.sparse(Xr)
  keep.rows <- which(row.max > tol)

  if (length(keep.rows)) {
    for (ii in keep.rows) {
      start <- p[[ii]] + 1L
      end <- p[[ii + 1L]]
      if (start <= end) {
        Xr@x[start:end] <- Xr@x[start:end] / row.max[[ii]]
      }
    }
  }

  Matrix::drop0(as(Xr, "CsparseMatrix"))
}

linf.choose.index <- function(candidates, tie.method, error.message) {
  if (!length(candidates)) {
    return(NA_integer_)
  }
  if (length(candidates) == 1L) {
    return(as.integer(candidates[[1L]]))
  }
  if (tie.method == "first") {
    return(as.integer(min(candidates)))
  }
  if (tie.method == "random") {
    return(as.integer(sample(candidates, 1L)))
  }
  stop(error.message)
}

linf.cells.sparse <- function(S,
                              feature.ids = NULL,
                              feature.labels = NULL,
                              tie.method = c("first", "random", "error"),
                              return.value = FALSE) {
  tie.method <- match.arg(tie.method)

  prep <- linf.prepare.matrix(S, backend = "sparse", fun.name = "linf.cells")
  X <- prep$X
  linf.validate.matrix(X, backend = "sparse", fun.name = "linf.cells")

  meta <- resolve.linf.feature.meta(X, feature.ids = feature.ids, feature.labels = feature.labels)
  id.lev <- meta$feature.ids
  lev <- meta$feature.labels

  Xr <- as(X, "RsparseMatrix")
  p <- Xr@p
  j <- Xr@j
  x <- Xr@x

  idx <- rep(NA_integer_, nrow(Xr))
  val <- numeric(nrow(Xr))

  for (ii in seq_len(nrow(Xr))) {
    start <- p[[ii]] + 1L
    end <- p[[ii + 1L]]
    if (start > end) {
      next
    }

    row.vals <- x[start:end]
    row.cols <- j[start:end] + 1L
    row.max <- max(row.vals)
    val[[ii]] <- row.max

    ties <- row.cols[row.vals == row.max]
    idx[[ii]] <- linf.choose.index(
      ties,
      tie.method = tie.method,
      error.message = "linf.cells: tie encountered and tie.method = 'error'"
    )
  }

  id <- ifelse(is.na(idx), NA_character_, id.lev[idx])
  lbl <- ifelse(is.na(idx), NA_character_, lev[idx])

  observed.id.levels <- id.lev[id.lev %in% id[!is.na(id)]]
  observed.levels <- lev[lev %in% lbl[!is.na(lbl)]]

  out <- list(
    index = idx,
    id = id,
    label = lbl,
    id.levels = id.lev,
    levels = lev,
    observed.id.levels = observed.id.levels,
    observed.levels = observed.levels
  )

  if (return.value) {
    out$value <- val
  }

  out
}

linf.absorb.sparse <- function(X,
                               to.absorb,
                               kept.idx,
                               tie.method = c("first", "random", "error"),
                               fallback.idx) {
  tie.method <- match.arg(tie.method)

  Xr <- as(X, "RsparseMatrix")
  p <- Xr@p
  j <- Xr@j
  x <- Xr@x

  kept.map <- logical(ncol(Xr))
  kept.map[kept.idx] <- TRUE
  new.idx <- integer(length(to.absorb))

  for (kk in seq_along(to.absorb)) {
    ii <- to.absorb[[kk]]
    start <- p[[ii]] + 1L
    end <- p[[ii + 1L]]

    if (start > end) {
      new.idx[[kk]] <- fallback.idx
      next
    }

    row.cols <- j[start:end] + 1L
    row.vals <- x[start:end]
    keep.mask <- kept.map[row.cols]

    if (!any(keep.mask)) {
      new.idx[[kk]] <- fallback.idx
      next
    }

    kept.cols <- row.cols[keep.mask]
    kept.vals <- row.vals[keep.mask]
    row.max <- max(kept.vals)

    if (!is.finite(row.max) || row.max <= 0) {
      new.idx[[kk]] <- fallback.idx
      next
    }

    ties <- kept.cols[kept.vals == row.max]
    new.idx[[kk]] <- linf.choose.index(
      ties,
      tie.method = tie.method,
      error.message = "linf.csts: tie during reassignment and tie.method = 'error'"
    )
  }

  new.idx
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", fig.width = 7, fig.height = 5)
library(linf)

## ----load-data----------------------------------------------------------------
data(valencia2k)

## Compositional matrix: 2000 samples x 178 taxa
dim(valencia2k$rel)

## Valencia CST assignments
table(valencia2k$cst$Val_CST)

## Top 10 most abundant taxa (by mean relative abundance)
means <- sort(colMeans(valencia2k$rel), decreasing = TRUE)
head(round(means, 4), 10)

## ----reconstruct-counts-------------------------------------------------------
count_mat <- sweep(valencia2k$rel, 1, valencia2k$reads, "*")
count_mat <- round(count_mat)
storage.mode(count_mat) <- "integer"

## Check: library sizes
summary(rowSums(count_mat))

## ----linf-normalize-----------------------------------------------------------
Z <- normalize.linf(valencia2k$rel)

## Every row max should be 1
summary(apply(Z, 1, max))

## ----linf-cells---------------------------------------------------------------
cells <- linf.cells(Z)

## How many distinct dominant taxa?
cat("Distinct dominant taxa:", length(cells$observed.levels), "\n")

## Dominance sample-set size distribution (top 15)
dominance_tab <- sort(table(cells$label), decreasing = TRUE)
head(dominance_tab, 15)

## ----depth1-dcsts-------------------------------------------------------------
dcst1 <- linf.csts(Z, n0 = 30, low.freq.policy = "pure")

## Retained depth-1 states
dcst1$kept.cells.lbl

## dCST assignment (pure policy)
dcst1_tab <- sort(table(dcst1$cell.label), decreasing = TRUE)
dcst1_tab

## ----concordance-depth1-------------------------------------------------------
## Build concordance table: rows = dCSTs, columns = Valencia CSTs
dcst_labels <- dcst1$cell.label.absorb  # use absorb view for clean comparison
val_cst     <- valencia2k$cst$Val_CST

concordance <- table(dCST = dcst_labels, Valencia = val_cst)

## Show as percentage: what fraction of each dCST falls in each Valencia CST?
pct <- round(100 * prop.table(concordance, margin = 1), 1)
pct

## ----concordance-reverse------------------------------------------------------
## Reverse view: what fraction of each Valencia CST falls in each dCST?
pct_rev <- round(100 * prop.table(concordance, margin = 2), 1)
pct_rev

## ----depth2-refine------------------------------------------------------------
dcst2 <- refine.linf.csts(Z, dcst1, n0 = 15, refinement.factor = 2,
                           low.freq.policy = "pure", verbose = FALSE)

## Depth-2 dCST table
dcst2_tab <- sort(table(dcst2$cell.label), decreasing = TRUE)
dcst2_tab

## ----concordance-depth2-------------------------------------------------------
dcst2_labels <- dcst2$cell.label.absorb
val_subcst   <- valencia2k$cst$Val_subCST

concordance2 <- table(dCST = dcst2_labels, Valencia = val_subcst)

## Percentage by dCST (row-wise)
pct2 <- round(100 * prop.table(concordance2, margin = 1), 1)

## Show only dCSTs with at least 20 samples for readability
dcst2_sizes <- rowSums(concordance2)
pct2[dcst2_sizes >= 20, , drop = FALSE]

## ----full-pipeline------------------------------------------------------------
out <- linf.dcst.landmark.pipeline(
  count_mat,
  feature.ids    = colnames(count_mat),
  feature.labels = colnames(count_mat),
  n0.depth1      = 30,
  n0.depth2      = 15,
  refinement.factor = 2,
  low.freq.policy   = "pure",
  landmark.view     = "absorb",
  verbose           = FALSE
)

names(out)

## ----landmarks----------------------------------------------------------------
## Depth-1 landmarks
lm1 <- out$landmarks.depth1$landmarks

## Show landmarks for a few key dCST dominance-lineages
key_lineages <- c("Lactobacillus_crispatus", "Lactobacillus_iners",
                  "Gardnerella_vaginalis")
key_lm <- lm1[lm1$cell.id %in% key_lineages, ]

## For each dCST, show the endpoint.max landmark's target value
## (how dominant is the dominant species in the most extreme sample?)
ep_max <- key_lm[key_lm$landmark.type == "endpoint.max",
                 c("cell.label", "point.name", "target.value")]
ep_max

## And the endpoint.min (least dominant while still assigned to this lineage)
ep_min <- key_lm[key_lm$landmark.type == "endpoint.min",
                 c("cell.label", "point.name", "target.value")]
ep_min

#' Extended homogeneous-coordinate hypercube embedding
#'
#' @description
#' Computes the zero-aware hypercube embedding associated with one reference
#' component of a nonnegative compositional matrix. For rows with positive
#' reference component, the function forms the ordinary homogeneous ratios
#' against that reference and radially maps them into the unit cube. For rows
#' whose reference component is zero, it uses the L-infinity boundary extension
#' so that the embedding remains defined.
#'
#' @details
#' Let \eqn{x = (x_1,\ldots,x_p)} be a nonnegative row and let \eqn{k} be the
#' reference component. When \eqn{x_k > 0}, define
#' \eqn{z = x_{-k}/x_k}. The embedded row is
#' \deqn{
#'   \sigma_\lambda(\|z\|_1)\frac{z}{\|z\|_\infty},
#'   \qquad
#'   \sigma_\lambda(t) = 1 - \exp(-\lambda t).
#' }
#' When \eqn{x_k = 0}, the embedded row is the L-infinity-normalized boundary
#' vector
#' \deqn{
#'   x_{-k}/\|x_{-k}\|_\infty.
#' }
#' All-zero rows are mapped to all-zero embedded rows by convention.
#'
#' If `lambda` is not supplied, it is chosen from the positive finite-reference
#' rows so that `sigma.target` is attained at the `sigma.quantile` quantile of
#' \eqn{\|z\|_1}. This is a numerical scaling convention for finite datasets; it
#' does not change the reference component or the boundary extension rule.
#'
#' @param X Nonnegative numeric matrix with samples in rows and features in
#'   columns.
#' @param reference Reference component. May be a column index, feature ID, or
#'   feature label.
#' @param lambda Positive numeric scalar. If `NULL`, choose a data-scaled value
#'   using `sigma.quantile` and `sigma.target`.
#' @param sigma.quantile Quantile of positive finite-reference \eqn{\|z\|_1}
#'   values used when `lambda = NULL`.
#' @param sigma.target Target value of \eqn{\sigma_\lambda(t)} at the selected
#'   quantile when `lambda = NULL`.
#' @param feature.ids Optional stable feature identifiers, length `ncol(X)`.
#' @param feature.labels Optional display labels, length `ncol(X)`.
#' @param tol Nonnegative tolerance. Reference entries `<= tol` are treated as
#'   zero, and L-infinity norms `<= tol` are treated as zero.
#' @param backend Matrix backend: `"auto"`, `"dense"`, or `"sparse"`. Sparse
#'   inputs are accepted, but the returned embedding is a dense matrix because
#'   homogeneous-coordinate embeddings are generally dense.
#'
#' @return A numeric matrix with `nrow(X)` rows and `ncol(X) - 1` columns. The
#'   columns correspond to the non-reference components. Attributes record the
#'   reference component, lambda choice, and finite/boundary row counts.
#'
#' @examples
#' X <- rbind(
#'   c(A = 2, B = 1, C = 1),
#'   c(A = 0, B = 2, C = 1)
#' )
#' linf.hypercube.embedding(X, reference = "A", lambda = log(2))
#'
#' @export
linf.hypercube.embedding <- function(X,
                                     reference,
                                     lambda = NULL,
                                     sigma.quantile = 0.95,
                                     sigma.target = 0.95,
                                     feature.ids = NULL,
                                     feature.labels = NULL,
                                     tol = 0,
                                     backend = c("auto", "dense", "sparse")) {
  prep <- linf.prepare.matrix(X, backend = backend, fun.name = "linf.hypercube.embedding")
  X <- prep$X
  backend <- prep$backend
  linf.validate.matrix(X, backend = backend, fun.name = "linf.hypercube.embedding")

  if (backend == "sparse") {
    X <- as.matrix(X)
  }

  if (!is.numeric(tol) || length(tol) != 1L || !is.finite(tol) || tol < 0) {
    stop("linf.hypercube.embedding: tol must be a single non-negative finite number")
  }
  if (!is.numeric(sigma.quantile) || length(sigma.quantile) != 1L ||
      !is.finite(sigma.quantile) || sigma.quantile <= 0 || sigma.quantile > 1) {
    stop("linf.hypercube.embedding: sigma.quantile must be in (0, 1]")
  }
  if (!is.numeric(sigma.target) || length(sigma.target) != 1L ||
      !is.finite(sigma.target) || sigma.target <= 0 || sigma.target >= 1) {
    stop("linf.hypercube.embedding: sigma.target must be in (0, 1)")
  }

  meta <- resolve.linf.feature.meta(
    X,
    feature.ids = feature.ids,
    feature.labels = feature.labels
  )
  feature.ids <- meta$feature.ids
  feature.labels <- meta$feature.labels

  ref.idx <- linf.resolve.reference.index(reference, feature.ids, feature.labels)
  ref.id <- feature.ids[[ref.idx]]
  ref.label <- feature.labels[[ref.idx]]
  other.idx <- setdiff(seq_len(ncol(X)), ref.idx)
  other.ids <- feature.ids[other.idx]
  other.labels <- feature.labels[other.idx]

  denom <- X[, ref.idx]
  others <- X[, other.idx, drop = FALSE]
  finite <- denom > tol

  z.norm1 <- rep(NA_real_, nrow(X))
  if (any(finite)) {
    z.norm1[finite] <- rowSums(others[finite, , drop = FALSE] / denom[finite])
  }

  lambda.policy <- "fixed"
  if (is.null(lambda)) {
    lambda.policy <- "quantile"
    positive.norms <- z.norm1[is.finite(z.norm1) & z.norm1 > tol]
    if (length(positive.norms)) {
      q <- stats::quantile(
        positive.norms,
        probs = sigma.quantile,
        names = FALSE,
        na.rm = TRUE,
        type = 7
      )
      lambda <- if (is.finite(q) && q > tol) -log(1 - sigma.target) / q else 1
    } else {
      lambda <- 1
    }
  }
  if (!is.numeric(lambda) || length(lambda) != 1L || !is.finite(lambda) || lambda <= 0) {
    stop("linf.hypercube.embedding: lambda must be a single positive finite number")
  }

  out <- matrix(0, nrow = nrow(X), ncol = length(other.idx))
  rownames(out) <- rownames(X)
  colnames(out) <- paste0(other.labels, "_rel_", ref.label)

  if (any(finite)) {
    z <- others[finite, , drop = FALSE] / denom[finite]
    z.inf <- apply(z, 1L, max)
    z.one <- rowSums(z)
    direction <- matrix(0, nrow = nrow(z), ncol = ncol(z))
    nz <- z.inf > tol
    direction[nz, ] <- z[nz, , drop = FALSE] / z.inf[nz]
    out[finite, ] <- direction * (1 - exp(-lambda * z.one))
  }

  boundary <- !finite
  if (any(boundary)) {
    boundary.x <- others[boundary, , drop = FALSE]
    boundary.inf <- apply(boundary.x, 1L, max)
    boundary.out <- matrix(0, nrow = nrow(boundary.x), ncol = ncol(boundary.x))
    nz <- boundary.inf > tol
    boundary.out[nz, ] <- boundary.x[nz, , drop = FALSE] / boundary.inf[nz]
    out[boundary, ] <- boundary.out
  }

  attr(out, "reference.index") <- ref.idx
  attr(out, "reference.id") <- ref.id
  attr(out, "reference.label") <- ref.label
  attr(out, "other.ids") <- other.ids
  attr(out, "other.labels") <- other.labels
  attr(out, "lambda") <- lambda
  attr(out, "lambda.policy") <- lambda.policy
  attr(out, "sigma.quantile") <- sigma.quantile
  attr(out, "sigma.target") <- sigma.target
  attr(out, "finite.reference.count") <- sum(finite)
  attr(out, "zero.reference.count") <- sum(!finite)
  out
}

linf.resolve.reference.index <- function(reference, feature.ids, feature.labels) {
  if (missing(reference) || length(reference) != 1L || is.na(reference)) {
    stop("linf.hypercube.embedding: reference must identify exactly one component")
  }

  if (is.numeric(reference)) {
    if (reference %% 1 != 0 || reference < 1 || reference > length(feature.ids)) {
      stop("linf.hypercube.embedding: numeric reference is out of range")
    }
    return(as.integer(reference))
  }

  ref <- as.character(reference)
  id.match <- which(feature.ids == ref)
  label.match <- which(feature.labels == ref)
  idx <- unique(c(id.match, label.match))

  if (length(idx) != 1L) {
    stop("linf.hypercube.embedding: reference must match exactly one feature ID or label")
  }
  idx
}

test_that("linf.hypercube.embedding handles finite ratios and boundary rows", {
  X <- rbind(
    finite = c(A = 2, B = 1, C = 1),
    boundary = c(A = 0, B = 2, C = 1),
    zero = c(A = 0, B = 0, C = 0),
    reference_only = c(A = 1, B = 0, C = 0)
  )

  emb <- linf.hypercube.embedding(X, reference = "A", lambda = log(2))

  expect_equal(dim(emb), c(4L, 2L))
  expect_equal(colnames(emb), c("B_rel_A", "C_rel_A"))
  expect_equal(unname(emb["finite", ]), c(0.5, 0.5), tolerance = 1e-12)
  expect_equal(unname(emb["boundary", ]), c(1, 0.5), tolerance = 1e-12)
  expect_equal(unname(emb["zero", ]), c(0, 0), tolerance = 1e-12)
  expect_equal(unname(emb["reference_only", ]), c(0, 0), tolerance = 1e-12)

  expect_equal(attr(emb, "reference.index"), 1L)
  expect_equal(attr(emb, "reference.id"), "A")
  expect_equal(attr(emb, "lambda"), log(2))
  expect_equal(attr(emb, "lambda.policy"), "fixed")
  expect_equal(attr(emb, "finite.reference.count"), 2L)
  expect_equal(attr(emb, "zero.reference.count"), 2L)
})

test_that("linf.hypercube.embedding is projectively scale invariant", {
  X <- rbind(
    c(A = 3, B = 6, C = 9),
    c(A = 0, B = 4, C = 8)
  )
  Y <- X * c(2, 7)

  emb.x <- linf.hypercube.embedding(X, reference = "A", lambda = 0.25)
  emb.y <- linf.hypercube.embedding(Y, reference = "A", lambda = 0.25)

  expect_equal(unname(emb.x), unname(emb.y), tolerance = 1e-12)
})

test_that("linf.hypercube.embedding resolves feature labels and auto lambda", {
  X <- rbind(
    c(2, 1, 1),
    c(4, 2, 2)
  )
  colnames(X) <- c("asv1", "asv2", "asv3")

  emb <- linf.hypercube.embedding(
    X,
    reference = "Li",
    feature.labels = c("Li", "Lc", "Gv"),
    sigma.quantile = 1,
    sigma.target = 0.5
  )

  expect_equal(attr(emb, "reference.id"), "asv1")
  expect_equal(attr(emb, "reference.label"), "Li")
  expect_equal(attr(emb, "lambda"), log(2), tolerance = 1e-12)
  expect_equal(unname(emb[1, ]), c(0.5, 0.5), tolerance = 1e-12)
})

test_that("linf.hypercube.embedding agrees for dense and sparse inputs", {
  skip_if_not_installed("Matrix")

  X <- rbind(
    c(A = 2, B = 1, C = 1),
    c(A = 0, B = 2, C = 1),
    c(A = 1, B = 0, C = 0)
  )

  dense <- linf.hypercube.embedding(X, reference = "A", lambda = log(2))
  sparse <- linf.hypercube.embedding(Matrix::Matrix(X, sparse = TRUE),
                                     reference = "A", lambda = log(2))

  expect_equal(unname(dense), unname(sparse), tolerance = 1e-12)
  expect_equal(colnames(sparse), colnames(dense))
})

test_that("linf.hypercube.embedding validates inputs", {
  expect_error(
    linf.hypercube.embedding(matrix(c(1, -1), nrow = 1), reference = 1),
    "negative entries"
  )
  expect_error(
    linf.hypercube.embedding(matrix(c(1, 2), nrow = 1), reference = 3),
    "out of range"
  )
  expect_error(
    linf.hypercube.embedding(matrix(c(1, 2), nrow = 1), reference = 1, lambda = 0),
    "lambda"
  )
})

#' Valencia 2k vaginal microbiome dataset
#'
#' A stratified subsample of 2,000 vaginal samples from the Valencia 13k
#' CST-classifier training set (France et al. 2020), bundled as an example
#' dataset for demonstrating dCST construction after L-infinity normalization.
#'
#' The subsample preserves proportional representation of all 13 Valencia
#' sub-CSTs and was drawn with \code{set.seed(42)}.
#'
#' @format A list with five components:
#' \describe{
#'   \item{rel}{Numeric matrix (2000 x 178). Compositional relative abundances;
#'     each row sums to 1. Rows are samples, columns are taxonomic features.}
#'   \item{cst}{Data frame (2000 x 3) with columns:
#'     \code{sample_id} (character),
#'     \code{Val_CST} (Valencia CST assignment: I, II, III, IV-A, IV-B, IV-C, V),
#'     \code{Val_subCST} (Valencia sub-CST assignment: I-A, I-B, II, III-A, III-B,
#'       IV-A, IV-B, IV-C0, IV-C1, IV-C2, IV-C3, IV-C4, V).}
#'   \item{reads}{Integer vector of length 2000. Per-sample read counts after
#'     taxonomic filtering. Use \code{sweep(valencia2k$rel, 1, valencia2k$reads, "*")}
#'     to reconstruct a count-like matrix.}
#'   \item{taxa}{Character vector of 178 taxon names (column names of \code{rel}).}
#'   \item{source}{Character string documenting provenance.}
#' }
#'
#' @details
#' The Valencia CST classifier (France et al. 2020) assigns vaginal microbiome
#' samples to community state types (CSTs) based on nearest-centroid
#' classification in relative-abundance space. The original training set
#' contains 12,881 samples and 178 taxonomic features after filtering.
#'
#' This 2,000-sample subsample is intended for vignette demonstrations. The
#' compositional matrix can be used directly with \code{\link{normalize.linf}}
#' and downstream dCST functions. For workflows that require count-like input,
#' reconstruct approximate counts using the \code{reads} vector.
#'
#' @references
#' France, M. T., Ma, B., Gajer, P., Brown, S., Humphrys, M. S., Holm, J. B.,
#' Waetjen, L. E., Brotman, R. M., & Ravel, J. (2020). VALENCIA: a nearest
#' centroid classification method for vaginal microbial communities based on
#' composition. \emph{Microbiome}, 8(1), 166.
#' \doi{10.1186/s40168-020-00934-6}
#'
#' @source
#' Subsampled from the VALENCIA training data at
#' \url{https://github.com/ravel-lab/VALENCIA}. See
#' \code{data-raw/build_valencia2k.R} and the installed
#' \code{DATA_PROVENANCE.md} file for construction and licensing details.
#'
#' @examples
#' data(valencia2k)
#' dim(valencia2k$rel)           # 2000 x 178
#' table(valencia2k$cst$Val_CST) # CST distribution
#' head(valencia2k$taxa, 10)     # first 10 taxon names
#'
"valencia2k"


#' Valencia 1k four-component hypercube embedding example
#'
#' A stratified 1,000-sample subset of the Valencia 13k vaginal microbiome
#' training set, reduced to four selected phylotype coordinates and
#' L1-normalized over those coordinates. The object is bundled as a lightweight
#' example for the zero-aware homogeneous-coordinate hypercube embedding in
#' \code{\link{linf.hypercube.embedding}}.
#'
#' @format A list with four components:
#' \describe{
#'   \item{rel4}{Numeric matrix (1000 x 4). Rows are samples and columns are
#'     \code{Li}, \code{Lc}, \code{Gv}, and \code{Bv}. Each row sums to 1 after
#'     restricting the original Valencia profile to the four mapped taxa.}
#'   \item{meta}{Data frame (1000 rows) with columns:
#'     \code{sample_id} (anonymized bundled-data identifier),
#'     \code{source_row} (row number in the filtered Valencia 13k source),
#'     \code{Val_CST}, \code{Val_subCST}, \code{selected_mass} (original
#'     relative-abundance mass carried by the four selected taxa), and
#'     \code{dominant_component} (largest of \code{Li}, \code{Lc},
#'     \code{Gv}, \code{Bv} after renormalization).}
#'   \item{component_map}{Named character vector mapping \code{Li},
#'     \code{Lc}, \code{Gv}, and \code{Bv} to the original Valencia taxon names:
#'     \code{Lactobacillus_iners}, \code{Lactobacillus_crispatus},
#'     \code{Gardnerella_vaginalis}, and \code{BVAB1}.}
#'   \item{source}{Character string documenting provenance.}
#' }
#'
#' @details
#' Rows with zero total mass in the four selected taxa are removed before
#' renormalization. Sampling is stratified by the dominant selected component,
#' using \code{set.seed(20261604)}. The object is not intended to replace the
#' full Valencia matrix; it is a compact reproducible example for visualizing
#' compositional projective-space coordinate charts.
#'
#' @source
#' Generated from the VALENCIA training data at
#' \url{https://github.com/ravel-lab/VALENCIA}. See
#' \code{data-raw/build_valencia_linf_hypercube_1k.R} and the installed
#' \code{DATA_PROVENANCE.md} file.
#'
#' @examples
#' data(valencia_linf_hypercube_1k)
#' dim(valencia_linf_hypercube_1k$rel4)
#' table(valencia_linf_hypercube_1k$meta$dominant_component)
#' emb <- linf.hypercube.embedding(
#'   valencia_linf_hypercube_1k$rel4,
#'   reference = "Li"
#' )
#' dim(emb)
#'
"valencia_linf_hypercube_1k"


#' Valencia 13k merged depth-2 dCST assignments
#'
#' A lightweight bundled assignment asset containing merged depth-2
#' Dominant community state type (dCST) labels for the full Valencia
#' 13k vaginal microbiome training set.
#'
#' @format A list with five components:
#' \describe{
#'   \item{assignments}{Data frame with one row per source sample and columns
#'     \code{sample_id} (anonymized bundled-data identifier), \code{source_row}
#'     (row number in the filtered Valencia 13k source), \code{Val_CST},
#'     \code{Val_subCST}, \code{dcst_depth1}, and \code{dcst_depth2}.}
#'   \item{summaries}{Named list of per-depth summary tables. Each table
#'     contains \code{depth}, \code{dcst_label}, \code{n}, \code{prop}, and
#'     \code{path_length}.}
#'   \item{feature_labels}{Character vector of source taxon labels.}
#'   \item{params}{List recording the construction parameters.}
#'   \item{source}{Character string documenting provenance.}
#' }
#'
#' @details
#' The asset is computed from the Valencia 13k compositional matrix after
#' L-infinity normalization. dCSTs use \code{n0 = 50} and the merged
#' \code{low.freq.policy = "absorb"} view, so samples from low-support
#' provisional states are reassigned to retained states rather than stored as
#' explicit rare buckets.
#'
#' @source
#' Generated from the VALENCIA training data at
#' \url{https://github.com/ravel-lab/VALENCIA}. See
#' \code{data-raw/build_valencia13k_merged_dcst_depths.R} and the installed
#' \code{DATA_PROVENANCE.md} file.
#'
#' @examples
#' data(valencia13k_dcst_depth2_merged)
#' nrow(valencia13k_dcst_depth2_merged$assignments)
#' head(valencia13k_dcst_depth2_merged$summaries$depth2)
#'
"valencia13k_dcst_depth2_merged"


#' Valencia 13k merged depth-3 dCST assignments
#'
#' A lightweight bundled assignment asset containing merged depth-3
#' Dominant community state type (dCST) labels for the full Valencia
#' 13k vaginal microbiome training set.
#'
#' @format A list with five components:
#' \describe{
#'   \item{assignments}{Data frame with one row per source sample and columns
#'     \code{sample_id} (anonymized bundled-data identifier), \code{source_row}
#'     (row number in the filtered Valencia 13k source), \code{Val_CST},
#'     \code{Val_subCST}, \code{dcst_depth1}, \code{dcst_depth2}, and
#'     \code{dcst_depth3}.}
#'   \item{summaries}{Named list of per-depth summary tables. Each table
#'     contains \code{depth}, \code{dcst_label}, \code{n}, \code{prop}, and
#'     \code{path_length}.}
#'   \item{feature_labels}{Character vector of source taxon labels.}
#'   \item{params}{List recording the construction parameters.}
#'   \item{source}{Character string documenting provenance.}
#' }
#'
#' @details
#' The depth-3 asset extends \code{\link{valencia13k_dcst_depth2_merged}} by one
#' additional hierarchical dCST refinement. All levels use \code{n0 = 50} and
#' \code{low.freq.policy = "absorb"}. This object is intended as a reusable
#' source for selecting richer VALENCIA-derived component sets without
#' recomputing the full hierarchy from the 13k source matrix.
#'
#' @source
#' Generated from the VALENCIA training data at
#' \url{https://github.com/ravel-lab/VALENCIA}. See
#' \code{data-raw/build_valencia13k_merged_dcst_depths.R} and the installed
#' \code{DATA_PROVENANCE.md} file.
#'
#' @examples
#' data(valencia13k_dcst_depth3_merged)
#' nrow(valencia13k_dcst_depth3_merged$assignments)
#' head(valencia13k_dcst_depth3_merged$summaries$depth3)
#'
"valencia13k_dcst_depth3_merged"


#' American Gut Project gut microbiome dataset
#'
#' A stratified subsample of 766 gut microbiome samples from the American Gut
#' Project (PRJEB11419, AGP-US-2015), bundled for demonstrating L-infinity
#' dCST construction in a gut ecosystem.
#'
#' The subsample includes all samples assigned to four uncommon demonstration
#' dCSTs (Prevotella_7, Pasteurellaceae, Akkermansia, and Staphylococcus).
#' The remaining slots are a simple random sample, drawn with seed 42, from
#' the eligible background after excluding Eukaryota and Unassigned labels.
#' Phenotype fields are joined only after membership is fixed and do not
#' influence selection. Because inclusion probabilities differ by dCST, this
#' object is a computational demonstration dataset rather than a probability
#' sample of the underlying cohort.
#'
#' @format A list with four components:
#' \describe{
#'   \item{counts}{Integer matrix (766 x 314). Raw 16S V4 read counts.
#'     Rows are samples, columns are SILVA species-level taxa.}
#'   \item{meta}{Data frame (766 rows) with columns:
#'     \code{Run} (SRA run accession),
#'     \code{dcst_depth1}, \code{dcst_depth2} (pre-computed dCST labels),
#'     \code{IBS}, \code{IBD}, \code{Diabetes}, \code{Autoimmune},
#'     \code{Seasonal_allergies}, \code{Migraine}, \code{Acid_reflux},
#'     \code{Lung_disease}, \code{Cardiovascular_disease}, \code{Skin_condition},
#'     \code{Obesity} (binary disease indicators from self-reported AGP metadata),
#'     \code{BMI} (numeric, self-reported), and
#'     \code{selection_reason} (target-dCST inclusion or seeded background
#'     sampling).}
#'   \item{taxa}{Character vector of 314 SILVA taxonomy strings.}
#'   \item{source}{Character string documenting provenance.}
#' }
#'
#' @details
#' The American Gut Project is a large citizen-science 16S rRNA survey of the
#' human microbiome. Health conditions are self-reported via questionnaire and
#' should be interpreted with appropriate caution.
#'
#' The phenotype fields must not be used with this dCST-stratified subset for
#' population prevalence estimates, effect-size estimation, or association
#' testing. The exact selection is generated by
#' \code{data-raw/create_agp_gut_subset.py}; run-ID membership, selection
#' reasons, and derived annotations are retained in
#' \code{inst/extdata/agp_gut_meta.csv}.
#'
#' The count matrix can be used directly with \code{\link{filter.asv}},
#' \code{\link{normalize.linf}}, and downstream dCST functions.
#'
#' Note on Escherichia-Shigella: this genus is inflated in 16S V4 data due to
#' primer cross-reactivity and should be interpreted with caution.
#'
#' @references
#' McDonald, D., Hyde, E., Debelius, J. W., et al. (2018). American Gut: an
#' Open Platform for Citizen Science Microbiome Research. \emph{mSystems},
#' 3(3), e00031-18. \doi{10.1128/mSystems.00031-18}
#'
#' @source
#' Derived from the public American Gut Project records under ENA accession
#' \href{https://www.ebi.ac.uk/ena/browser/view/PRJEB11419}{PRJEB11419} via the
#' PRIME pipeline. See \code{data-raw/create_agp_gut_subset.py},
#' \code{data-raw/build_agp_gut.R}, and the installed
#' \code{DATA_PROVENANCE.md} file.
#'
#' @examples
#' data(agp_gut)
#' dim(agp_gut$counts)                     # 766 x 314
#' table(agp_gut$meta$dcst_depth1)         # dCST distribution
#' sum(agp_gut$meta$IBS)                   # IBS cases
#'
"agp_gut"

# L-infinity normalization utilities

#' L-infinity normalization (row-wise)
#'
#' Scales each row of a numeric matrix by its L-infinity norm (row maximum).
#' Rows whose maximum is zero (or below tolerance) are left unchanged
#' and remain all-zero.
#'
#' @param X Numeric matrix (samples x features).
#' @param tol Numeric >= 0. Values with row max <= tol are treated as zero rows.
#'   Default: 0 (exact zero only).
#' @param backend Character. Matrix backend to use: \code{"auto"},
#'   \code{"dense"}, or \code{"sparse"}. The default \code{"auto"} preserves
#'   sparse input and otherwise uses the dense path.
#'
#' @return Numeric matrix of same dimensions as X, L-infinity normalized.
#'
#' @details
#' Zero rows have undefined L-infinity direction. By convention, they are
#' preserved as all-zero rows and will yield NA labels in downstream
#' dominant-feature or dCST assignment.
#'
#' @export
normalize.linf <- function(X,
                           tol = 0,
                           backend = c("auto", "dense", "sparse")) {
  prep <- linf.prepare.matrix(X, backend = backend, fun.name = "normalize.linf")
  X <- prep$X
  backend <- prep$backend

  linf.validate.matrix(X, backend = backend, fun.name = "normalize.linf")
  if (!is.numeric(tol) || length(tol) != 1L || tol < 0) {
    stop("normalize.linf: tol must be a single non-negative number")
  }

  if (backend == "sparse") {
    return(linf.normalize.sparse(X, tol = tol))
  }

  m <- apply(X, 1, max)

  ## Identify rows with meaningful L-infinity norm
  keep <- m > tol

  if (any(keep)) {
    X[keep, ] <- X[keep, , drop = FALSE] / m[keep]
  }

  X
}

#' Dominant-feature assignment
#'
#' @description
#' Assigns each row to the column achieving its maximum.
#'
#' For each sample (row) of a nonnegative matrix, identifies the dominant
#' feature as the column with the maximum value. Samples with the same dominant
#' feature form a depth-1 dominance sample set. Ties are broken by the first
#' maximum (as in \code{max.col(..., ties.method = "first")}). Rows that are
#' all zero are assigned \code{NA}.
#'
#' Feature IDs default to \code{colnames(S)}; if absent, synthetic IDs
#' \code{"V1", "V2", ..., "Vp"} are generated. Display labels default to the
#' feature IDs unless \code{feature.labels} is supplied. To guarantee a 1-1
#' mapping between columns and both IDs and labels, duplicates are
#' disambiguated via \code{make.unique()}.
#'
#' @param S Numeric matrix (samples x features), typically L-infinity-normalized.
#' @param feature.ids Optional character vector of stable feature identifiers,
#'   length \code{ncol(S)}.
#' @param feature.labels Optional character vector of display labels, length
#'   \code{ncol(S)}.
#' @param tie.method Character. How to resolve ties during dominant-feature
#'   assignment.
#' @param return.value Logical. If `TRUE`, include a `value` vector with row maxima.
#' @param backend Character. Matrix backend to use: \code{"auto"},
#'   \code{"dense"}, or \code{"sparse"}. The default \code{"auto"} preserves
#'   sparse input and otherwise uses the dense path.
#'
#' @return A list with components:
#' \itemize{
#'   \item \code{index}: integer index of the dominant column per sample (`NA` for all-zero rows)
#'   \item \code{id}: dominant feature ID per sample (`NA` for all-zero rows)
#'   \item \code{label}: dominant column label per sample (`NA` for all-zero rows)
#'   \item \code{id.levels}: full feature ID set after `make.unique(..., sep = "_")`
#'   \item \code{levels}: full column label set after `make.unique(..., sep = "_")`
#'   \item \code{observed.id.levels}: subset of \code{id.levels} that appear in \code{id}
#'   \item \code{observed.levels}: subset of \code{levels} that appear in \code{label}
#'   \item \code{value}: row maxima (only when \code{return.value = TRUE})
#' }
#'
#' @examples
#' # Basic example with named columns
#' S <- rbind(
#'   a = c(A = 10, B = 5,  C = 0),   # -> A
#'   b = c(A = 0,  B = 0,  C = 0),   # -> NA
#'   c = c(A = 1,  B = 4,  C = 4)    # tie -> first max: B
#' )
#' out <- linf.cells(S)
#' out$index
#' out$label
#' out$levels
#' out$observed.levels
#'
#' # Unnamed columns (synthetic labels V1..Vp), duplicate names disambiguated
#' T <- matrix(c(0,2,  3,1,  0,0), nrow = 3, byrow = TRUE)
#' colnames(T) <- c("X", "X")  # duplicates -> X, X_1
#' linf.cells(T)$levels
#'
#' # With L-infinity normalization in a pipeline
#' M <- normalize.linf(S)
#' linf.cells(M)$label
#'
#' @seealso \code{\link{normalize.linf}}, \code{\link{linf.csts}}
#' @export
linf.cells <- function(S,
                       feature.ids = NULL,
                       feature.labels = NULL,
                       tie.method = c("first", "random", "error"),
                       return.value = FALSE,
                       backend = c("auto", "dense", "sparse")) {

  tie.method <- match.arg(tie.method)
  backend <- linf.resolve.backend(S, backend)

  if (backend == "sparse") {
    return(linf.cells.sparse(
      S,
      feature.ids = feature.ids,
      feature.labels = feature.labels,
      tie.method = tie.method,
      return.value = return.value
    ))
  }

  prep <- linf.prepare.matrix(S, backend = "dense", fun.name = "linf.cells")
  X <- prep$X
  linf.validate.matrix(X, backend = "dense", fun.name = "linf.cells")

  meta <- resolve.linf.feature.meta(X, feature.ids = feature.ids, feature.labels = feature.labels)
  id.lev <- meta$feature.ids
  lev <- meta$feature.labels

  idx <- rep(NA_integer_, nrow(X))
  val <- apply(X, 1, max)

  nz <- val > 0
  if (any(nz)) {
    idx[nz] <- apply(X[nz, , drop = FALSE], 1, function(r) {
      m <- max(r)
      j <- which(r == m)
      if (length(j) == 1L) return(j)
      if (tie.method == "first") return(j[1L])
      if (tie.method == "random") return(sample(j, 1L))
      stop("linf.cells: tie encountered and tie.method = 'error'")
    })
  }

  id <- ifelse(is.na(idx), NA_character_, id.lev[idx])
  lbl <- ifelse(is.na(idx), NA_character_, lev[idx])

  observed.id.levels <- id.lev[id.lev %in% id[!is.na(id)]]
  observed.levels <- lev[lev %in% lbl[!is.na(lbl)]]

  out <- list(
    index = idx,
    id = id,
    label = lbl,
    id.levels = id.lev,
    levels = lev,
    observed.id.levels = observed.id.levels,
    observed.levels = observed.levels
  )

  if (return.value) {
    out$value <- val
  }

  out
}

linf.normalize.low.freq.policy <- function(low.freq.policy, fun.name) {
  if (length(low.freq.policy) > 1L) {
    low.freq.policy <- low.freq.policy[[1L]]
  }

  low.freq.policy <- match.arg(
    low.freq.policy,
    choices = c("pure", "absorb", "rare")
  )

  if (identical(low.freq.policy, "rare")) {
    warning(
      sprintf(
        '%s: low.freq.policy = "rare" is deprecated; use "pure" instead',
        fun.name
      ),
      call. = FALSE
    )
    return("pure")
  }

  low.freq.policy
}

linf.active.low.freq.view <- function(low.freq.policy) {
  if (is.null(low.freq.policy)) {
    return("active")
  }

  if (identical(low.freq.policy, "pure") || identical(low.freq.policy, "rare")) {
    return("rare")
  }

  low.freq.policy
}

#' Truncated dominant community state types with configurable low-support handling
#'
#' @description
#' Forms provisional depth-1 dominance sample sets from the dominant feature
#' of each sample and then applies the minimum support threshold \code{n0}.
#' Sets with fewer than \code{n0} samples are handled according to
#' \code{low.freq.policy}:
#' \itemize{
#'   \item \code{"pure"}: retain only sets with support >= \code{n0} as named
#'     dCSTs and collapse all low-support sets into \code{rare.label}.
#'   \item \code{"absorb"}: reassign each low-support sample to the retained
#'     state with the largest value among the retained features (ties handled
#'     by \code{tie.method}).
#' }
#'
#' @param S Numeric matrix (samples x features), typically L-infinity relatives.
#' @param feature.ids Optional character vector of stable feature identifiers,
#'   length \code{ncol(S)}.
#' @param feature.labels Optional character vector of display labels, length
#'   \code{ncol(S)}.
#' @param n0 Integer >= 1. Minimum support required to retain a dominance
#'   sample set.
#' @param low.freq.policy Character. One of \code{"pure"} or \code{"absorb"}.
#'   Default: \code{"pure"}. The legacy value \code{"rare"} is still accepted as
#'   a deprecated alias for \code{"pure"}.
#' @param rare.label Character scalar used when \code{low.freq.policy = "pure"}.
#'   Default: \code{"RARE_DOMINANT"}.
#' @param tie.method Character. Tie handling passed to \code{linf.cells()} and used
#'   during absorb reassignment ("first", "random", "error").
#' @param return.diagnostics Logical. If TRUE, return reassignment diagnostics.
#' @param return.landmarks Logical. If TRUE, attach a depth-1 landmark summary
#'   computed by \code{\link{linf.landmarks}}.
#' @param landmark.types Character vector of landmark types passed to
#'   \code{\link{linf.landmarks}} when \code{return.landmarks = TRUE}.
#' @param landmark.view Character. Landmark view passed to
#'   \code{\link{linf.landmarks}} when \code{return.landmarks = TRUE}.
#' @param backend Character. Matrix backend to use: \code{"auto"},
#'   \code{"dense"}, or \code{"sparse"}. The default \code{"auto"} preserves
#'   sparse input and otherwise uses the dense path.
#'
#' @return List with:
#'   \itemize{
#'     \item \code{cell.index}, \code{cell.id}, \code{cell.label}: active labeling per \code{low.freq.policy}
#'     \item \code{cell.index.rare}, \code{cell.id.rare}, \code{cell.label.rare}
#'     \item \code{cell.index.absorb}, \code{cell.id.absorb}, \code{cell.label.absorb}
#'     \item \code{kept.cells.idx}, \code{kept.cells.id}, \code{kept.cells.lbl}
#'     \item \code{raw.index}, \code{raw.id}, \code{raw.label}
#'     \item \code{feature.ids}, \code{feature.labels}
#'     \item \code{size.table}, \code{size.table.id}
#'     \item \code{n0}, \code{low.freq.policy}, \code{rare.label}
#'     \item \code{diagnostics} (if \code{return.diagnostics = TRUE})
#'     \item \code{landmarks} (if \code{return.landmarks = TRUE})
#'   }
#'
#' @export
linf.csts <- function(S,
                      feature.ids = NULL,
                      feature.labels = NULL,
                      n0 = 50,
                      low.freq.policy = c("pure", "absorb"),
                      rare.label = "RARE_DOMINANT",
                      tie.method = c("first", "random", "error"),
                      return.diagnostics = FALSE,
                      return.landmarks = FALSE,
                      landmark.types = c("endpoint.max", "endpoint.min"),
                      landmark.view = c("active", "rare", "absorb"),
                      backend = c("auto", "dense", "sparse")) {

    low.freq.policy <- linf.normalize.low.freq.policy(low.freq.policy, "linf.csts")
    tie.method <- match.arg(tie.method)
    landmark.view <- match.arg(landmark.view)
    prep <- linf.prepare.matrix(S, backend = backend, fun.name = "linf.csts")
    X <- prep$X
    backend <- prep$backend

    if (!is.numeric(n0) || length(n0) != 1L || n0 < 1 || n0 %% 1 != 0) {
        stop("linf.csts: n0 must be integer >= 1")
    }
    if (!is.character(rare.label) || length(rare.label) != 1L || !nzchar(rare.label)) {
        stop("linf.csts: rare.label must be a non-empty character scalar")
    }

    linf.validate.matrix(X, backend = backend, fun.name = "linf.csts")

    meta <- resolve.linf.feature.meta(X, feature.ids = feature.ids, feature.labels = feature.labels)
    fid <- meta$feature.ids
    lev <- meta$feature.labels

    raw <- linf.cells(X,
                      feature.ids = fid,
                      feature.labels = lev,
                      tie.method = tie.method,
                      backend = backend)
    tab <- sort(table(raw$label[!is.na(raw$label)]), decreasing = TRUE)
    tab.id <- sort(table(raw$id[!is.na(raw$id)]), decreasing = TRUE)

    kept.lbl <- names(tab[tab >= n0])
    kept.idx <- match(kept.lbl, lev)
    kept.id <- fid[kept.idx]

    n <- nrow(X)

    ## Pure-policy labels: retain only states with support >= n0.
    is.kept <- !is.na(raw$label) & (raw$label %in% kept.lbl)

    cell.idx.rare <- raw$index
    cell.id.rare <- raw$id
    cell.lbl.rare <- raw$label
    cell.idx.rare[!is.kept] <- NA_integer_
    cell.id.rare[!is.kept] <- rare.label
    cell.lbl.rare[!is.kept] <- rare.label

    ## Absorb-policy labels: reassign low-support (and zero-row) samples
    ## into retained states.
    cell.idx.absorb <- raw$index
    cell.id.absorb <- raw$id
    cell.lbl.absorb <- raw$label

    reassigned <- logical(n)
    reassigned.from <- rep(NA_character_, n)
    reassigned.to   <- rep(NA_character_, n)

    if (length(kept.idx) > 0L) {

        ## Fallback target for degenerate cases (e.g., all kept values are zero)
        fallback.idx <- kept.idx[1L]
        fallback.lbl <- lev[fallback.idx]

        ## Absorb: (i) low-frequency raw labels, (ii) raw NA labels (e.g., all-zero rows)
        to.absorb <- which(!is.kept)

        if (length(to.absorb)) {
            if (backend == "sparse") {
                new.idx <- linf.absorb.sparse(
                    X,
                    to.absorb = to.absorb,
                    kept.idx = kept.idx,
                    tie.method = tie.method,
                    fallback.idx = fallback.idx
                )
            } else {
                new.idx <- apply(X[to.absorb, , drop = FALSE], 1, function(r) {
                    kvals <- r[kept.idx]
                    m <- max(kvals)

                    ## If there is no positive evidence among kept taxa, avoid arbitrary ties
                    if (!is.finite(m) || m <= 0) return(fallback.idx)

                    j <- kept.idx[kvals == m]
                    if (length(j) == 1L) return(j)
                    if (tie.method == "first") return(j[1L])
                    if (tie.method == "random") return(sample(j, 1L))
                    stop("linf.csts: tie during reassignment and tie.method = 'error'")
                })
            }

            reassigned[to.absorb] <- TRUE
            reassigned.from[to.absorb] <- raw$label[to.absorb]
            reassigned.to[to.absorb]   <- lev[new.idx]

            cell.idx.absorb[to.absorb] <- new.idx
            cell.id.absorb[to.absorb] <- fid[new.idx]
            cell.lbl.absorb[to.absorb] <- lev[new.idx]
        }

    } else {

        ## No retained states at this n0:
        ## - rare-policy: everyone is rare.label (already set above)
        ## - absorb-policy: undefined; keep NA labels
        cell.idx.absorb[] <- NA_integer_
        cell.id.absorb[] <- NA_character_
        cell.lbl.absorb[] <- NA_character_
    }

    ## Select active labeling
    if (low.freq.policy == "pure") {
        cell.idx <- cell.idx.rare
        cell.id <- cell.id.rare
        cell.lbl <- cell.lbl.rare
    } else {
        cell.idx <- cell.idx.absorb
        cell.id <- cell.id.absorb
        cell.lbl <- cell.lbl.absorb
    }

    names(cell.idx) <- rownames(X)
    names(cell.id) <- rownames(X)
    names(cell.lbl) <- rownames(X)

    names(cell.idx.rare) <- rownames(X)
    names(cell.id.rare) <- rownames(X)
    names(cell.lbl.rare) <- rownames(X)

    names(cell.idx.absorb) <- rownames(X)
    names(cell.id.absorb) <- rownames(X)
    names(cell.lbl.absorb) <- rownames(X)

    out <- list(
        cell.index       = cell.idx,
        cell.id          = cell.id,
        cell.label       = cell.lbl,
        cell.index.rare  = cell.idx.rare,
        cell.id.rare     = cell.id.rare,
        cell.label.rare  = cell.lbl.rare,
        cell.index.absorb = cell.idx.absorb,
        cell.id.absorb   = cell.id.absorb,
        cell.label.absorb = cell.lbl.absorb,
        kept.cells.idx   = kept.idx,
        kept.cells.id    = kept.id,
        kept.cells.lbl   = kept.lbl,
        raw.index        = raw$index,
        raw.id           = raw$id,
        raw.label        = raw$label,
        size.table       = tab,
        size.table.id    = tab.id,
        feature.ids      = fid,
        feature.labels   = lev,
        matrix.backend   = backend,
        n0               = as.integer(n0),
        low.freq.policy  = low.freq.policy,
        rare.label       = rare.label
    )

    if (return.diagnostics) {
        out$diagnostics <- list(
            reassigned      = reassigned,
            reassigned.from = reassigned.from,
            reassigned.to   = reassigned.to
        )
    }

    out$cst.id.levels <- list(level1 = out$cell.id)
    out$cst.id.levels.rare <- list(level1 = out$cell.id.rare)
    out$cst.id.levels.absorb <- list(level1 = out$cell.id.absorb)
    out$cst.levels <- list(level1 = out$cell.label)
    out$cst.levels.rare <- list(level1 = out$cell.label.rare)
    out$cst.levels.absorb <- list(level1 = out$cell.label.absorb)
    out$cst.depth <- 1L

    class(out) <- "linf.csts"

    if (isTRUE(return.landmarks)) {
        out$landmarks <- linf.landmarks(
            X,
            out,
            depth = 1L,
            view = landmark.view,
            landmark.types = landmark.types,
            tie.method = tie.method,
            backend = backend
        )
    }

    out
}

#' From counts to filtered, normalized, truncated dCSTs
#'
#' Convenience pipeline: runs \code{filter.asv()} on counts, computes
#' L-infinity-normalized relatives on the filtered counts, and assigns
#' truncated dCSTs.
#'
#' @param S.counts Numeric matrix of counts (samples x features).
#' @param ... Arguments passed to \code{filter.asv()} (e.g., \code{min.lib},
#'   \code{prev.prop}, \code{min.count}, \code{min.rel}).
#' @param backend Character. Matrix backend to use: \code{"auto"},
#'   \code{"dense"}, or \code{"sparse"}.
#' @return A list with:
#' \itemize{
#'   \item \code{filter}: the full result from \code{filter.asv()}.
#'   \item \code{linf.rel}: L-infinity-normalized matrix on filtered counts.
#'   \item \code{csts}: result from \code{linf.csts()} on \code{linf.rel}.
#' }
#'
#' @examples
#' set.seed(1)
#' S <- matrix(rpois(100, lambda = 5), nrow = 20, ncol = 5)
#' res <- asv.to.linf.csts(S, min.lib = 10, prev.prop = 0.1, min.count = 1)
#' names(res)
#' apply(res$linf.rel, 1, max)  # should be 1 (or 0 for all-zero rows)
#'
#' @export
asv.to.linf.csts <- function(S.counts,
                             ...,
                             backend = c("auto", "dense", "sparse")) {
  filt <- filter.asv(S.counts, ...)
  M <- normalize.linf(filt$counts, backend = backend)
  cst <- linf.csts(M, backend = backend)
  list(filter = filt, linf.rel = M, csts = cst)
}

validate.linf.csts <- function(obj) {

  stopifnot(is.list(obj))
  stopifnot(!is.null(obj$cell.label))

  n <- length(obj$cell.label)

  ## Case 1: explicit hierarchy (refined dCSTs)
  if (!is.null(obj$cst.levels)) {

    stopifnot(is.list(obj$cst.levels))
    stopifnot(!is.null(obj$cst.depth))
    stopifnot(obj$cst.depth == length(obj$cst.levels))

    for (lvl in obj$cst.levels) {
      stopifnot(length(lvl) == n)
    }

    return(invisible(TRUE))
  }

  ## Case 2: depth-1 dCST (plain linf.csts output)
  ## This is VALID and should be accepted
  return(invisible(TRUE))
}

#' Refine a dCST hierarchy by one level
#'
#' @description
#' Selects well-supported leaf dominance-lineages and refines them by dropping
#' the dominant feature(s) encoded in the parent label path and re-applying
#' \code{\link{linf.csts}} to the remaining features. The resulting child
#' labels are appended to the parent label using \code{sep}.
#'
#' Low-support child lineages are handled by \code{low.freq.policy}. When
#' \code{low.freq.policy = "pure"}, rare buckets at depth >= 2 become
#' parent-prefixed automatically via the hierarchical \code{paste(parent, child, sep = sep)}.
#'
#' @param M Numeric matrix (samples x features) used for refinement. Column names
#'   should match feature labels used in dCST names.
#' @param csts A \code{"linf.csts"} object.
#' @param n0 Integer >= 1. Minimum support required to retain a child lineage
#'   (passed to \code{linf.csts}).
#' @param refinement.factor Numeric > 0. Auto-refine parent lineages with
#'   support >= \code{refinement.factor * n0}.
#' @param sep Character scalar used to concatenate hierarchical labels.
#' @param low.freq.policy Character. One of \code{"pure"} or \code{"absorb"}.
#'   Default: \code{"pure"}. The legacy value \code{"rare"} is still accepted as
#'   a deprecated alias for \code{"pure"}.
#' @param rare.label Character scalar for rare buckets when \code{low.freq.policy = "pure"}.
#' @param verbose Logical. If TRUE, print progress information.
#' @param backend Character. Matrix backend to use: \code{"auto"},
#'   \code{"dense"}, or \code{"sparse"}. The default \code{"auto"} inherits the
#'   backend from \code{M} or from \code{csts} when available.
#'
#' @return Updated \code{"linf.csts"} object with \code{cst.depth} increased by one and
#'   updated \code{cell.label}. Policy-specific views are stored in
#'   \code{cell.label.rare} and \code{cell.label.absorb}.
#'
#' @export
refine.linf.csts <- function(M,
                             csts,
                             n0 = 50,
                             refinement.factor = 2,
                             sep = "__",
                             low.freq.policy = c("pure", "absorb"),
                             rare.label = "RARE_DOMINANT",
                             verbose = TRUE,
                             backend = c("auto", "dense", "sparse")) {

    validate.linf.csts(csts)

    low.freq.policy <- linf.normalize.low.freq.policy(low.freq.policy, "refine.linf.csts")
    if (missing(backend) && !is.null(csts$matrix.backend)) {
        backend <- csts$matrix.backend
    }
    prep <- linf.prepare.matrix(M, backend = backend, fun.name = "refine.linf.csts")
    M <- prep$X
    backend <- prep$backend
    linf.validate.matrix(M, backend = backend, fun.name = "refine.linf.csts")

    if (!is.numeric(n0) || length(n0) != 1L || n0 < 1 || n0 %% 1 != 0) {
        stop("refine.linf.csts: n0 must be integer >= 1")
    }
    if (!is.numeric(refinement.factor) || length(refinement.factor) != 1L ||
        !is.finite(refinement.factor) || refinement.factor <= 0) {
        stop("refine.linf.csts: refinement.factor must be a finite numeric > 0")
    }
    if (!is.character(sep) || length(sep) != 1L || !nzchar(sep)) {
        stop("refine.linf.csts: sep must be a non-empty character scalar")
    }
    if (!is.character(rare.label) || length(rare.label) != 1L || !nzchar(rare.label)) {
        stop("refine.linf.csts: rare.label must be a non-empty character scalar")
    }

    ## Initialize hierarchy if missing
    if (is.null(csts$cst.levels)) {
        csts$cst.levels <- list(level1 = csts$cell.label)
        csts$cst.depth <- 1L
        csts$sep <- sep

        ## Ensure both policy-specific views exist at depth 1
        if (is.null(csts$cell.label.rare)) {
            csts$cell.label.rare <- csts$cell.label
        }
        if (is.null(csts$cell.label.absorb)) {
            csts$cell.label.absorb <- csts$cell.label
        }

        csts$cst.levels.rare <- list(level1 = csts$cell.label.rare)
        csts$cst.levels.absorb <- list(level1 = csts$cell.label.absorb)
    }

    if (is.null(csts$cst.id.levels)) {
        csts$cst.id.levels <- list(level1 = csts$cell.id %||% csts$cell.label)
    }

    if (is.null(csts$feature.ids)) {
        csts$feature.ids <- colnames(M)
    }
    if (is.null(csts$feature.labels)) {
        csts$feature.labels <- csts$feature.ids
    }

    ## Backward-compatible defaults if stored views are missing
    if (is.null(csts$cst.levels.rare)) csts$cst.levels.rare <- csts$cst.levels
    if (is.null(csts$cst.levels.absorb)) csts$cst.levels.absorb <- csts$cst.levels
    if (is.null(csts$cst.id.levels.rare)) csts$cst.id.levels.rare <- csts$cst.id.levels
    if (is.null(csts$cst.id.levels.absorb)) csts$cst.id.levels.absorb <- csts$cst.id.levels

    depth <- csts$cst.depth + 1L
    parent.ids <- csts$cst.id.levels[[depth - 1L]]
    parent.ids.rare <- csts$cst.id.levels.rare[[depth - 1L]]
    parent.ids.absorb <- csts$cst.id.levels.absorb[[depth - 1L]]
    parent.labels <- csts$cst.levels[[depth - 1L]]

    parent.labels.rare <- csts$cst.levels.rare[[depth - 1L]]
    parent.labels.absorb <- csts$cst.levels.absorb[[depth - 1L]]

    refined.ids.rare <- parent.ids.rare
    refined.ids.absorb <- parent.ids.absorb
    refined.labels.rare <- parent.labels.rare
    refined.labels.absorb <- parent.labels.absorb

    cell.sizes <- sort(table(parent.ids), decreasing = TRUE)

    threshold <- refinement.factor * n0
    refine.cells <- names(cell.sizes[cell.sizes >= threshold])

    if (verbose) {
        cat("========================================\n")
        cat("AUTO-REFINEMENT MODE\n")
        cat("========================================\n")
        cat("Refinement threshold:", threshold, "\n")
        cat("Dominance-lineages selected for refinement:", length(refine.cells), "\n\n")
    }

    for (cell in refine.cells) {
        ## Rare buckets are synthetic catch-all groups, not real feature paths.
        ## They should never be refined further.
        if (identical(cell, rare.label)) next

        idx <- which(parent.ids == cell)
        parent.id.rare <- parent.ids.rare[idx[1L]]
        parent.id.absorb <- parent.ids.absorb[idx[1L]]
        parent.label.rare <- parent.labels.rare[idx[1L]]
        parent.label.absorb <- parent.labels.absorb[idx[1L]]

        parent.features <- strsplit(cell, sep, fixed = TRUE)[[1]]
        drop.idx <- match(parent.features, csts$feature.ids)
        drop.idx <- drop.idx[!is.na(drop.idx)]

        ## If no parent features match columns (e.g., the lineage is a rare
        ## bucket), do not drop any columns.
        ## Note: x[, -integer(0)] selects *zero* columns, so we must handle this explicitly.
        if (length(drop.idx) == 0L) {
            M.sub <- M[idx, , drop = FALSE]
        } else if (length(drop.idx) >= ncol(M)) {
            ## Dropping all columns would yield an empty matrix; nothing to refine.
            next
        } else {
            M.sub <- M[idx, -drop.idx, drop = FALSE]
        }

        if (nrow(M.sub) == 0L || ncol(M.sub) == 0L) next

        sub.csts <- linf.csts(M.sub,
                             feature.ids = csts$feature.ids[-drop.idx],
                             feature.labels = csts$feature.labels[-drop.idx],
                             n0 = n0,
                             low.freq.policy = low.freq.policy,
                             rare.label = rare.label,
                             backend = backend)

        sub.ids.rare <- sub.csts$cell.id.rare %||% sub.csts$cell.label.rare
        sub.ids.absorb <- sub.csts$cell.id.absorb %||% sub.csts$cell.label.absorb
        sub.labels.rare <- sub.csts$cell.label.rare
        sub.labels.absorb <- sub.csts$cell.label.absorb

        ## If refinement yields no retained child lineages, skip.
        if (all(sub.labels.rare == rare.label)) next

        refined.ids.rare[idx] <- paste(parent.id.rare, sub.ids.rare, sep = sep)
        refined.ids.absorb[idx] <- paste(parent.id.absorb, sub.ids.absorb, sep = sep)
        refined.labels.rare[idx] <- paste(parent.label.rare, sub.labels.rare, sep = sep)
        refined.labels.absorb[idx] <- paste(parent.label.absorb, sub.labels.absorb, sep = sep)
    }

    ## Update dCST object (store both views; active view chosen by low.freq.policy)
    csts$cst.id.levels[[depth]] <- if (low.freq.policy == "pure") refined.ids.rare else refined.ids.absorb
    csts$cst.id.levels.rare[[depth]] <- refined.ids.rare
    csts$cst.id.levels.absorb[[depth]] <- refined.ids.absorb
    csts$cst.levels.rare[[depth]] <- refined.labels.rare
    csts$cst.levels.absorb[[depth]] <- refined.labels.absorb

    if (low.freq.policy == "pure") {
        refined.labels <- refined.labels.rare
    } else {
        refined.labels <- refined.labels.absorb
    }

    csts$cst.levels[[depth]] <- refined.labels
    csts$cst.depth <- depth
    csts$sep <- sep

    csts$cell.id.rare <- refined.ids.rare
    csts$cell.id.absorb <- refined.ids.absorb
    csts$cell.id <- csts$cst.id.levels[[depth]]
    csts$cell.label.rare <- refined.labels.rare
    csts$cell.label.absorb <- refined.labels.absorb
    csts$cell.label <- refined.labels

    csts$low.freq.policy <- low.freq.policy
    csts$rare.label <- rare.label
    csts$matrix.backend <- backend
    csts$landmarks <- NULL

    class(csts) <- "linf.csts"

    csts
}

#' Iteratively refine dCSTs by one additional level
#'
#' @description
#' Appends one refinement level to an existing dCST hierarchy produced by
#' \code{\link{refine.linf.csts}} or \code{\link{refine.linf.csts.iter}}.
#' Dominance-lineages to refine can be provided explicitly or selected
#' automatically based on \code{refinement.factor * n0}.
#'
#' @param M Numeric matrix used for refinement.
#' @param refined A \code{"linf.csts"} object with an existing hierarchy.
#' @param cells.to.refine Character vector of leaf dominance-lineage IDs to
#'   refine, or NULL for auto-selection. The argument name is retained for
#'   backward compatibility.
#' @param n0 Integer >= 1. Minimum support required to retain a child lineage.
#' @param refinement.factor Numeric > 0. Auto-selection threshold multiplier.
#' @param sep Character scalar used to concatenate hierarchical labels.
#' @param low.freq.policy Character. One of \code{"pure"} or \code{"absorb"}.
#'   Default: \code{"pure"}. The legacy value \code{"rare"} is still accepted as
#'   a deprecated alias for \code{"pure"}.
#' @param rare.label Character scalar for rare buckets when \code{low.freq.policy = "pure"}.
#' @param verbose Logical. If TRUE, print progress information.
#' @param backend Character. Matrix backend to use: \code{"auto"},
#'   \code{"dense"}, or \code{"sparse"}. The default \code{"auto"} inherits the
#'   backend from \code{M} or from \code{refined} when available.
#'
#' @return Updated \code{"linf.csts"} object with \code{cst.depth} increased by one.
#'
#' @export
refine.linf.csts.iter <- function(M,
                                  refined,
                                  cells.to.refine = NULL,
                                  n0 = 50,
                                  refinement.factor = 5,
                                  sep = "__",
                                  low.freq.policy = c("pure", "absorb"),
                                  rare.label = "RARE_DOMINANT",
                                  verbose = TRUE,
                                  backend = c("auto", "dense", "sparse")) {

    validate.linf.csts(refined)

    low.freq.policy <- linf.normalize.low.freq.policy(low.freq.policy, "refine.linf.csts.iter")
    if (missing(backend) && !is.null(refined$matrix.backend)) {
        backend <- refined$matrix.backend
    }
    prep <- linf.prepare.matrix(M, backend = backend, fun.name = "refine.linf.csts.iter")
    M <- prep$X
    backend <- prep$backend
    linf.validate.matrix(M, backend = backend, fun.name = "refine.linf.csts.iter")

    if (!is.numeric(n0) || length(n0) != 1L || n0 < 1 || n0 %% 1 != 0) {
        stop("refine.linf.csts.iter: n0 must be integer >= 1")
    }
    if (!is.numeric(refinement.factor) || length(refinement.factor) != 1L ||
        !is.finite(refinement.factor) || refinement.factor <= 0) {
        stop("refine.linf.csts.iter: refinement.factor must be a finite numeric > 0")
    }
    if (!is.character(sep) || length(sep) != 1L || !nzchar(sep)) {
        stop("refine.linf.csts.iter: sep must be a non-empty character scalar")
    }
    if (!is.character(rare.label) || length(rare.label) != 1L || !nzchar(rare.label)) {
        stop("refine.linf.csts.iter: rare.label must be a non-empty character scalar")
    }

    ## Ensure stored views exist (backward compatible)
    if (is.null(refined$cst.levels.rare)) refined$cst.levels.rare <- refined$cst.levels
    if (is.null(refined$cst.levels.absorb)) refined$cst.levels.absorb <- refined$cst.levels
    if (is.null(refined$cst.id.levels)) refined$cst.id.levels <- list(level1 = refined$cell.id %||% refined$cell.label)
    if (is.null(refined$cst.id.levels.rare)) refined$cst.id.levels.rare <- refined$cst.id.levels
    if (is.null(refined$cst.id.levels.absorb)) refined$cst.id.levels.absorb <- refined$cst.id.levels
    if (is.null(refined$feature.ids)) refined$feature.ids <- colnames(M)
    if (is.null(refined$feature.labels)) refined$feature.labels <- refined$feature.ids

    depth <- refined$cst.depth

    parent.ids <- refined$cst.id.levels[[depth]]
    parent.ids.rare <- refined$cst.id.levels.rare[[depth]]
    parent.ids.absorb <- refined$cst.id.levels.absorb[[depth]]
    parent.labels <- refined$cst.levels[[depth]]
    parent.labels.rare <- refined$cst.levels.rare[[depth]]
    parent.labels.absorb <- refined$cst.levels.absorb[[depth]]

    cell.sizes <- sort(table(parent.ids), decreasing = TRUE)
    threshold <- refinement.factor * n0

    if (is.null(cells.to.refine)) {
        cells.to.refine <- names(cell.sizes[cell.sizes >= threshold])
    }

    if (verbose) {
        cat("Auto-selecting dominance-lineages for iterative refinement (threshold =", threshold, "):\n")
        for (c in cells.to.refine) {
            cat(" -", c, ":", cell.sizes[c], "samples\n")
        }
        cat("\n")
    }

    new.ids.rare <- parent.ids.rare
    new.ids.absorb <- parent.ids.absorb
    new.labels.rare <- parent.labels.rare
    new.labels.absorb <- parent.labels.absorb

    for (cell in cells.to.refine) {
        ## Rare buckets are synthetic catch-all groups, not real feature paths.
        ## They should never be refined further.
        if (identical(cell, rare.label)) next

        idx <- which(parent.ids == cell)
        parent.id.rare <- parent.ids.rare[idx[1L]]
        parent.id.absorb <- parent.ids.absorb[idx[1L]]
        parent.label.rare <- parent.labels.rare[idx[1L]]
        parent.label.absorb <- parent.labels.absorb[idx[1L]]

        parents <- strsplit(cell, sep, fixed = TRUE)[[1]]
        drop.idx <- match(parents, refined$feature.ids)
        drop.idx <- drop.idx[!is.na(drop.idx)]

        ## If no parent features match columns (e.g., the lineage is a rare
        ## bucket), do not drop any columns.
        ## Note: x[, -integer(0)] selects *zero* columns, so we must handle this explicitly.
        if (length(drop.idx) == 0L) {
            M.sub <- M[idx, , drop = FALSE]
        } else if (length(drop.idx) >= ncol(M)) {
            ## Dropping all columns would yield an empty matrix; nothing to refine.
            next
        } else {
            M.sub <- M[idx, -drop.idx, drop = FALSE]
        }

        if (nrow(M.sub) == 0L || ncol(M.sub) == 0L) next

        sub.csts <- linf.csts(M.sub,
                             feature.ids = refined$feature.ids[-drop.idx],
                             feature.labels = refined$feature.labels[-drop.idx],
                             n0 = n0,
                             low.freq.policy = low.freq.policy,
                             rare.label = rare.label,
                             backend = backend)

        sub.ids.rare <- sub.csts$cell.id.rare %||% sub.csts$cell.label.rare
        sub.ids.absorb <- sub.csts$cell.id.absorb %||% sub.csts$cell.label.absorb
        sub.labels.rare <- sub.csts$cell.label.rare
        sub.labels.absorb <- sub.csts$cell.label.absorb

        ## If refinement yields no retained child lineages, skip.
        if (all(sub.labels.rare == rare.label)) next

        new.ids.rare[idx] <- paste(parent.id.rare, sub.ids.rare, sep = sep)
        new.ids.absorb[idx] <- paste(parent.id.absorb, sub.ids.absorb, sep = sep)
        new.labels.rare[idx] <- paste(parent.label.rare, sub.labels.rare, sep = sep)
        new.labels.absorb[idx] <- paste(parent.label.absorb, sub.labels.absorb, sep = sep)
    }

    refined$cst.id.levels[[depth + 1L]] <- if (low.freq.policy == "pure") new.ids.rare else new.ids.absorb
    refined$cst.id.levels.rare[[depth + 1L]] <- new.ids.rare
    refined$cst.id.levels.absorb[[depth + 1L]] <- new.ids.absorb
    refined$cst.levels.rare[[depth + 1L]] <- new.labels.rare
    refined$cst.levels.absorb[[depth + 1L]] <- new.labels.absorb

    if (low.freq.policy == "pure") {
        new.labels <- new.labels.rare
    } else {
        new.labels <- new.labels.absorb
    }

    refined$cst.levels[[depth + 1L]] <- new.labels
    refined$cst.depth <- depth + 1L

    refined$cell.id.rare <- new.ids.rare
    refined$cell.id.absorb <- new.ids.absorb
    refined$cell.id <- refined$cst.id.levels[[depth + 1L]]
    refined$cell.label.rare <- new.labels.rare
    refined$cell.label.absorb <- new.labels.absorb
    refined$cell.label <- new.labels

    refined$low.freq.policy <- low.freq.policy
    refined$rare.label <- rare.label
    refined$matrix.backend <- backend
    refined$sep <- sep
    refined$landmarks <- NULL

    class(refined) <- "linf.csts"

    refined
}

#' Switch a dCST hierarchy to the "absorb" view (collapse rare buckets)
#'
#' @description
#' Returns a copy of a \code{"linf.csts"} object with \code{cell.label} (and, if
#' present, \code{cst.levels}) replaced by the precomputed "absorb" labeling.
#' This does not recompute dCSTs; it only switches between labelings already
#' stored in the object.
#'
#' @param csts A \code{"linf.csts"} object produced by \code{\link{linf.csts}}
#'   and optionally refined by \code{\link{refine.linf.csts}} /
#'   \code{\link{refine.linf.csts.iter}}.
#'
#' @return A \code{"linf.csts"} object using the "absorb" view.
#'
#' @export
collapse.rare <- function(csts) {

  validate.linf.csts(csts)

  if (is.null(csts$cell.label.absorb)) {
    stop("collapse.rare: no precomputed absorb labels found in object")
  }

  if (!is.null(csts$cell.index.absorb)) {
    csts$cell.index <- csts$cell.index.absorb
  }
  if (!is.null(csts$cell.id.absorb)) {
    csts$cell.id <- csts$cell.id.absorb
  }
  csts$cell.label <- csts$cell.label.absorb
  csts$low.freq.policy <- "absorb"

  if (!is.null(csts$cst.id.levels)) {
    if (is.null(csts$cst.id.levels.absorb)) {
      depth <- csts$cst.depth
      csts$cst.id.levels[[depth]] <- csts$cell.id %||% csts$cell.label
    } else {
      csts$cst.id.levels <- csts$cst.id.levels.absorb
    }
  }

  if (!is.null(csts$cst.levels)) {
    if (is.null(csts$cst.levels.absorb)) {
      ## Fall back to updating only the leaf level
      depth <- csts$cst.depth
      csts$cst.levels[[depth]] <- csts$cell.label.absorb
    } else {
      csts$cst.levels <- csts$cst.levels.absorb
    }
  }

  csts$landmarks <- NULL
  class(csts) <- "linf.csts"
  csts
}

#' Switch a dCST hierarchy to the "rare" view (explicit rare buckets)
#'
#' @description
#' Returns a copy of a \code{"linf.csts"} object with \code{cell.label} (and, if
#' present, \code{cst.levels}) replaced by the precomputed "rare" labeling.
#' This does not recompute dCSTs; it only switches between labelings already
#' stored in the object. This explicit rare-bucket view is the active labeling
#' used by \code{low.freq.policy = "pure"}.
#'
#' @param csts A \code{"linf.csts"} object produced by \code{\link{linf.csts}}
#'   and optionally refined by \code{\link{refine.linf.csts}} /
#'   \code{\link{refine.linf.csts.iter}}.
#'
#' @return A \code{"linf.csts"} object using the "rare" view.
#'
#' @export
expand.rare <- function(csts) {

  validate.linf.csts(csts)

  if (is.null(csts$cell.label.rare)) {
    stop("expand.rare: no precomputed rare labels found in object")
  }

  if (!is.null(csts$cell.index.rare)) {
    csts$cell.index <- csts$cell.index.rare
  }
  if (!is.null(csts$cell.id.rare)) {
    csts$cell.id <- csts$cell.id.rare
  }
  csts$cell.label <- csts$cell.label.rare
  csts$low.freq.policy <- "pure"

  if (!is.null(csts$cst.id.levels)) {
    if (is.null(csts$cst.id.levels.rare)) {
      depth <- csts$cst.depth
      csts$cst.id.levels[[depth]] <- csts$cell.id %||% csts$cell.label
    } else {
      csts$cst.id.levels <- csts$cst.id.levels.rare
    }
  }

  if (!is.null(csts$cst.levels)) {
    if (is.null(csts$cst.levels.rare)) {
      ## Fall back to updating only the leaf level
      depth <- csts$cst.depth
      csts$cst.levels[[depth]] <- csts$cell.label.rare
    } else {
      csts$cst.levels <- csts$cst.levels.rare
    }
  }

  csts$landmarks <- NULL
  class(csts) <- "linf.csts"
  csts
}

#' Print method for \code{"linf.csts"}
#'
#' @param x A \code{"linf.csts"} object.
#' @param ... Unused.
#' @return The input object, invisibly.
#' @export
print.linf.csts <- function(x, ...) {

  validate.linf.csts(x)

  ## Determine hierarchy
  if (is.null(x$cst.levels)) {
    levels <- list(level1 = x$cell.label)
    max.depth <- 1L
  } else {
    levels <- x$cst.levels
    max.depth <- x$cst.depth
  }

  cat("\n================================================================================\n")
  cat("Dominant Community State Type Hierarchy\n")
  cat("================================================================================\n")
  cat("Total samples: ", length(x$cell.label), "\n")
  cat("Max depth:     ", max.depth, "\n")

  ## Policy stamp (if available)
  if (!is.null(x$low.freq.policy) && !is.null(x$rare.label)) {
    cat("Low-freq policy: ", x$low.freq.policy,
        " (rare.label: ", x$rare.label, ")\n", sep = "")
  }

  cat("--------------------------------------------------------------------------------\n\n")

  for (d in seq_along(levels)) {
    cat("Depth", d, "\n")
    tab <- sort(table(levels[[d]]), decreasing = TRUE)
    for (nm in names(tab)) {
      cat("  ", nm, ": ", tab[[nm]], "\n", sep = "")
    }
    cat("\n")
  }

  invisible(x)
}

#' Summarize dCST hierarchy statistics
#'
#' @param object A \code{"linf.csts"} object.
#' @param ... Unused.
#' @return Data frame with depth-wise statistics
#' @export
summary.linf.csts <- function(object, ...) {

  validate.linf.csts(object)

  ## Determine hierarchy
  if (is.null(object$cst.levels)) {
    levels <- list(level1 = object$cell.label)
  } else {
    levels <- object$cst.levels
  }

  out <- data.frame(
    depth = integer(0),
    n.cells = integer(0),
    total.samples = integer(0),
    mean.size = numeric(0),
    median.size = numeric(0),
    min.size = numeric(0),
    max.size = numeric(0)
  )

  for (d in seq_along(levels)) {
    lbl <- levels[[d]]
    tab <- table(lbl)

    out <- rbind(out, data.frame(
      depth = d,
      n.cells = length(tab),
      total.samples = sum(tab),
      mean.size = mean(tab),
      median.size = stats::median(tab),
      min.size = min(tab),
      max.size = max(tab)
    ))
  }

  out
}

escape.latex <- function(x) {
  x <- gsub("\\\\", "\\\\textbackslash{}", x, fixed = TRUE)
  x <- gsub("([#$%&_{}])", "\\\\\\1", x, perl = TRUE)
  x <- gsub("~", "\\\\textasciitilde{}", x, fixed = TRUE)
  x <- gsub("\\^", "\\\\textasciicircum{}", x, perl = TRUE)
  x
}

df.to.latex <- function(df,
                        caption = NULL,
                        label = NULL,
                        digits = 1,
                        include.rownames = FALSE,
                        booktabs = TRUE,
                        use.float = TRUE,
                        print = FALSE) {
  if (!is.data.frame(df)) stop("df.to.latex: df must be a data.frame")

  if (!include.rownames) {
    row_spec <- "l"
  } else {
    row_spec <- "l"
    df <- cbind(Row = rownames(df), df, stringsAsFactors = FALSE)
  }

  col_spec <- paste0(row_spec, paste(rep("r", ncol(df) - 1L), collapse = ""))
  if (ncol(df) == 1L) col_spec <- "l"

  tab_top <- if (booktabs) "\\toprule" else "\\hline"
  tab_mid <- if (booktabs) "\\midrule" else "\\hline"
  tab_bot <- if (booktabs) "\\bottomrule" else "\\hline"

  lines <- character(0)
  if (use.float) lines <- c(lines, "\\begin{table}[htbp]", "\\centering")
  if (!is.null(caption)) lines <- c(lines, paste0("\\caption{", escape.latex(caption), "}"))
  if (!is.null(label)) lines <- c(lines, paste0("\\label{", escape.latex(label), "}"))

  lines <- c(lines, paste0("\\begin{tabular}{", col_spec, "}"), tab_top)

  header <- paste(escape.latex(colnames(df)), collapse = " & ")
  lines <- c(lines, paste0(header, " \\\\"), tab_mid)

  for (i in seq_len(nrow(df))) {
    vals <- vapply(df[i, , drop = FALSE], function(v) {
      if (is.numeric(v)) {
        format(round(v, digits), trim = TRUE, scientific = FALSE)
      } else {
        as.character(v)
      }
    }, character(1))
    vals <- escape.latex(vals)
    lines <- c(lines, paste0(paste(vals, collapse = " & "), " \\\\"))
  }

  lines <- c(lines, tab_bot, "\\end{tabular}")
  if (use.float) lines <- c(lines, "\\end{table}")

  if (isTRUE(print)) cat(paste(lines, collapse = "\n"), "\n")
  lines
}

#' Render dCSTs as a LaTeX table
#'
#' @description
#' Generates LaTeX code summarizing dCSTs at a specified hierarchy depth.
#' Uses the explicit dCST hierarchy stored in \code{cst.levels} when available.
#'
#' @param csts A dCST object produced by \code{linf.csts()} and optionally refined
#'   by \code{refine.linf.csts()} or \code{refine.linf.csts.iter()}.
#' @param depth Integer. dCST depth to render. Default is the leaf level
#'   (\code{csts$cst.depth}). If \code{cst.levels} is missing, depth is ignored.
#' @param caption Character or NULL. LaTeX table caption.
#' @param label Character or NULL. LaTeX label for referencing the table.
#' @param digits Integer. Digits to display for percentages. Default: 1.
#' @param include.percent Logical. If TRUE, include percent of total samples.
#'
#' @return Character vector containing LaTeX table code.
#'
#' @details
#' This function is hierarchy-aware. It does not infer dCST depth from label
#' strings. If the dCST object does not contain \code{cst.levels}, the function
#' falls back to using \code{cell.label} as a single-level dCST.
#'
#' @export
latex.linf.csts <- function(csts,
                            depth = NULL,
                            caption = NULL,
                            label = NULL,
                            digits = 1,
                            include.percent = TRUE) {

  validate.linf.csts(csts)

  ## Determine labels at requested depth
  if (!is.null(csts$cst.levels)) {
    if (is.null(depth)) {
      depth <- csts$cst.depth
    }
    if (depth < 1 || depth > csts$cst.depth) {
      stop("latex.linf.csts: invalid depth")
    }
    labels <- csts$cst.levels[[depth]]
  } else {
    ## Backward compatibility: flat dCST
    labels <- csts$cell.label
    depth <- 1L
  }

  tab <- sort(table(labels), decreasing = TRUE)
  total <- sum(tab)

  df <- data.frame(
    CST = names(tab),
    Samples = as.integer(tab),
    stringsAsFactors = FALSE
  )

  if (include.percent) {
    df$Percent <- round(100 * df$Samples / total, digits)
  }

  ## Build LaTeX using df.to.latex()
  latex <- df.to.latex(
    df,
    caption = caption,
    label = label,
    digits = digits,
    include.rownames = FALSE,
    booktabs = TRUE,
    use.float = TRUE,
    print = FALSE
  )

  invisible(latex)
}

test_that("transfer.dcsts reproduces fitted labels on aligned training data", {
  X <- rbind(
    s1 = c(A = 10, B = 5, C = 1, D = 0),
    s2 = c(A = 9, B = 4, C = 2, D = 0),
    s3 = c(A = 8, B = 1, C = 6, D = 0),
    s4 = c(A = 7, B = 1, C = 5, D = 0),
    s5 = c(C = 10, A = 5, B = 1, D = 0),
    s6 = c(C = 9, A = 4, B = 1, D = 0),
    s7 = c(C = 8, A = 1, B = 6, D = 0),
    s8 = c(C = 7, A = 1, B = 5, D = 0)
  )
  X <- X[, c("A", "B", "C", "D")]
  M <- normalize.linf(X)
  fit <- linf.csts(M, n0 = 2, low.freq.policy = "absorb")
  fit <- refine.linf.csts(M, fit, n0 = 2, low.freq.policy = "absorb", verbose = FALSE)

  tr <- transfer.dcsts(X, fit, match.by = "feature.labels")

  expect_equal(tr$assignment[, "depth1"], fit$cst.levels.absorb[[1]])
  expect_equal(tr$assignment[, "depth2"], fit$cst.levels.absorb[[2]])
})

test_that("transfer.dcsts aligns by feature metadata rather than column order", {
  X <- rbind(
    s1 = c(A = 10, B = 1, C = 0),
    s2 = c(A = 9, B = 2, C = 0),
    s3 = c(A = 1, B = 10, C = 0),
    s4 = c(A = 2, B = 9, C = 0)
  )
  fit <- linf.csts(normalize.linf(X), n0 = 2, low.freq.policy = "absorb")
  shuffled <- X[, c("C", "B", "A")]

  tr <- transfer.dcsts(shuffled, fit, match.by = "feature.labels")

  expect_equal(tr$assignment[, "depth1"], fit$cst.levels.absorb[[1]])
})

test_that("transfer.dcsts can use support tie-breaking in frozen transfer", {
  X <- rbind(
    a1 = c(A = 10, B = 1),
    a2 = c(A = 9, B = 1),
    a3 = c(A = 8, B = 1),
    b1 = c(A = 1, B = 10)
  )
  fit <- linf.csts(normalize.linf(X), n0 = 1, low.freq.policy = "absorb")
  tied <- rbind(new = c(A = 5, B = 5))

  tr <- transfer.dcsts(tied, fit, match.by = "feature.labels", tie.method = "support")

  expect_equal(unname(tr$assignment[1, "depth1"]), "A")
})

test_that("linf.feature.labels applies abbreviations and global indices", {
  ids <- c("asv_1", "asv_4", "asv_5", "asv_6")
  tax <- c(
    "Lactobacillus iners",
    "Lactobacillus iners",
    "Megasphaera lornae",
    "Ca_Lachnocurva_vaginae"
  )
  abbr <- c(
    Lactobacillus = "L.",
    Streptococcus = "Strep",
    Gardnerella = "Gard.",
    Fannyhessea = "F.",
    Megasphaera = "Mega.",
    Ureaplasma = "U."
  )
  aliases <- c(
    "Ca. Lachnocurva vaginae" = "BVAB1",
    "Ca_Lachnocurva_vaginae" = "BVAB1"
  )

  out <- linf.feature.labels(
    feature.ids = ids,
    taxonomy = tax,
    abbreviations = abbr,
    aliases = aliases,
    duplicate.index = "global"
  )

  expect_equal(out, c("L. iners 1", "L. iners 4", "Mega. lornae", "BVAB1"))
})

test_that("linf.cells keeps ids separate from labels", {
  S <- rbind(c(0.7, 0.3), c(0, 0), c(0.1, 0.9))
  ids <- c("asv_1", "asv_4")
  labels <- c("L. iners 1", "L. iners 4")

  out <- linf.cells(S, feature.ids = ids, feature.labels = labels)

  expect_equal(out$index, c(1L, NA_integer_, 2L))
  expect_equal(out$id, c("asv_1", NA_character_, "asv_4"))
  expect_equal(out$label, c("L. iners 1", NA_character_, "L. iners 4"))
  expect_equal(out$id.levels, ids)
  expect_equal(out$levels, labels)
})

test_that("refine.linf.csts tracks id paths separately from label paths", {
  M <- rbind(
    c(0.90, 0.08, 0.02),
    c(0.88, 0.09, 0.03),
    c(0.87, 0.10, 0.03),
    c(0.86, 0.11, 0.03),
    c(0.05, 0.92, 0.03),
    c(0.06, 0.91, 0.03)
  )
  ids <- c("asv_1", "asv_4", "asv_5")
  labels <- c("L. iners 1", "L. iners 4", "Mega. lornae")

  d1 <- linf.csts(
    M,
    feature.ids = ids,
    feature.labels = labels,
    n0 = 2,
    low.freq.policy = "pure"
  )
  d2 <- refine.linf.csts(
    M,
    d1,
    n0 = 1,
    refinement.factor = 2,
    sep = "__",
    low.freq.policy = "pure",
    verbose = FALSE
  )

  expect_true("cst.id.levels" %in% names(d2))
  expect_true(all(grepl("^asv_1(__|$)|^asv_4(__|$)", d2$cst.id.levels[[2]][!is.na(d2$cst.id.levels[[2]])])))
  expect_true(all(grepl("^L\\. iners 1(__|$)|^L\\. iners 4(__|$)", d2$cst.levels[[2]][!is.na(d2$cst.levels[[2]])])))
})

test_that("refine.linf.csts skips refinement of the synthetic rare bucket", {
  M <- rbind(
    a1 = c(0.95, 0.03, 0.01, 0.01, 0.00, 0.00),
    a2 = c(0.94, 0.04, 0.01, 0.01, 0.00, 0.00),
    a3 = c(0.93, 0.04, 0.02, 0.01, 0.00, 0.00),
    a4 = c(0.92, 0.05, 0.02, 0.01, 0.00, 0.00),
    b1 = c(0.04, 0.94, 0.01, 0.01, 0.00, 0.00),
    b2 = c(0.05, 0.93, 0.01, 0.01, 0.00, 0.00),
    b3 = c(0.05, 0.92, 0.02, 0.01, 0.00, 0.00),
    b4 = c(0.06, 0.91, 0.02, 0.01, 0.00, 0.00),
    c1 = c(0.05, 0.05, 0.90, 0.00, 0.00, 0.00),
    d1 = c(0.05, 0.05, 0.00, 0.90, 0.00, 0.00),
    e1 = c(0.05, 0.05, 0.00, 0.00, 0.90, 0.00),
    f1 = c(0.05, 0.05, 0.00, 0.00, 0.00, 0.90)
  )
  ids <- paste0("asv_", seq_len(ncol(M)))
  labels <- LETTERS[seq_len(ncol(M))]

  d1 <- linf.csts(
    M,
    feature.ids = ids,
    feature.labels = labels,
    n0 = 2,
    low.freq.policy = "pure"
  )

  expect_equal(sum(d1$cell.id == d1$rare.label), 4L)

  d2 <- refine.linf.csts(
    M,
    d1,
    n0 = 2,
    refinement.factor = 2,
    sep = "__",
    low.freq.policy = "pure",
    verbose = FALSE
  )

  expect_equal(d2$cst.depth, 2L)
  expect_equal(sum(d2$cst.id.levels[[2]] == d2$rare.label), 4L)
  expect_false(any(grepl(paste0("^", d2$rare.label, "__"), d2$cst.id.levels[[2]])))
})

test_that("linf.landmarks computes feature-specific landmarks at depth 1", {
  M <- rbind(
    s1 = c(1.00, 0.20, 0.10),
    s2 = c(0.90, 0.40, 0.30),
    s3 = c(0.70, 0.60, 0.10),
    s4 = c(0.20, 1.00, 0.10),
    s5 = c(0.30, 0.90, 0.20)
  )
  ids <- c("asv_1", "asv_2", "asv_3")
  labels <- c("A", "B", "C")

  csts <- linf.csts(
    M,
    feature.ids = ids,
    feature.labels = labels,
    n0 = 2,
    low.freq.policy = "absorb"
  )
  out <- linf.landmarks(
    M,
    csts,
    landmark.types = c("endpoint.max", "endpoint.min", "mean.rep", "median.rep")
  )

  a.rows <- out$landmarks[out$landmarks$cell.id == "asv_1", , drop = FALSE]

  expect_s3_class(out, "linf.landmarks")
  expect_equal(a.rows$landmark.type, c("endpoint.max", "endpoint.min", "mean.rep", "median.rep"))
  expect_equal(a.rows$point.name, c("s1", "s3", "s2", "s2"))
  expect_true(all(a.rows$target.feature.id == "asv_1"))
  expect_true(all(a.rows$target.feature.label == "A"))
  expect_equal(a.rows$target.value[a.rows$landmark.type == "mean.rep"], mean(c(1.0, 0.9, 0.7)))
  expect_equal(a.rows$abs.deviation[a.rows$landmark.type == "mean.rep"], abs(0.9 - mean(c(1.0, 0.9, 0.7))))
})

test_that("linf.csts can attach landmarks and skip rare buckets", {
  A <- matrix(c(5, 1, 0), nrow = 6, ncol = 3, byrow = TRUE)
  B <- matrix(c(1, 5, 0), nrow = 2, ncol = 3, byrow = TRUE)
  C <- matrix(c(1, 0, 5), nrow = 2, ncol = 3, byrow = TRUE)
  M <- rbind(A, B, C)
  M <- sweep(M, 1, rowSums(M), "/")
  colnames(M) <- c("Dom1", "Dom2", "Dom3")

  out <- linf.csts(
    M,
    n0 = 5,
    low.freq.policy = "pure",
    return.landmarks = TRUE,
    landmark.types = "endpoint.max",
    landmark.view = "rare"
  )

  rare.row <- out$landmarks$cells[out$landmarks$cells$cell.id == out$rare.label, , drop = FALSE]

  expect_s3_class(out$landmarks, "linf.landmarks")
  expect_identical(out$landmarks$view, "rare")
  expect_equal(nrow(rare.row), 1L)
  expect_false(rare.row$landmarks.computable)
  expect_true(all(out$landmarks$landmarks$cell.id != out$rare.label))
})

test_that("linf.landmarks uses leaf feature ids for refined depth-2 dominance-lineages", {
  M <- rbind(
    r1 = c(1.00, 0.80, 0.20),
    r2 = c(0.95, 0.70, 0.30),
    r3 = c(0.90, 0.20, 0.70),
    r4 = c(0.85, 0.10, 0.80),
    r5 = c(0.20, 1.00, 0.10),
    r6 = c(0.30, 0.95, 0.10)
  )
  ids <- c("asv_1", "asv_2", "asv_3")
  labels <- c("A", "B", "C")

  d1 <- linf.csts(
    M,
    feature.ids = ids,
    feature.labels = labels,
    n0 = 2,
    low.freq.policy = "pure"
  )
  d2 <- refine.linf.csts(
    M,
    d1,
    n0 = 2,
    refinement.factor = 2,
    low.freq.policy = "pure",
    verbose = FALSE
  )
  out <- linf.landmarks(M, d2, depth = 2, view = "rare", landmark.types = "endpoint.max")

  c.row <- out$landmarks[out$landmarks$cell.id == "asv_1__asv_3", , drop = FALSE]

  expect_true("cst.id.levels.rare" %in% names(d2))
  expect_true("cst.id.levels.absorb" %in% names(d2))
  expect_equal(c.row$target.feature.id, "asv_3")
  expect_equal(c.row$target.feature.label, "C")
  expect_equal(c.row$point.name, "r4")
})

test_that("switching dCST views updates ids and drops stale attached landmarks", {
  M <- rbind(
    s1 = c(1.0, 0.2, 0.1),
    s2 = c(0.9, 0.3, 0.1),
    s3 = c(0.1, 1.0, 0.1)
  )
  colnames(M) <- c("A", "B", "C")

  csts <- linf.csts(
    M,
    n0 = 3,
    low.freq.policy = "pure",
    return.landmarks = TRUE
  )
  collapsed <- collapse.rare(csts)

  expect_null(collapsed$landmarks)
  expect_identical(collapsed$cell.id, collapsed$cell.id.absorb)
  expect_identical(collapsed$cst.id.levels[[collapsed$cst.depth]], collapsed$cst.id.levels.absorb[[collapsed$cst.depth]])
})

test_that('low.freq.policy = "rare" is accepted as a deprecated alias for "pure"', {
  M <- rbind(
    s1 = c(1.0, 0.2, 0.1),
    s2 = c(0.9, 0.3, 0.1),
    s3 = c(0.1, 1.0, 0.1)
  )
  colnames(M) <- c("A", "B", "C")

  expect_warning(
    csts <- linf.csts(M, n0 = 2, low.freq.policy = "rare"),
    'deprecated; use "pure" instead'
  )

  expect_identical(csts$low.freq.policy, "pure")
  expect_identical(csts$cell.label, csts$cell.label.rare)
})

test_that("normalize.linf max-per-row", {
  S <- rbind(c(1, 2, 4), c(0, 0, 0))
  Z <- normalize.linf(S)
  expect_equal(apply(Z, 1, max), c(1, 0))
})

test_that("linf.cells ties, zeros, labels, levels", {
  S <- rbind(c(0.7, 0.3),
             c(0, 0),
             c(1, 1))   # tie -> first
  colnames(S) <- c("A", "B")

  out <- linf.cells(S)

  # index behavior
  expect_equal(out$index, c(1L, NA_integer_, 1L))

  # label behavior (mapped from levels)
  expect_equal(out$label, c("A", NA_character_, "A"))

  # levels must be full column-ordered set, 1–1 with columns
  expect_equal(out$levels, c("A", "B"))

  # observed.levels is a subset of levels, in column order, no NA
  expect_equal(out$observed.levels, "A")
})

test_that("linf.cells unnamed columns synthesize unique levels", {
  T <- matrix(c(0,2,  3,1,  0,0), nrow = 3, byrow = TRUE)
  # no colnames -> V1, V2
  out <- linf.cells(T)

  expect_equal(out$levels, c("V1", "V2"))     # synthesized
  expect_equal(out$index, c(2L, 1L, NA_integer_))
  expect_equal(out$label, c("V2", "V1", NA_character_))
  expect_equal(out$observed.levels, c("V1", "V2"))  # both appeared
})

test_that("linf.cells duplicate column names get disambiguated", {
  T <- matrix(c(0, 2,
                3, 1,
                0, 0), nrow = 3, byrow = TRUE)
  colnames(T) <- c("X", "X") # duplicates

  out <- linf.cells(T)
  # expected: make.unique -> "X", "X_1"
  expect_equal(out$levels, c("X", "X_1"))
  expect_equal(out$label[1:2], c("X_1", "X"))
})

test_that("linf.csts retains supported states and reassigns low-support ones", {
  # 6 rows dominated by col1, 2 by col2, 2 by col3
  A <- matrix(c(5, 1, 0), nrow = 6, ncol = 3, byrow = TRUE)
  B <- matrix(c(1, 5, 0), nrow = 2, ncol = 3, byrow = TRUE)
  C <- matrix(c(1, 0, 5), nrow = 2, ncol = 3, byrow = TRUE)
  S <- rbind(A, B, C)
  S <- sweep(S, 1, rowSums(S), "/")
  colnames(S) <- c("Dom1", "Dom2", "Dom3")

  out <- linf.csts(S, n0 = 5, low.freq.policy = "absorb")

  # only col1 kept
  expect_identical(out$kept.cells.idx, 1L)
  expect_identical(out$kept.cells.lbl, "Dom1")

  # supported state unchanged
  expect_true(all(out$cell.index[1:6] == 1L))
  expect_true(all(out$cell.label[1:6] == "Dom1"))

  # low-support samples reassigned to the retained state
  expect_true(all(out$cell.index[7:10] == 1L))
  expect_true(all(out$cell.label[7:10] == "Dom1"))
})

test_that("linf.csts returns all-NA when no states meet the support threshold", {
  S <- rbind(
    c(1, 0, 0),  # 1 sample in col1
    c(0, 1, 0),  # 1 sample in col2
    c(0, 0, 1)   # 1 sample in col3
  )
  colnames(S) <- c("A","B","C")
  out <- linf.csts(S, n0 = 2)

  expect_true(all(is.na(out$cell.index)))
  expect_true(all(out$cell.label == out$rare.label))
  expect_length(out$kept.cells.idx, 0)
  expect_length(out$kept.cells.lbl, 0)

  # size.table reports raw sizes by label
  expect_identical(as.integer(unclass(out$size.table)), c(1L,1L,1L))
  expect_identical(names(out$size.table), c("A","B","C"))
})

test_that("normalize.linf returns the same values under dense and sparse backends", {
  M <- rbind(
    s1 = c(5, 2, 0, 0),
    s2 = c(0, 0, 0, 0),
    s3 = c(1, 4, 4, 0),
    s4 = c(0, 3, 0, 6)
  )

  dense <- normalize.linf(M, backend = "dense")
  sparse <- normalize.linf(Matrix::Matrix(M, sparse = TRUE), backend = "auto")

  expect_true(inherits(sparse, "dgCMatrix"))
  expect_equal(as.matrix(sparse), dense)
})

test_that("linf.cells matches between dense and sparse backends", {
  M <- rbind(
    s1 = c(1, 0.5, 0),
    s2 = c(0, 0, 0),
    s3 = c(0.2, 0.8, 0.8),
    s4 = c(0, 0.4, 0.9)
  )
  ids <- c("asv_1", "asv_2", "asv_3")
  labels <- c("A", "B", "C")

  dense <- linf.cells(M, feature.ids = ids, feature.labels = labels, backend = "dense")
  sparse <- linf.cells(Matrix::Matrix(M, sparse = TRUE), feature.ids = ids, feature.labels = labels, backend = "auto")

  expect_equal(sparse$index, dense$index)
  expect_equal(sparse$id, dense$id)
  expect_equal(sparse$label, dense$label)
  expect_equal(sparse$observed.id.levels, dense$observed.id.levels)
})

test_that("linf.csts and refinement agree across dense and sparse backends", {
  M <- rbind(
    r1 = c(1.00, 0.70, 0.10, 0.00),
    r2 = c(0.95, 0.75, 0.10, 0.00),
    r3 = c(0.90, 0.20, 0.80, 0.00),
    r4 = c(0.88, 0.10, 0.82, 0.00),
    r5 = c(0.10, 1.00, 0.05, 0.00),
    r6 = c(0.12, 0.95, 0.08, 0.00),
    r7 = c(0.05, 0.10, 0.05, 1.00),
    r8 = c(0.06, 0.12, 0.04, 0.95)
  )
  ids <- paste0("asv_", seq_len(ncol(M)))
  labels <- LETTERS[seq_len(ncol(M))]

  dense.d1 <- linf.csts(
    M,
    feature.ids = ids,
    feature.labels = labels,
    n0 = 2,
    low.freq.policy = "absorb",
    backend = "dense"
  )
  sparse.input <- Matrix::Matrix(M, sparse = TRUE)
  sparse.d1 <- linf.csts(
    sparse.input,
    feature.ids = ids,
    feature.labels = labels,
    n0 = 2,
    low.freq.policy = "absorb",
    backend = "auto"
  )

  expect_identical(sparse.d1$matrix.backend, "sparse")
  expect_equal(sparse.d1$cell.id, dense.d1$cell.id)
  expect_equal(sparse.d1$cell.label, dense.d1$cell.label)
  expect_equal(sparse.d1$cell.label.absorb, dense.d1$cell.label.absorb)
  expect_equal(sparse.d1$cell.label.rare, dense.d1$cell.label.rare)

  dense.d2 <- refine.linf.csts(
    M,
    dense.d1,
    n0 = 2,
    refinement.factor = 2,
    low.freq.policy = "absorb",
    verbose = FALSE,
    backend = "dense"
  )
  sparse.d2 <- refine.linf.csts(
    sparse.input,
    sparse.d1,
    n0 = 2,
    refinement.factor = 2,
    low.freq.policy = "absorb",
    verbose = FALSE
  )

  expect_identical(sparse.d2$matrix.backend, "sparse")
  expect_equal(sparse.d2$cst.levels[[2]], dense.d2$cst.levels[[2]])
  expect_equal(sparse.d2$cst.id.levels[[2]], dense.d2$cst.id.levels[[2]])
})

test_that("linf.landmarks and the landmark pipeline accept sparse matrices", {
  M <- rbind(
    s1 = c(10, 8, 1),
    s2 = c(9, 7, 2),
    s3 = c(8, 2, 7),
    s4 = c(7, 1, 8),
    s5 = c(1, 10, 2),
    s6 = c(2, 9, 1)
  )
  ids <- c("asv_1", "asv_2", "asv_3")
  labels <- c("A", "B", "C")
  sparse.M <- Matrix::Matrix(M, sparse = TRUE)

  sparse.rel <- normalize.linf(sparse.M, backend = "auto")
  sparse.d1 <- linf.csts(
    sparse.rel,
    feature.ids = ids,
    feature.labels = labels,
    n0 = 2,
    low.freq.policy = "pure",
    return.landmarks = TRUE,
    backend = "auto"
  )

  expect_s3_class(sparse.d1$landmarks, "linf.landmarks")

  pipe.out <- linf.dcst.landmark.pipeline(
    sparse.M,
    feature.ids = ids,
    feature.labels = labels,
    n0.depth1 = 2,
    n0.depth2 = 2,
    refinement.factor = 2,
    low.freq.policy = "pure",
    landmark.view = "absorb",
    verbose = FALSE,
    backend = "auto"
  )

  expect_identical(pipe.out$dcst.depth1$matrix.backend, "sparse")
  expect_s3_class(pipe.out$landmarks.depth2, "linf.landmarks")
})

test_that("public dCST objects use dominance-lineage terminology", {
  M <- rbind(
    s1 = c(A = 1.0, B = 0.2),
    s2 = c(A = 0.9, B = 0.3),
    s3 = c(A = 0.1, B = 1.0)
  )

  fit <- linf.csts(M, n0 = 2, low.freq.policy = "pure")

  expect_true(all(c(
    "depth1.feature.index", "depth1.feature.id", "depth1.feature.label",
    "lineage.id", "lineage.label", "lineage.ids", "lineage.labels", "depth"
  ) %in% names(fit)))
  expect_false(any(c(
    "cell.index", "cell.id", "cell.label", "kept.cells.idx",
    "cst.id.levels", "cst.levels", "cst.depth"
  ) %in% names(fit)))

  landmarks <- linf.landmarks(M, fit, view = "pure")
  expect_true("lineages" %in% names(landmarks))
  expect_false("cells" %in% names(landmarks))
  expect_true(all(c("lineage.id", "lineage.label", "lineage.size") %in%
                    names(landmarks$lineages)))
})

test_that("removed public aliases are not present", {
  ns <- asNamespace("linf")

  expect_false(exists("linf.cells", envir = ns, inherits = FALSE))
  expect_false(exists("collapse.rare", envir = ns, inherits = FALSE))
  expect_false(exists("expand.rare", envir = ns, inherits = FALSE))
})

`%||%` <- function(a, b) if (is.null(a)) b else a

#' Filter ASV count matrix by library size (samples) and prevalence (features)
#'
#' @description
#' Filters an ASV **count** matrix by:
#' (1) removing samples with library size below \code{min.lib};
#' (2) keeping features (taxa) that are "present" in at least
#' \code{ceiling(prev.prop * nrow(S.counts))} samples, where presence is
#' defined by either \code{min.count} (counts) or \code{min.rel} (relative).
#' After feature filtering, zero-total samples are dropped and a row-normalized
#' relative-abundance matrix is returned alongside the filtered counts.
#'
#' @param S.counts Numeric matrix (samples x features) of nonnegative counts.
#' @param min.lib Integer. Minimum library size (row sum of counts) to keep a sample. Default: 1000.
#' @param prev.prop Numeric in (0,1]. Minimum fraction of samples where a feature must be present.
#'   Default: 0.05.
#' @param min.count Integer (>=1) or NULL. Reads to call "present" (ignored if \code{min.rel} set). Default: 2.
#' @param min.rel Numeric in (0,1) or NULL. Relative abundance to call "present" (overrides \code{min.count}). Default: NULL.
#' @param min.feat.total Integer (>=0) or NULL. Optional minimum total reads across all samples per feature. Default: NULL.
#' @param verbose Logical. Print keep/drop summaries. Default: TRUE.
#'
#' @details
#' Sample filtering is applied on raw counts first. Prevalence is computed on the
#' post-sample-filter matrix using either a count or relative rule. A feature is
#' retained if prevalence \eqn{\ge \lceil \text{prev.prop} \times n_{\text{samples}} \rceil}.
#' After feature filtering, samples with zero remaining counts are dropped and
#' the relative matrix \code{rel} is row-normalized.
#'
#' @return A list with elements:
#' \describe{
#'   \item{counts}{Filtered count matrix (samples x features).}
#'   \item{rel}{Row-normalized relative-abundance matrix.}
#'   \item{kept.sample.idx}{Kept sample indices (original order).}
#'   \item{kept.feature.idx}{Kept feature indices (original order).}
#'   \item{prevalence}{Per-feature prevalence counts for kept features.}
#'   \item{thresholds}{List of thresholds actually used.}
#' }
#'
#' @section Conventions:
#' Dot-delimited names; rows are samples, columns are features. Supply raw counts.
#'
#' @examples
#' set.seed(1)
#' S <- matrix(rpois(100 * 20, lambda = 5), nrow = 100, ncol = 20)
#' res <- filter.asv(S, min.lib = 50, prev.prop = 0.1, min.count = 2)
#' dim(res$counts); dim(res$rel)
#' range(rowSums(res$rel))
#'
#' @export
filter.asv <- function(S.counts,
                       min.lib        = 1000,
                       prev.prop      = 0.05,
                       min.count      = 2,
                       min.rel        = NULL,
                       min.feat.total = NULL,
                       verbose        = TRUE) {

  stopifnot(is.matrix(S.counts) || is.data.frame(S.counts))
  S.counts <- as.matrix(S.counts)
  storage.mode(S.counts) <- "numeric"

  # --- sample filter on raw counts ---
  libsize <- rowSums(S.counts)
  keep.samp <- libsize >= min.lib
  if (verbose) message("Samples kept: ", sum(keep.samp), " / ", nrow(S.counts),
                       "  (min.lib = ", min.lib, ")")
  S.counts <- S.counts[keep.samp, , drop = FALSE]

  if (nrow(S.counts) == 0L || ncol(S.counts) == 0L) {
    warning("No data left after sample filtering.")
    return(list(counts = S.counts, rel = S.counts,
                kept.sample.idx = integer(0), kept.feature.idx = integer(0),
                prevalence = numeric(0)))
  }

  # --- prevalence rule ---
  n <- nrow(S.counts)
  if (!is.null(min.rel)) {
    rs <- rowSums(S.counts)
    present.mat <- sweep(S.counts, 1, rs, "/") >= min.rel
  } else {
    present.mat <- S.counts >= (min.count %||% 1)
  }
  prev <- colSums(present.mat)
  prev.thld <- ceiling(prev.prop * n)
  keep.feat <- prev >= prev.thld

  if (!is.null(min.feat.total)) {
    keep.feat <- keep.feat & (colSums(S.counts) >= min.feat.total)
  }

  if (verbose) {
    message("Features kept: ", sum(keep.feat), " / ", ncol(S.counts),
            "  (prev.prop = ", prev.prop, ", prev.thld = ", prev.thld, ")")
  }

  S.counts <- S.counts[, keep.feat, drop = FALSE]

  # --- drop zero rows; renormalize ---
  lib2 <- rowSums(S.counts)
  keep.samp2 <- lib2 > 0
  if (verbose && any(!keep.samp2)) {
    message("Dropping ", sum(!keep.samp2), " samples with zero counts after feature filtering.")
  }
  S.counts <- S.counts[keep.samp2, , drop = FALSE]

  rs2 <- rowSums(S.counts)
  S.rel <- sweep(S.counts, 1, rs2, "/")

  if (nrow(S.rel) > 0) {
    s <- rowSums(S.rel)
    stopifnot(all(is.finite(S.rel)), all(abs(s - 1) < 1e-8))
  }

  list(
    counts = S.counts,
    rel = S.rel,
    kept.sample.idx  = which(keep.samp)[keep.samp2],
    kept.feature.idx = which(keep.feat),
    prevalence = prev[keep.feat],
    thresholds = list(min.lib = min.lib,
                      prev.prop = prev.prop,
                      prev.thld = prev.thld,
                      min.count = min.count,
                      min.rel = min.rel,
                      min.feat.total = min.feat.total)
  )
}

resolve.linf.cst.levels <- function(csts,
                                    kind = c("id", "label"),
                                    view = c("active", "rare", "absorb")) {
  kind <- match.arg(kind)
  view <- match.arg(view)

  if (kind == "id") {
    active <- csts$cst.id.levels
    if (is.null(active)) active <- list(level1 = csts$cell.id %||% csts$cell.label)

    rare <- csts$cst.id.levels.rare
    if (is.null(rare) && !is.null(csts$cell.id.rare)) rare <- list(level1 = csts$cell.id.rare)

    absorb <- csts$cst.id.levels.absorb
    if (is.null(absorb) && !is.null(csts$cell.id.absorb)) absorb <- list(level1 = csts$cell.id.absorb)
  } else {
    active <- csts$cst.levels
    if (is.null(active)) active <- list(level1 = csts$cell.label)

    rare <- csts$cst.levels.rare
    if (is.null(rare) && !is.null(csts$cell.label.rare)) rare <- list(level1 = csts$cell.label.rare)

    absorb <- csts$cst.levels.absorb
    if (is.null(absorb) && !is.null(csts$cell.label.absorb)) absorb <- list(level1 = csts$cell.label.absorb)
  }

  rare <- rare %||% active
  absorb <- absorb %||% active

  switch(view,
         active = active,
         rare = rare,
         absorb = absorb)
}

resolve.linf.landmark.view <- function(csts,
                                       view = c("active", "rare", "absorb")) {
  view <- match.arg(view)
  if (view == "active") {
    return(linf.active.low.freq.view(csts$low.freq.policy %||% "active"))
  }
  view
}

resolve.linf.landmark.feature <- function(leaf.id,
                                          feature.ids,
                                          feature.labels,
                                          rare.label) {
  is.rare <- !is.na(leaf.id) && identical(leaf.id, rare.label)

  if (is.na(leaf.id) || is.rare) {
    return(list(
      computable = FALSE,
      is.rare = is.rare,
      index = NA_integer_,
      feature.id = NA_character_,
      feature.label = NA_character_
    ))
  }

  idx <- match(leaf.id, feature.ids)
  if (is.na(idx)) {
    idx <- match(leaf.id, feature.labels)
  }

  if (is.na(idx)) {
    return(list(
      computable = FALSE,
      is.rare = FALSE,
      index = NA_integer_,
      feature.id = NA_character_,
      feature.label = NA_character_
    ))
  }

  list(
    computable = TRUE,
    is.rare = FALSE,
    index = idx,
    feature.id = feature.ids[[idx]],
    feature.label = feature.labels[[idx]]
  )
}

choose.linf.landmark.index <- function(scores,
                                       tie.method = c("first", "random", "error"),
                                       minimize = FALSE) {
  tie.method <- match.arg(tie.method)

  best <- if (isTRUE(minimize)) min(scores) else max(scores)
  idx <- which(scores == best)

  if (length(idx) == 1L) return(idx)
  if (tie.method == "first") return(idx[1L])
  if (tie.method == "random") return(sample(idx, 1L))

  stop("linf.landmarks: tie encountered and tie.method = 'error'")
}

empty.linf.landmark.cells <- function() {
  data.frame(
    cell.id = character(0),
    cell.label = character(0),
    cell.size = integer(0),
    target.feature.id = character(0),
    target.feature.label = character(0),
    is.rare = logical(0),
    landmarks.computable = logical(0),
    stringsAsFactors = FALSE
  )
}

empty.linf.landmark.rows <- function() {
  data.frame(
    cell.id = character(0),
    cell.label = character(0),
    landmark.type = character(0),
    point.index = integer(0),
    point.name = character(0),
    target.feature.id = character(0),
    target.feature.label = character(0),
    observed.value = numeric(0),
    target.value = numeric(0),
    abs.deviation = numeric(0),
    stringsAsFactors = FALSE
  )
}

#' Landmark points for dCST dominance-lineages
#'
#' @description
#' Computes representative landmark points for the dominance-lineages of a
#' \code{"linf.csts"} object at a chosen depth and view.
#'
#' Landmark types are defined with respect to the leaf feature of the dCST path:
#' the last feature ID in the lineage ID path.
#' Lineages whose leaf token is \code{rare.label} are reported but skipped for
#' landmark computation because they do not correspond to a unique target
#' feature.
#'
#' @param M Numeric matrix (samples x features) used to build or refine the dCSTs.
#' @param csts A \code{"linf.csts"} object.
#' @param depth Integer. dCST depth to inspect. Defaults to the leaf depth
#'   \code{csts$cst.depth}.
#' @param view Character. One of \code{"active"}, \code{"rare"}, or
#'   \code{"absorb"}.
#' @param landmark.types Character vector containing any of
#'   \code{"endpoint.max"}, \code{"endpoint.min"}, \code{"mean.rep"}, or
#'   \code{"median.rep"}.
#' @param tie.method Character. Tie handling for landmark selection:
#'   \code{"first"}, \code{"random"}, or \code{"error"}.
#' @param backend Character. Matrix backend to use: \code{"auto"},
#'   \code{"dense"}, or \code{"sparse"}. The default \code{"auto"} preserves
#'   sparse input and otherwise uses the dense path.
#'
#' @return A list of class \code{"linf.landmarks"} with components:
#' \itemize{
#'   \item \code{depth}, \code{view}, \code{sep}, \code{rare.label}
#'   \item \code{feature.ids}, \code{feature.labels}
#'   \item \code{cells}: one row per dominance-lineage with computability
#'     metadata; the component name is retained for backward compatibility
#'   \item \code{landmarks}: one row per computed landmark point
#' }
#'
#' @export
linf.landmarks <- function(M,
                           csts,
                           depth = NULL,
                           view = c("active", "rare", "absorb"),
                           landmark.types = c("endpoint.max",
                                              "endpoint.min",
                                              "mean.rep",
                                              "median.rep"),
                           tie.method = c("first", "random", "error"),
                           backend = c("auto", "dense", "sparse")) {
  validate.linf.csts(csts)

  view <- match.arg(view)
  tie.method <- match.arg(tie.method)
  landmark.types <- match.arg(landmark.types, several.ok = TRUE)
  if (missing(backend) && !is.null(csts$matrix.backend)) {
    backend <- csts$matrix.backend
  }

  prep <- linf.prepare.matrix(M, backend = backend, fun.name = "linf.landmarks")
  X <- prep$X
  backend <- prep$backend
  linf.validate.matrix(X, backend = backend, fun.name = "linf.landmarks")
  if (nrow(X) != length(csts$cell.label)) {
    stop("linf.landmarks: nrow(M) must match the number of samples in csts")
  }

  max.depth <- csts$cst.depth %||% 1L
  if (is.null(depth)) depth <- max.depth

  if (!is.numeric(depth) || length(depth) != 1L || depth < 1L || depth %% 1 != 0) {
    stop("linf.landmarks: depth must be an integer >= 1")
  }
  if (depth > max.depth) {
    stop("linf.landmarks: requested depth exceeds csts$cst.depth")
  }

  resolved.view <- resolve.linf.landmark.view(csts, view)
  id.levels <- resolve.linf.cst.levels(csts, kind = "id", view = resolved.view)
  label.levels <- resolve.linf.cst.levels(csts, kind = "label", view = resolved.view)

  if (length(id.levels) < depth || length(label.levels) < depth) {
    stop("linf.landmarks: requested view does not contain the requested depth")
  }

  cell.ids <- as.character(id.levels[[depth]])
  cell.labels <- as.character(label.levels[[depth]])

  if (length(cell.ids) != nrow(X) || length(cell.labels) != nrow(X)) {
    stop("linf.landmarks: dCST levels at the requested depth must align with rows of M")
  }

  meta <- resolve.linf.feature.meta(
    X,
    feature.ids = csts$feature.ids,
    feature.labels = csts$feature.labels
  )

  sep <- csts$sep %||% "__"
  rare.label <- csts$rare.label %||% "RARE_DOMINANT"
  row.ids <- rownames(X)
  if (is.null(row.ids)) row.ids <- rep(NA_character_, nrow(X))

  cells <- empty.linf.landmark.cells()
  landmarks <- empty.linf.landmark.rows()
  unique.cells <- unique(cell.ids[!is.na(cell.ids)])

  for (cell in unique.cells) {
    members <- which(cell.ids == cell)
    cell.label <- cell.labels[members[1L]]
    parts <- strsplit(cell, sep, fixed = TRUE)[[1L]]
    leaf.id <- parts[[length(parts)]]
    target <- resolve.linf.landmark.feature(
      leaf.id,
      feature.ids = meta$feature.ids,
      feature.labels = meta$feature.labels,
      rare.label = rare.label
    )

    cells <- rbind(cells, data.frame(
      cell.id = cell,
      cell.label = cell.label,
      cell.size = length(members),
      target.feature.id = target$feature.id,
      target.feature.label = target$feature.label,
      is.rare = target$is.rare,
      landmarks.computable = target$computable,
      stringsAsFactors = FALSE
    ))

    if (!isTRUE(target$computable)) next

    vals <- as.numeric(X[members, target$index, drop = FALSE])

    for (landmark.type in landmark.types) {
      if (landmark.type == "endpoint.max") {
        pick <- choose.linf.landmark.index(vals, tie.method = tie.method, minimize = FALSE)
        observed <- vals[[pick]]
        target.value <- observed
      } else if (landmark.type == "endpoint.min") {
        pick <- choose.linf.landmark.index(vals, tie.method = tie.method, minimize = TRUE)
        observed <- vals[[pick]]
        target.value <- observed
      } else if (landmark.type == "mean.rep") {
        target.value <- mean(vals)
        dev <- abs(vals - target.value)
        pick <- choose.linf.landmark.index(dev, tie.method = tie.method, minimize = TRUE)
        observed <- vals[[pick]]
      } else if (landmark.type == "median.rep") {
        target.value <- stats::median(vals)
        dev <- abs(vals - target.value)
        pick <- choose.linf.landmark.index(dev, tie.method = tie.method, minimize = TRUE)
        observed <- vals[[pick]]
      } else {
        stop("linf.landmarks: unsupported landmark type")
      }

      point.index <- members[[pick]]
      landmarks <- rbind(landmarks, data.frame(
        cell.id = cell,
        cell.label = cell.label,
        landmark.type = landmark.type,
        point.index = point.index,
        point.name = row.ids[[point.index]],
        target.feature.id = target$feature.id,
        target.feature.label = target$feature.label,
        observed.value = observed,
        target.value = target.value,
        abs.deviation = abs(observed - target.value),
        stringsAsFactors = FALSE
      ))
    }
  }

  structure(
    list(
      depth = as.integer(depth),
      view = resolved.view,
      sep = sep,
      rare.label = rare.label,
      feature.ids = meta$feature.ids,
      feature.labels = meta$feature.labels,
      cells = cells,
      landmarks = landmarks
    ),
    class = "linf.landmarks"
  )
}

test_that("bundled data provenance strings identify the intended sources", {
  data("valencia2k", package = "linf")
  expect_match(
    valencia2k$source,
    "doi:10[.]1186/s40168-020-00934-6",
    fixed = FALSE
  )
  expect_false(grepl("10[.]1128/mSystems[.]00149-20", valencia2k$source))

  data("agp_gut", package = "linf")
  expect_match(agp_gut$source, "Phenotypes do not influence selection")
  expect_match(agp_gut$source, "selection reasons")
  expect_match(agp_gut$source, "DATA_PROVENANCE[.]md")
  expect_setequal(
    unique(agp_gut$meta$selection_reason),
    c("all_target_dcst", "seeded_random_background")
  )
  expect_equal(
    unname(table(agp_gut$meta$selection_reason)["all_target_dcst"]),
    287L
  )
  expect_equal(
    unname(table(agp_gut$meta$selection_reason)["seeded_random_background"]),
    479L
  )
})

test_that("installed data provenance notice is present", {
  provenance <- system.file("DATA_PROVENANCE.md", package = "linf")
  expect_true(nzchar(provenance))
  expect_true(file.exists(provenance))
})

---
title: "dCSTs for Vaginal Microbiome Data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{dCSTs for Vaginal Microbiome Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>", fig.width = 7, fig.height = 5)
library(linf)
```

## Introduction

Community state types (CSTs) are the standard way to describe the structure of
vaginal microbial communities. They were originally defined by hierarchical
clustering of 16S rRNA data (Ravel et al. 2011) and later standardized by the
VALENCIA nearest-centroid classifier (France et al. 2020).

**Dominant Community State Types (dCSTs)** offer an alternative that requires no
clustering and no training data. A dCST is defined by
$L^\infty$-normalization: divide each sample's abundance vector by its maximum
value, then group samples by which taxon achieves that maximum. This is a
purely geometric operation on compositional data --- it identifies the dominant
species in each sample and groups samples accordingly.

This vignette demonstrates the dCST workflow on a real vaginal 16S dataset and
shows that dCSTs recover the established CST structure with high concordance.

**Reference:** Gajer, P. & Ravel, J. (2025). A New Approach to Compositional
Data Analysis using $L^\infty$-normalization with Applications to Vaginal
Microbiome. arXiv:2503.21543.

## The Valencia 2k dataset

The `valencia2k` dataset is a stratified subsample of 2,000 vaginal samples from
the Valencia CST-classifier training set (France et al. 2020). It contains
species-level relative abundances for 178 taxa along with the original Valencia
CST and sub-CST assignments.

```{r load-data}
data(valencia2k)

## Compositional matrix: 2000 samples x 178 taxa
dim(valencia2k$rel)

## Valencia CST assignments
table(valencia2k$cst$Val_CST)

## Top 10 most abundant taxa (by mean relative abundance)
means <- sort(colMeans(valencia2k$rel), decreasing = TRUE)
head(round(means, 4), 10)
```

The dominant taxa are familiar vaginal species: *Lactobacillus iners*,
*L. crispatus*, *Gardnerella vaginalis*, and others. This is exactly the kind
of data where dominant-feature assignments are most intuitive: most samples
are dominated by a single species, and the identity of that species is
biologically meaningful.

## Reconstructing counts

The bundled matrix is compositional (rows sum to 1). For workflows that require
count-like input (e.g., `filter.asv()`), we can reconstruct approximate counts
using the per-sample read depths:

```{r reconstruct-counts}
count_mat <- sweep(valencia2k$rel, 1, valencia2k$reads, "*")
count_mat <- round(count_mat)
storage.mode(count_mat) <- "integer"

## Check: library sizes
summary(rowSums(count_mat))
```

Since the Valencia 13k data was already filtered (min 3,000 reads, taxa present
in $\ge$ 1% of samples), we skip `filter.asv()` here and work directly with
the compositional matrix.

## $L^\infty$ normalization

$L^\infty$-normalization divides each row by its maximum value. After
normalization, every sample has at least one coordinate equal to 1, and all
other coordinates lie in $[0, 1]$.

```{r linf-normalize}
Z <- normalize.linf(valencia2k$rel)

## Every row max should be 1
summary(apply(Z, 1, max))
```

## Dominance sample sets: dominant species assignment

The dominant species is the column that achieves the maximum in each row.
Samples with the same dominant species form a depth-1 dominance sample set.
The assignment is deterministic and sample-independent: it depends only on
the sample's own abundance profile, never on other samples.

```{r linf-cells}
cells <- linf.cells(Z)

## How many distinct dominant taxa?
cat("Distinct dominant taxa:", length(cells$observed.levels), "\n")

## Dominance sample-set size distribution (top 15)
dominance_tab <- sort(table(cells$label), decreasing = TRUE)
head(dominance_tab, 15)
```

We see the expected vaginal community structure: a few well-supported
dominance sample sets (the *Lactobacillus* species, *Gardnerella*, BVAB1) and
a long tail of rarer dominant taxa.

## Depth-1 dCSTs

In practice, many provisional dominance sample sets contain very few samples.
**Dominant Community State Types (dCSTs)** retain only states above a minimum
support threshold ($n_0$) and handle samples from low-support sets according
to the selected policy.

The package supports two policies for handling low-support provisional sets:

- **"pure"**: named dCSTs stay pure, and samples from low-support sets are
  labeled `RARE_DOMINANT`
- **"absorb"**: samples from low-support sets are reassigned to a retained
  state by restricted argmax

The legacy policy name `"rare"` is still accepted as a deprecated alias for `"pure"`.

```{r depth1-dcsts}
dcst1 <- linf.csts(Z, n0 = 30, low.freq.policy = "pure")

## Retained depth-1 states
dcst1$kept.cells.lbl

## dCST assignment (pure policy)
dcst1_tab <- sort(table(dcst1$cell.label), decreasing = TRUE)
dcst1_tab
```

### Concordance with Valencia CSTs

The key test: do dCSTs correspond to the established Valencia CSTs? Let's
build a cross-tabulation:

```{r concordance-depth1}
## Build concordance table: rows = dCSTs, columns = Valencia CSTs
dcst_labels <- dcst1$cell.label.absorb  # use absorb view for clean comparison
val_cst     <- valencia2k$cst$Val_CST

concordance <- table(dCST = dcst_labels, Valencia = val_cst)

## Show as percentage: what fraction of each dCST falls in each Valencia CST?
pct <- round(100 * prop.table(concordance, margin = 1), 1)
pct
```

The concordance is striking. Each dCST maps predominantly to a single Valencia
CST:

- *Lactobacillus crispatus* dCST $\to$ CST I (dominated by *L. crispatus*)
- *Lactobacillus iners* dCST $\to$ CST III (dominated by *L. iners*)
- *Gardnerella vaginalis* dCST $\to$ CST IV-B
- BVAB1 dCST $\to$ CST IV-A
- *Lactobacillus gasseri* dCST $\to$ CST II
- *Lactobacillus jensenii* dCST $\to$ CST V

```{r concordance-reverse}
## Reverse view: what fraction of each Valencia CST falls in each dCST?
pct_rev <- round(100 * prop.table(concordance, margin = 2), 1)
pct_rev
```

This two-way concordance confirms that dCSTs are an alternative construct of
CSTs: they recover the same groupings without any clustering algorithm or
training data.

## Depth-2 refinement

Large dCSTs can be refined by dropping the dominant species and repeating the
procedure on the remaining taxa. This produces a two-level hierarchy that
reveals sub-community structure.

```{r depth2-refine}
dcst2 <- refine.linf.csts(Z, dcst1, n0 = 15, refinement.factor = 2,
                           low.freq.policy = "pure", verbose = FALSE)

## Depth-2 dCST table
dcst2_tab <- sort(table(dcst2$cell.label), decreasing = TRUE)
dcst2_tab
```

### Concordance with Valencia sub-CSTs

```{r concordance-depth2}
dcst2_labels <- dcst2$cell.label.absorb
val_subcst   <- valencia2k$cst$Val_subCST

concordance2 <- table(dCST = dcst2_labels, Valencia = val_subcst)

## Percentage by dCST (row-wise)
pct2 <- round(100 * prop.table(concordance2, margin = 1), 1)

## Show only dCSTs with at least 20 samples for readability
dcst2_sizes <- rowSums(concordance2)
pct2[dcst2_sizes >= 20, , drop = FALSE]
```

The depth-2 refinement recovers sub-CST structure. For example, the
*Lactobacillus crispatus* dCST splits based on the second-most-abundant taxon,
and these sub-dCSTs align with Valencia sub-CSTs I-A and I-B.

## The full pipeline

For convenience, `linf.dcst.landmark.pipeline()` runs the entire workflow in
one call: normalization, depth-1 dCSTs, depth-2 refinement, and landmark
computation.

```{r full-pipeline}
out <- linf.dcst.landmark.pipeline(
  count_mat,
  feature.ids    = colnames(count_mat),
  feature.labels = colnames(count_mat),
  n0.depth1      = 30,
  n0.depth2      = 15,
  refinement.factor = 2,
  low.freq.policy   = "pure",
  landmark.view     = "absorb",
  verbose           = FALSE
)

names(out)
```

## Landmark points

Each dCST dominance-lineage has characteristic landmark points: the most
extreme sample
(highest value of the dominant taxon), the least extreme (lowest value while
still dominant), and mean/median representatives. These help characterize what
a "typical" or "extreme" member of each community type looks like.

```{r landmarks}
## Depth-1 landmarks
lm1 <- out$landmarks.depth1$landmarks

## Show landmarks for a few key dCST dominance-lineages
key_lineages <- c("Lactobacillus_crispatus", "Lactobacillus_iners",
                  "Gardnerella_vaginalis")
key_lm <- lm1[lm1$cell.id %in% key_lineages, ]

## For each dCST, show the endpoint.max landmark's target value
## (how dominant is the dominant species in the most extreme sample?)
ep_max <- key_lm[key_lm$landmark.type == "endpoint.max",
                 c("cell.label", "point.name", "target.value")]
ep_max

## And the endpoint.min (least dominant while still assigned to this lineage)
ep_min <- key_lm[key_lm$landmark.type == "endpoint.min",
                 c("cell.label", "point.name", "target.value")]
ep_min
```

The endpoint.max landmarks show samples where the dominant species comprises
nearly the entire community, while endpoint.min landmarks show the boundary
cases --- samples where dominance is marginal.

## Summary

This vignette demonstrated the dCST workflow on real vaginal 16S data:

1. **$L^\infty$-normalization** maps compositional data to the $L^\infty$-simplex
2. **Dominant-feature assignment** assigns each sample to its dominant taxon
3. **Truncated dCSTs** retain well-supported states and handle rare dominance
   patterns
4. **Depth-2 refinement** reveals sub-community structure
5. **Landmark points** characterize extreme and typical community members

The concordance with established Valencia CSTs is high, validating dCSTs as a
principled, deterministic alternative to clustering-based community typing. Key
advantages:

- **No clustering algorithm:** assignment depends only on the sample itself
- **No training data:** new samples can be assigned without a reference database
- **Interpretable:** dCST labels directly name the dominant species
- **Hierarchical:** refinement naturally produces sub-types

## References

- France, M. T., Ma, B., Gajer, P., et al. (2020). VALENCIA: a nearest
  centroid classification method for vaginal microbial communities based on
  composition. *Microbiome*, 8(1), 166.
- Gajer, P. & Ravel, J. (2025). A New Approach to Compositional Data Analysis
  using $L^\infty$-normalization with Applications to Vaginal Microbiome.
  arXiv:2503.21543.
- Ravel, J., Gajer, P., Abdo, Z., et al. (2011). Vaginal microbiome of
  reproductive-age women. *PNAS*, 108(Suppl 1), 4680--4687.

# Bundled data provenance and upstream terms

The package code is distributed under the package's MIT license. The bundled
microbiome data are derived from the upstream resources described below. The
package license does not replace or narrow applicable upstream attribution
terms.

## VALENCIA-derived objects

The `valencia2k`, `valencia_linf_hypercube_1k`,
`valencia13k_dcst_depth2_merged`, and `valencia13k_dcst_depth3_merged` objects
are derived from the VALENCIA training data:

- Source repository: <https://github.com/ravel-lab/VALENCIA>
- The primary article states that the training dataset is available in that
  repository and applies the Creative Commons CC0 public-domain dedication to
  data made available with the article unless otherwise stated.
- Upstream software repository license: MIT, copyright 2023 ravel-lab
- Primary reference: France et al. (2020), "VALENCIA: a nearest centroid
  classification method for vaginal microbial communities based on
  composition," <https://doi.org/10.1186/s40168-020-00934-6>

The source-preparation scripts in `data-raw/` record the deterministic
subsampling and dCST construction parameters. Original sample identifiers are
not retained in the merged 13k assignment objects.

## American Gut Project demonstration object

The `agp_gut` object is derived from public American Gut Project 16S records
and deidentified metadata:

- ENA/BioProject accession: PRJEB11419
  (<https://www.ebi.ac.uk/ena/browser/view/PRJEB11419>)
- Primary reference: McDonald et al. (2018), "American Gut: an Open Platform
  for Citizen Science Microbiome Research,"
  <https://doi.org/10.1128/mSystems.00031-18>
- The primary article is distributed under CC BY 4.0. It states that the
  project protocol permits public deposition of non-identifying data and that
  sequence data and deidentified participant responses are placed in the
  public domain without access restrictions.
- Processing source: PRIME SILVA species-level outputs dated 2026-03-24.

The full analysis first selected at most 5,000 samples with library size at
least 1,000 using random seed 42. `data-raw/create_agp_gut_subset.py` then
creates the 766-sample package object using an explicit deterministic rule:

1. include every sample assigned to `Prevotella_7`, `Pasteurellaceae`,
   `Akkermansia`, or `Staphylococcus`;
2. exclude depth-1 `Eukaryota` and `Unassigned` labels; and
3. fill the remaining slots by a seed-42 simple random sample from the
   eligible background.

Phenotype fields are joined only after membership is fixed and cannot
influence selection. Exact run-ID membership, selection reasons, dCST labels,
and derived phenotype fields are retained in
`inst/extdata/agp_gut_meta.csv`. The script also reconstructs the bundled
count matrix from the full abundance table and retained 5,000-sample dCST
assignments. `data-raw/build_agp_gut.R` then constructs the package object.

Because inclusion probabilities differ by dCST, `agp_gut` is suitable for
demonstrating data alignment, filtering, normalization, and dCST construction
only. Its phenotype frequencies, effect sizes, and p-values must not be
interpreted as population estimates.

---
title: "Dominant Community State Types: From Normalization to a Gut Demonstration"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Dominant Community State Types: From Normalization to a Gut Demonstration}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{css, echo=FALSE}
table td, table th { white-space: nowrap; }
```

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment  = "#>",
  fig.width = 7,
  fig.height = 5
)
library(linf)
```

The `linf` package implements L-infinity normalization and Dominant Community
State Types (dCSTs) for compositional data. This vignette demonstrates the
complete workflow in two parts: first a quick-start with toy data to introduce
the core functions, then a real-data analysis of gut microbiome samples from
the American Gut Project showing how to inspect dominant and subdominant
community structure.

## Part 1: Quick Start with Toy Data

### L-infinity normalization

L-infinity normalization scales each row of a count matrix by its maximum
value, projecting samples onto the L-infinity unit ball. Every normalized row
has a maximum of exactly 1.

```{r toy-normalize}
set.seed(1)
S.counts <- matrix(
  rpois(30, lambda = 5),
  nrow = 10, ncol = 3,
  dimnames = list(paste0("s", 1:10), c("Taxon_A", "Taxon_B", "Taxon_C"))
)

Z <- normalize.linf(S.counts)
apply(Z, 1, max)   # all 1
```

### Dominance sample sets

Each sample is assigned to its dominant feature: the column achieving the
within-sample maximum. Samples with the same dominant feature form a
depth-1 *dominance sample set*.

```{r toy-cells}
cells <- linf.cells(Z)
table(cells$label)
```

### Truncated dCSTs

`linf.csts()` groups samples by dominant feature and collapses small groups
(below threshold `n0`) into a `RARE_DOMINANT` bucket. This produces
*Dominant Community State Types* (dCSTs).

```{r toy-csts}
A <- matrix(c(5, 1, 0), nrow = 6, ncol = 3, byrow = TRUE)
B <- matrix(c(1, 5, 0), nrow = 2, ncol = 3, byrow = TRUE)
C <- matrix(c(1, 0, 5), nrow = 2, ncol = 3, byrow = TRUE)
S <- rbind(A, B, C)
S <- sweep(S, 1, rowSums(S), "/")
colnames(S) <- c("Dom1", "Dom2", "Dom3")

res <- linf.csts(S, n0 = 5)
table(res$cell.label, useNA = "ifany")
```

Only Dom1 (6 samples) meets the `n0 = 5` threshold. Dom2 and Dom3 are
collapsed into `RARE_DOMINANT`.


## Part 2: Gut Microbiome dCST Demonstration

### Background

Single-species dominance is common in some microbial ecosystems and less
common in others. dCSTs provide a deterministic way to summarize the dominant
feature and then refine large groups by subdominant features. The gut example
below demonstrates those mechanics without treating the selected package
subset as an epidemiologic sample.

### The data

The `agp_gut` dataset bundled with this package contains `r nrow(agp_gut$counts)`
gut microbiome samples from the American Gut Project (PRJEB11419), a citizen-science
16S rRNA V4 survey. The dataset is a dCST-stratified demonstration subset
that includes all samples in four selected uncommon dCSTs and fills
the remaining slots with a seed-42 simple random sample from the eligible
background. Phenotypes are joined only after selection and do not influence
membership. Because inclusion probabilities differ by dCST, it is not a
probability sample of the underlying cohort; phenotype frequencies, effect
sizes, and hypothesis tests must not be interpreted as population results.

```{r load-data}
data(agp_gut)
str(agp_gut, max.level = 1)

dim(agp_gut$counts)   # samples x taxa
```

Metadata includes self-reported health conditions parsed from the AGP
questionnaire:

```{r meta-summary}
disease_cols <- c("IBS", "IBD", "Obesity", "Cardiovascular_disease",
                  "Autoimmune", "Acid_reflux")
sapply(disease_cols, function(col) sum(agp_gut$meta[[col]], na.rm = TRUE))
```

### Step 1: Quality filtering

`filter.asv()` removes low-depth samples and rare taxa in one call. We
require at least 1,000 reads per sample and taxa present in at least 5% of
samples.

```{r filter}
filt <- filter.asv(agp_gut$counts, min.lib = 1000, prev.prop = 0.05,
                   min.count = 2)
dim(filt$counts)
dim(filt$rel)
```

### Step 2: L-infinity normalization

```{r normalize}
M <- normalize.linf(filt$counts)

# Verify: every row max is 1
stopifnot(all(abs(apply(M, 1, max) - 1) < 1e-10))
```

### Step 3: Depth-1 dCSTs

```{r dcst-depth1}
csts <- linf.csts(M, n0 = 30)

# dCST size distribution
sort(table(csts$cell.label), decreasing = TRUE)
```

The landscape is dominated by *Bacteroides* and *Escherichia-Shigella*, with
*Prevotella*, *Faecalibacterium*, and several smaller dCSTs forming the tail.
The `RARE_DOMINANT` bucket collects samples dominated by taxa too infrequent
to form their own dCST at this threshold.

Note on *Escherichia-Shigella*: this genus is known to be inflated in 16S V4
data due to primer cross-reactivity. Its high prevalence should be interpreted
with caution.

### Step 4: Depth-2 refinement

`refine.linf.csts()` subdivides large dCSTs by their *subdominant* species.
For instance, a Bacteroides-dominated sample might be further classified by
whether its second-most-abundant taxon is *Faecalibacterium* or
*Lachnospiraceae*. This captures co-dominance patterns that are especially
important in the diverse gut environment.

```{r dcst-depth2}
csts2 <- refine.linf.csts(M, csts, n0 = 30)

# Show depth-2 dCSTs with >= 20 samples
tab2 <- sort(table(csts2$cell.label), decreasing = TRUE)
tab2[tab2 >= 20]
```

### Dominance strength

A useful diagnostic: the relative abundance of the dominant species in each
sample. In the gut, most samples have modest dominance (< 50%), consistent
with the diverse nature of the ecosystem.

```{r dominance-dist, fig.cap = "Distribution of dominance strength across gut samples. Most samples show moderate dominance, with a tail of strongly dominated communities."}
rel <- filt$rel
dom_strength <- apply(rel, 1, max)
hist(dom_strength, breaks = 50, col = "#2ecc71", border = "white",
     main = "Dominance Strength in Gut Microbiome",
     xlab = "Relative abundance of dominant species",
     ylab = "Number of samples")
abline(v = 0.5, col = "red", lty = 2, lwd = 2)
legend("topright", "50% dominance", col = "red", lty = 2, lwd = 2, bty = "n")
```


### dCST size distribution

```{r dcst-barplot, fig.cap = "dCST size distribution at depth 1. Bacteroides and Escherichia-Shigella dominate; rare dCSTs in the tail are collapsed into RARE_DOMINANT."}
tab1 <- sort(table(csts$cell.label), decreasing = TRUE)
par(mar = c(10, 4, 3, 1))
bp <- barplot(tab1,
        col = ifelse(names(tab1) == "RARE_DOMINANT", "#e74c3c", "#3498db"),
        las = 2, cex.names = 0.7,
        ylab = "Number of samples",
        main = "Gut dCST Size Distribution (depth 1)")
text(bp, tab1 + 5, labels = tab1, cex = 0.7, pos = 3)
```


### Phenotype metadata

The bundle retains selected self-reported phenotype fields so that users can
inspect the object structure and practice data alignment. The following table
is purely descriptive; dCST-stratified selection precludes population
prevalence estimation or treating association calculations as cohort results.

```{r phenotype-summary}
meta <- agp_gut$meta[match(rownames(M), agp_gut$meta$Run), ]
stopifnot(all(meta$Run == rownames(M)))

conditions <- c("IBS", "IBD", "Obesity", "Cardiovascular_disease",
                "Autoimmune", "Acid_reflux", "Lung_disease")
phenotype_summary <- data.frame(
  Condition = conditions,
  Recorded_cases = vapply(
    conditions,
    function(x) sum(meta[[x]] == 1, na.rm = TRUE),
    integer(1)
  ),
  Non_missing = vapply(
    conditions,
    function(x) sum(!is.na(meta[[x]])),
    integer(1)
  )
)
knitr::kable(
  phenotype_summary,
  caption = "Recorded phenotypes in the selected demonstration subset"
)
```


## Limitations

This analysis has several important caveats:

- **Self-reported metadata.** AGP health conditions are self-reported by
  citizen-science participants, with unknown accuracy.
- **16S resolution.** The Escherichia-Shigella inflated abundance is a known
  V4 artifact. Shotgun metagenomics would resolve this.
- **dCST-stratified demonstration subset.** Four uncommon dCSTs are included
  exhaustively and the eligible background is sampled with seed 42.
  Phenotypes do not affect membership, but differing inclusion probabilities
  mean the object must not be used for population prevalence estimates,
  effect-size estimation, or association testing.

The repository also contains the source for a separate 5,000-sample companion
analysis. It is not installed as a package vignette.

## Summary

The dCST framework provides a clustering-free approach to
community typing that is deterministic, hierarchical, and biologically
interpretable. The combination of depth-1 dCSTs (dominant species) and depth-2
refinement (co-dominance patterns) captures structure at multiple resolutions
without requiring any parameter tuning beyond the minimum support threshold
`n0`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hypercube_embedding.R
\name{linf.hypercube.embedding}
\alias{linf.hypercube.embedding}
\title{Extended homogeneous-coordinate hypercube embedding}
\usage{
linf.hypercube.embedding(
  X,
  reference,
  lambda = NULL,
  sigma.quantile = 0.95,
  sigma.target = 0.95,
  feature.ids = NULL,
  feature.labels = NULL,
  tol = 0,
  backend = c("auto", "dense", "sparse")
)
}
\arguments{
\item{X}{Nonnegative numeric matrix with samples in rows and features in
columns.}

\item{reference}{Reference component. May be a column index, feature ID, or
feature label.}

\item{lambda}{Positive numeric scalar. If \code{NULL}, choose a data-scaled value
using \code{sigma.quantile} and \code{sigma.target}.}

\item{sigma.quantile}{Quantile of positive finite-reference \eqn{\|z\|_1}
values used when \code{lambda = NULL}.}

\item{sigma.target}{Target value of \eqn{\sigma_\lambda(t)} at the selected
quantile when \code{lambda = NULL}.}

\item{feature.ids}{Optional stable feature identifiers, length \code{ncol(X)}.}

\item{feature.labels}{Optional display labels, length \code{ncol(X)}.}

\item{tol}{Nonnegative tolerance. Reference entries \verb{<= tol} are treated as
zero, and L-infinity norms \verb{<= tol} are treated as zero.}

\item{backend}{Matrix backend: \code{"auto"}, \code{"dense"}, or \code{"sparse"}. Sparse
inputs are accepted, but the returned embedding is a dense matrix because
homogeneous-coordinate embeddings are generally dense.}
}
\value{
A numeric matrix with \code{nrow(X)} rows and \code{ncol(X) - 1} columns. The
columns correspond to the non-reference components. Attributes record the
reference component, lambda choice, and finite/boundary row counts.
}
\description{
Computes the zero-aware hypercube embedding associated with one reference
component of a nonnegative compositional matrix. For rows with positive
reference component, the function forms the ordinary homogeneous ratios
against that reference and radially maps them into the unit cube. For rows
whose reference component is zero, it uses the L-infinity boundary extension
so that the embedding remains defined.
}
\details{
Let \eqn{x = (x_1,\ldots,x_p)} be a nonnegative row and let \eqn{k} be the
reference component. When \eqn{x_k > 0}, define
\eqn{z = x_{-k}/x_k}. The embedded row is
\deqn{
  \sigma_\lambda(\|z\|_1)\frac{z}{\|z\|_\infty},
  \qquad
  \sigma_\lambda(t) = 1 - \exp(-\lambda t).
}
When \eqn{x_k = 0}, the embedded row is the L-infinity-normalized boundary
vector
\deqn{
  x_{-k}/\|x_{-k}\|_\infty.
}
All-zero rows are mapped to all-zero embedded rows by convention.

If \code{lambda} is not supplied, it is chosen from the positive finite-reference
rows so that \code{sigma.target} is attained at the \code{sigma.quantile} quantile of
\eqn{\|z\|_1}. This is a numerical scaling convention for finite datasets; it
does not change the reference component or the boundary extension rule.
}
\examples{
X <- rbind(
  c(A = 2, B = 1, C = 1),
  c(A = 0, B = 2, C = 1)
)
linf.hypercube.embedding(X, reference = "A", lambda = log(2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dcst_landmark_pipeline.R
\name{linf.dcst.landmark.pipeline}
\alias{linf.dcst.landmark.pipeline}
\title{Landmark-aware dCST pipeline}
\usage{
linf.dcst.landmark.pipeline(
  X,
  feature.ids = NULL,
  feature.labels = NULL,
  n0.depth1 = 50,
  n0.depth2 = 25,
  refinement.factor = 2,
  sep = "__",
  low.freq.policy = c("pure", "absorb"),
  rare.label = "RARE_DOMINANT",
  depth1.landmark.types = c("endpoint.max", "endpoint.min", "mean.rep", "median.rep"),
  depth2.landmark.types = c("endpoint.max", "endpoint.min", "mean.rep", "median.rep"),
  landmark.view = c("absorb", "active", "rare"),
  tie.method = c("first", "random", "error"),
  verbose = FALSE,
  backend = c("auto", "dense", "sparse")
)
}
\arguments{
\item{X}{Numeric matrix (samples x features). Must be finite and nonnegative.}

\item{feature.ids}{Optional character vector of stable feature identifiers,
length \code{ncol(X)}.}

\item{feature.labels}{Optional character vector of display labels, length
\code{ncol(X)}.}

\item{n0.depth1}{Integer >= 1. Minimum support for a depth-1 dominance
sample set.}

\item{n0.depth2}{Integer >= 1. Minimum support for a depth-2 child lineage.}

\item{refinement.factor}{Numeric > 0. Auto-refinement threshold multiplier
passed to \code{\link{refine.linf.csts}}.}

\item{sep}{Character scalar used to join depth-refined dCST path tokens.}

\item{low.freq.policy}{Character. One of \code{"pure"} or \code{"absorb"}.
Controls the active dCST view while still preserving parallel rare/absorb
views inside the returned dCST objects. The legacy value \code{"rare"} is
still accepted as a deprecated alias for \code{"pure"}.}

\item{rare.label}{Character scalar for rare buckets.}

\item{depth1.landmark.types}{Character vector of landmark types for depth 1.}

\item{depth2.landmark.types}{Character vector of landmark types for depth 2.}

\item{landmark.view}{Character. One of \code{"absorb"}, \code{"active"}, or
\code{"rare"}. Defaults to \code{"absorb"} because landmark reporting is
typically requested on the absorb view.}

\item{tie.method}{Character. Tie handling passed through to
\code{\link{linf.csts}} and \code{\link{linf.landmarks}}.}

\item{verbose}{Logical. Passed to \code{\link{refine.linf.csts}}.}

\item{backend}{Character. Matrix backend to use: \code{"auto"},
\code{"dense"}, or \code{"sparse"}.}
}
\value{
A named list with components:
\itemize{
\item \code{linf.rel}: L-infinity-normalized matrix.
\item \code{dcst.depth1}: depth-1 \code{"linf.csts"} object.
\item \code{dcst.depth2}: depth-2 \code{"linf.csts"} object.
\item \code{landmarks.depth1}: \code{"linf.landmarks"} object at depth 1.
\item \code{landmarks.depth2}: \code{"linf.landmarks"} object at depth 2.
\item \code{params}: compact record of the pipeline arguments used.
}
}
\description{
Small wrapper that normalizes a nonnegative matrix with
\code{\link{normalize.linf}}, computes depth-1 dCSTs with
\code{\link{linf.csts}}, refines once to depth 2 with
\code{\link{refine.linf.csts}}, and then computes landmark points for both
depths with \code{\link{linf.landmarks}}.

The function is intentionally explicit rather than highly abstracted. It
keeps the intermediate objects visible and returns them together in one list
so downstream workflows can inspect the normalized matrix, the depth-1 dCSTs,
the depth-2 dCSTs, and the landmark tables without re-running the pipeline.
}
\examples{
X <- rbind(
  s1 = c(10, 8, 1),
  s2 = c(9, 7, 2),
  s3 = c(8, 2, 7),
  s4 = c(7, 1, 8),
  s5 = c(1, 10, 2),
  s6 = c(2, 9, 1)
)
ids <- c("asv_1", "asv_2", "asv_3")
labels <- c("L. iners 1", "Gard. vaginalis 2", "BVAB1 3")

out <- linf.dcst.landmark.pipeline(
  X,
  feature.ids = ids,
  feature.labels = labels,
  n0.depth1 = 2,
  n0.depth2 = 2,
  refinement.factor = 2,
  low.freq.policy = "pure",
  landmark.view = "absorb",
  verbose = FALSE
)

names(out)
out$dcst.depth2$cst.depth
out$landmarks.depth2$view
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{valencia2k}
\alias{valencia2k}
\title{Valencia 2k vaginal microbiome dataset}
\format{
A list with five components:
\describe{
\item{rel}{Numeric matrix (2000 x 178). Compositional relative abundances;
each row sums to 1. Rows are samples, columns are taxonomic features.}
\item{cst}{Data frame (2000 x 3) with columns:
\code{sample_id} (character),
\code{Val_CST} (Valencia CST assignment: I, II, III, IV-A, IV-B, IV-C, V),
\code{Val_subCST} (Valencia sub-CST assignment: I-A, I-B, II, III-A, III-B,
IV-A, IV-B, IV-C0, IV-C1, IV-C2, IV-C3, IV-C4, V).}
\item{reads}{Integer vector of length 2000. Per-sample read counts after
taxonomic filtering. Use \code{sweep(valencia2k$rel, 1, valencia2k$reads, "*")}
to reconstruct a count-like matrix.}
\item{taxa}{Character vector of 178 taxon names (column names of \code{rel}).}
\item{source}{Character string documenting provenance.}
}
}
\source{
Subsampled from the VALENCIA training data at
\url{https://github.com/ravel-lab/VALENCIA}. See
\code{data-raw/build_valencia2k.R} and the installed
\code{DATA_PROVENANCE.md} file for construction and licensing details.
}
\usage{
valencia2k
}
\description{
A stratified subsample of 2,000 vaginal samples from the Valencia 13k
CST-classifier training set (France et al. 2020), bundled as an example
dataset for demonstrating dCST construction after L-infinity normalization.
}
\details{
The subsample preserves proportional representation of all 13 Valencia
sub-CSTs and was drawn with \code{set.seed(42)}.

The Valencia CST classifier (France et al. 2020) assigns vaginal microbiome
samples to community state types (CSTs) based on nearest-centroid
classification in relative-abundance space. The original training set
contains 12,881 samples and 178 taxonomic features after filtering.

This 2,000-sample subsample is intended for vignette demonstrations. The
compositional matrix can be used directly with \code{\link{normalize.linf}}
and downstream dCST functions. For workflows that require count-like input,
reconstruct approximate counts using the \code{reads} vector.
}
\examples{
data(valencia2k)
dim(valencia2k$rel)           # 2000 x 178
table(valencia2k$cst$Val_CST) # CST distribution
head(valencia2k$taxa, 10)     # first 10 taxon names

}
\references{
France, M. T., Ma, B., Gajer, P., Brown, S., Humphrys, M. S., Holm, J. B.,
Waetjen, L. E., Brotman, R. M., & Ravel, J. (2020). VALENCIA: a nearest
centroid classification method for vaginal microbial communities based on
composition. \emph{Microbiome}, 8(1), 166.
\doi{10.1186/s40168-020-00934-6}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linf.R
\name{dcst.view}
\alias{dcst.view}
\title{Select a stored dCST policy view}
\usage{
dcst.view(csts, view = c("absorb", "pure"))
}
\arguments{
\item{csts}{A \code{"linf.csts"} object produced by \code{\link{linf.csts}}
and optionally refined by \code{\link{refine.linf.csts}} /
\code{\link{refine.linf.csts.iter}}.}

\item{view}{Character. The stored policy view to activate: \code{"pure"} or
\code{"absorb"}.}
}
\value{
A \code{"linf.csts"} object using the requested view.
}
\description{
Returns a copy of a \code{"linf.csts"} object using either the stored
\code{"pure"} or \code{"absorb"} hierarchy. This does not recompute dCSTs.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/landmarks.R
\name{linf.landmarks}
\alias{linf.landmarks}
\title{Landmark points for dCST dominance-lineages}
\usage{
linf.landmarks(
  M,
  csts,
  depth = NULL,
  view = c("active", "rare", "absorb"),
  landmark.types = c("endpoint.max", "endpoint.min", "mean.rep", "median.rep"),
  tie.method = c("first", "random", "error"),
  backend = c("auto", "dense", "sparse")
)
}
\arguments{
\item{M}{Numeric matrix (samples x features) used to build or refine the dCSTs.}

\item{csts}{A \code{"linf.csts"} object.}

\item{depth}{Integer. dCST depth to inspect. Defaults to the leaf depth
\code{csts$cst.depth}.}

\item{view}{Character. One of \code{"active"}, \code{"rare"}, or
\code{"absorb"}.}

\item{landmark.types}{Character vector containing any of
\code{"endpoint.max"}, \code{"endpoint.min"}, \code{"mean.rep"}, or
\code{"median.rep"}.}

\item{tie.method}{Character. Tie handling for landmark selection:
\code{"first"}, \code{"random"}, or \code{"error"}.}

\item{backend}{Character. Matrix backend to use: \code{"auto"},
\code{"dense"}, or \code{"sparse"}. The default \code{"auto"} preserves
sparse input and otherwise uses the dense path.}
}
\value{
A list of class \code{"linf.landmarks"} with components:
\itemize{
\item \code{depth}, \code{view}, \code{sep}, \code{rare.label}
\item \code{feature.ids}, \code{feature.labels}
\item \code{cells}: one row per dominance-lineage with computability
metadata; the component name is retained for backward compatibility
\item \code{landmarks}: one row per computed landmark point
}
}
\description{
Computes representative landmark points for the dominance-lineages of a
\code{"linf.csts"} object at a chosen depth and view.

Landmark types are defined with respect to the leaf feature of the dCST path:
the last feature ID in the lineage ID path.
Lineages whose leaf token is \code{rare.label} are reported but skipped for
landmark computation because they do not correspond to a unique target
feature.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transfer_dcsts.R
\name{transfer.dcsts}
\alias{transfer.dcsts}
\title{Transfer samples into a frozen dCST hierarchy}
\usage{
transfer.dcsts(
  X,
  csts,
  depth = NULL,
  view = c("absorb", "active", "rare"),
  match.by = c("feature.ids", "feature.labels"),
  feature.ids = NULL,
  feature.labels = NULL,
  tie.method = c("support", "first", "random", "error"),
  carry.forward.terminal.depths = TRUE,
  sep = NULL,
  backend = c("auto", "dense", "sparse")
)
}
\arguments{
\item{X}{Nonnegative sample-by-feature matrix to transfer. Columns may be in
any order and may include features not present in \code{csts}; missing
retained features are treated as zero.}

\item{csts}{A fitted \code{"linf.csts"} object produced by
\code{\link{linf.csts}} and optionally refined by
\code{\link{refine.linf.csts}} or \code{\link{refine.linf.csts.iter}}.}

\item{depth}{Integer vector of requested depths. Defaults to all fitted
depths in \code{csts}.}

\item{view}{Which fitted hierarchy view to transfer into. \code{"absorb"}
uses \code{cst.levels.absorb}; \code{"rare"} uses
\code{cst.levels.rare}; \code{"active"} uses \code{cst.levels}.}

\item{match.by}{Whether columns in \code{X} are aligned to
\code{csts$feature.ids} or \code{csts$feature.labels}.}

\item{feature.ids}{Optional feature identifiers for columns of \code{X}.
Defaults to \code{colnames(X)}.}

\item{feature.labels}{Optional feature labels for columns of \code{X}.
Defaults to \code{feature.ids}.}

\item{tie.method}{How to resolve ties among frozen child lineages with equal
sample abundance. \code{"support"} chooses the tied lineage with largest
reference support and then lexical order; \code{"first"} uses frozen child
order; \code{"random"} samples one tied lineage; \code{"error"} stops.}

\item{carry.forward.terminal.depths}{Logical. If \code{TRUE}, an unmatched
depth-1 sample may fall back to its dominant retained feature. Terminal
lineages at later depths are assigned only when they appear as realized
stable lineage sets in the fitted hierarchy.}

\item{sep}{Separator used in lineage labels. Defaults to \code{csts$sep} or
\code{"__"}.}

\item{backend}{Matrix backend; passed to the internal matrix preparer.}
}
\value{
A list of class \code{"linf.dcst.transfer"} with components:
\itemize{
\item \code{assignment}: character matrix of transferred labels for the
requested depths.
\item \code{all.depths}: character matrix for all fitted depths.
\item \code{depth}: requested depth vector.
\item \code{view}, \code{match.by}, \code{tie.method}: settings used.
}
}
\description{
Assigns rows of a new sample-by-feature matrix into the realized lineage sets
of a fitted \code{"linf.csts"} hierarchy. The hierarchy is not refit: each
sample is walked through the frozen tree by choosing, at each depth, the
realized child lineage whose newly added feature has the largest abundance in
that sample.
}
\examples{
X <- rbind(
  s1 = c(A = 10, B = 2, C = 1),
  s2 = c(A = 9, B = 3, C = 1),
  s3 = c(A = 1, B = 10, C = 2),
  s4 = c(A = 1, B = 9, C = 3)
)
M <- normalize.linf(X)
fit <- linf.csts(M, n0 = 2, low.freq.policy = "absorb")
transfer.dcsts(X, fit)$assignment

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linf.R
\name{asv.to.linf.csts}
\alias{asv.to.linf.csts}
\title{From counts to filtered, normalized, truncated dCSTs}
\usage{
asv.to.linf.csts(S.counts, ..., backend = c("auto", "dense", "sparse"))
}
\arguments{
\item{S.counts}{Numeric matrix of counts (samples x features).}

\item{...}{Arguments passed to \code{filter.asv()} (e.g., \code{min.lib},
\code{prev.prop}, \code{min.count}, \code{min.rel}).}

\item{backend}{Character. Matrix backend to use: \code{"auto"},
\code{"dense"}, or \code{"sparse"}.}
}
\value{
A list with:
\itemize{
\item \code{filter}: the full result from \code{filter.asv()}.
\item \code{linf.rel}: L-infinity-normalized matrix on filtered counts.
\item \code{csts}: result from \code{linf.csts()} on \code{linf.rel}.
}
}
\description{
Convenience pipeline: runs \code{filter.asv()} on counts, computes
L-infinity-normalized relatives on the filtered counts, and assigns
truncated dCSTs.
}
\examples{
set.seed(1)
S <- matrix(rpois(100, lambda = 5), nrow = 20, ncol = 5)
res <- asv.to.linf.csts(S, min.lib = 10, prev.prop = 0.1, min.count = 1)
names(res)
apply(res$linf.rel, 1, max)  # should be 1 (or 0 for all-zero rows)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{agp_gut}
\alias{agp_gut}
\title{American Gut Project gut microbiome dataset}
\format{
A list with four components:
\describe{
\item{counts}{Integer matrix (766 x 314). Raw 16S V4 read counts.
Rows are samples, columns are SILVA species-level taxa.}
\item{meta}{Data frame (766 rows) with columns:
\code{Run} (SRA run accession),
\code{dcst_depth1}, \code{dcst_depth2} (pre-computed dCST labels),
\code{IBS}, \code{IBD}, \code{Diabetes}, \code{Autoimmune},
\code{Seasonal_allergies}, \code{Migraine}, \code{Acid_reflux},
\code{Lung_disease}, \code{Cardiovascular_disease}, \code{Skin_condition},
\code{Obesity} (binary disease indicators from self-reported AGP metadata),
\code{BMI} (numeric, self-reported), and
\code{selection_reason} (target-dCST inclusion or seeded background
sampling).}
\item{taxa}{Character vector of 314 SILVA taxonomy strings.}
\item{source}{Character string documenting provenance.}
}
}
\source{
Derived from the public American Gut Project records under ENA accession
\href{https://www.ebi.ac.uk/ena/browser/view/PRJEB11419}{PRJEB11419} via the
PRIME pipeline. See \code{data-raw/create_agp_gut_subset.py},
\code{data-raw/build_agp_gut.R}, and the installed
\code{DATA_PROVENANCE.md} file.
}
\usage{
agp_gut
}
\description{
A stratified subsample of 766 gut microbiome samples from the American Gut
Project (PRJEB11419, AGP-US-2015), bundled for demonstrating L-infinity
dCST construction in a gut ecosystem.
}
\details{
The subsample includes all samples assigned to four uncommon demonstration
dCSTs (Prevotella_7, Pasteurellaceae, Akkermansia, and Staphylococcus).
The remaining slots are a simple random sample, drawn with seed 42, from
the eligible background after excluding Eukaryota and Unassigned labels.
Phenotype fields are joined only after membership is fixed and do not
influence selection. Because inclusion probabilities differ by dCST, this
object is a computational demonstration dataset rather than a probability
sample of the underlying cohort.

The American Gut Project is a large citizen-science 16S rRNA survey of the
human microbiome. Health conditions are self-reported via questionnaire and
should be interpreted with appropriate caution.

The phenotype fields must not be used with this dCST-stratified subset for
population prevalence estimates, effect-size estimation, or association
testing. The exact selection is generated by
\code{data-raw/create_agp_gut_subset.py}; run-ID membership, selection
reasons, and derived annotations are retained in
\code{inst/extdata/agp_gut_meta.csv}.

The count matrix can be used directly with \code{\link{filter.asv}},
\code{\link{normalize.linf}}, and downstream dCST functions.

Note on Escherichia-Shigella: this genus is inflated in 16S V4 data due to
primer cross-reactivity and should be interpreted with caution.
}
\examples{
data(agp_gut)
dim(agp_gut$counts)                     # 766 x 314
table(agp_gut$meta$dcst_depth1)         # dCST distribution
sum(agp_gut$meta$IBS)                   # IBS cases

}
\references{
McDonald, D., Hyde, E., Debelius, J. W., et al. (2018). American Gut: an
Open Platform for Citizen Science Microbiome Research. \emph{mSystems},
3(3), e00031-18. \doi{10.1128/mSystems.00031-18}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linf.R
\name{summary.linf.csts}
\alias{summary.linf.csts}
\title{Summarize dCST hierarchy statistics}
\usage{
\method{summary}{linf.csts}(object, ...)
}
\arguments{
\item{object}{A \code{"linf.csts"} object.}

\item{...}{Unused.}
}
\value{
Data frame with depth-wise statistics
}
\description{
Summarize dCST hierarchy statistics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linf.R
\name{refine.linf.csts}
\alias{refine.linf.csts}
\title{Refine a dCST hierarchy by one level}
\usage{
refine.linf.csts(
  M,
  csts,
  n0 = 50,
  refinement.factor = 2,
  sep = "__",
  low.freq.policy = c("pure", "absorb"),
  rare.label = "RARE_DOMINANT",
  verbose = TRUE,
  backend = c("auto", "dense", "sparse")
)
}
\arguments{
\item{M}{Numeric matrix (samples x features) used for refinement. Column names
should match feature labels used in dCST names.}

\item{csts}{A \code{"linf.csts"} object.}

\item{n0}{Integer >= 1. Minimum support required to retain a child lineage
(passed to \code{linf.csts}).}

\item{refinement.factor}{Numeric > 0. Auto-refine parent lineages with
support >= \code{refinement.factor * n0}.}

\item{sep}{Character scalar used to concatenate hierarchical labels.}

\item{low.freq.policy}{Character. One of \code{"pure"} or \code{"absorb"}.
Default: \code{"pure"}. The legacy value \code{"rare"} is still accepted as
a deprecated alias for \code{"pure"}.}

\item{rare.label}{Character scalar for rare buckets when \code{low.freq.policy = "pure"}.}

\item{verbose}{Logical. If TRUE, print progress information.}

\item{backend}{Character. Matrix backend to use: \code{"auto"},
\code{"dense"}, or \code{"sparse"}. The default \code{"auto"} inherits the
backend from \code{M} or from \code{csts} when available.}
}
\value{
Updated \code{"linf.csts"} object with \code{cst.depth} increased by one and
updated \code{cell.label}. Policy-specific views are stored in
\code{cell.label.rare} and \code{cell.label.absorb}.
}
\description{
Selects well-supported leaf dominance-lineages and refines them by dropping
the dominant feature(s) encoded in the parent label path and re-applying
\code{\link{linf.csts}} to the remaining features. The resulting child
labels are appended to the parent label using \code{sep}.

Low-support child lineages are handled by \code{low.freq.policy}. When
\code{low.freq.policy = "pure"}, rare buckets at depth >= 2 become
parent-prefixed automatically via the hierarchical \code{paste(parent, child, sep = sep)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linf.R
\name{linf.cells}
\alias{linf.cells}
\title{Dominant-feature assignment}
\usage{
linf.cells(
  S,
  feature.ids = NULL,
  feature.labels = NULL,
  tie.method = c("first", "random", "error"),
  return.value = FALSE,
  backend = c("auto", "dense", "sparse")
)
}
\arguments{
\item{S}{Numeric matrix (samples x features), typically L-infinity-normalized.}

\item{feature.ids}{Optional character vector of stable feature identifiers,
length \code{ncol(S)}.}

\item{feature.labels}{Optional character vector of display labels, length
\code{ncol(S)}.}

\item{tie.method}{Character. How to resolve ties during dominant-feature
assignment.}

\item{return.value}{Logical. If \code{TRUE}, include a \code{value} vector with row maxima.}

\item{backend}{Character. Matrix backend to use: \code{"auto"},
\code{"dense"}, or \code{"sparse"}. The default \code{"auto"} preserves
sparse input and otherwise uses the dense path.}
}
\value{
A list with components:
\itemize{
\item \code{index}: integer index of the dominant column per sample (\code{NA} for all-zero rows)
\item \code{id}: dominant feature ID per sample (\code{NA} for all-zero rows)
\item \code{label}: dominant column label per sample (\code{NA} for all-zero rows)
\item \code{id.levels}: full feature ID set after \code{make.unique(..., sep = "_")}
\item \code{levels}: full column label set after \code{make.unique(..., sep = "_")}
\item \code{observed.id.levels}: subset of \code{id.levels} that appear in \code{id}
\item \code{observed.levels}: subset of \code{levels} that appear in \code{label}
\item \code{value}: row maxima (only when \code{return.value = TRUE})
}
}
\description{
Assigns each row to the column achieving its maximum.

For each sample (row) of a nonnegative matrix, identifies the dominant
feature as the column with the maximum value. Samples with the same dominant
feature form a depth-1 dominance sample set. Ties are broken by the first
maximum (as in \code{max.col(..., ties.method = "first")}). Rows that are
all zero are assigned \code{NA}.

Feature IDs default to \code{colnames(S)}; if absent, synthetic IDs
\code{"V1", "V2", ..., "Vp"} are generated. Display labels default to the
feature IDs unless \code{feature.labels} is supplied. To guarantee a 1-1
mapping between columns and both IDs and labels, duplicates are
disambiguated via \code{make.unique()}.
}
\examples{
# Basic example with named columns
S <- rbind(
  a = c(A = 10, B = 5,  C = 0),   # -> A
  b = c(A = 0,  B = 0,  C = 0),   # -> NA
  c = c(A = 1,  B = 4,  C = 4)    # tie -> first max: B
)
out <- linf.cells(S)
out$index
out$label
out$levels
out$observed.levels

# Unnamed columns (synthetic labels V1..Vp), duplicate names disambiguated
T <- matrix(c(0,2,  3,1,  0,0), nrow = 3, byrow = TRUE)
colnames(T) <- c("X", "X")  # duplicates -> X, X_1
linf.cells(T)$levels

# With L-infinity normalization in a pipeline
M <- normalize.linf(S)
linf.cells(M)$label

}
\seealso{
\code{\link{normalize.linf}}, \code{\link{linf.csts}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{valencia13k_dcst_depth2_merged}
\alias{valencia13k_dcst_depth2_merged}
\title{Valencia 13k merged depth-2 dCST assignments}
\format{
A list with five components:
\describe{
\item{assignments}{Data frame with one row per source sample and columns
\code{sample_id} (anonymized bundled-data identifier), \code{source_row}
(row number in the filtered Valencia 13k source), \code{Val_CST},
\code{Val_subCST}, \code{dcst_depth1}, and \code{dcst_depth2}.}
\item{summaries}{Named list of per-depth summary tables. Each table
contains \code{depth}, \code{dcst_label}, \code{n}, \code{prop}, and
\code{path_length}.}
\item{feature_labels}{Character vector of source taxon labels.}
\item{params}{List recording the construction parameters.}
\item{source}{Character string documenting provenance.}
}
}
\source{
Generated from the VALENCIA training data at
\url{https://github.com/ravel-lab/VALENCIA}. See
\code{data-raw/build_valencia13k_merged_dcst_depths.R} and the installed
\code{DATA_PROVENANCE.md} file.
}
\usage{
valencia13k_dcst_depth2_merged
}
\description{
A lightweight bundled assignment asset containing merged depth-2
Dominant community state type (dCST) labels for the full Valencia
13k vaginal microbiome training set.
}
\details{
The asset is computed from the Valencia 13k compositional matrix after
L-infinity normalization. dCSTs use \code{n0 = 50} and the merged
\code{low.freq.policy = "absorb"} view, so samples from low-support
provisional states are reassigned to retained states rather than stored as
explicit rare buckets.
}
\examples{
data(valencia13k_dcst_depth2_merged)
nrow(valencia13k_dcst_depth2_merged$assignments)
head(valencia13k_dcst_depth2_merged$summaries$depth2)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linf.R
\name{print.linf.csts}
\alias{print.linf.csts}
\title{Print method for \code{"linf.csts"}}
\usage{
\method{print}{linf.csts}(x, ...)
}
\arguments{
\item{x}{A \code{"linf.csts"} object.}

\item{...}{Unused.}
}
\value{
The input object, invisibly.
}
\description{
Print method for \code{"linf.csts"}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linf.R
\name{refine.linf.csts.iter}
\alias{refine.linf.csts.iter}
\title{Iteratively refine dCSTs by one additional level}
\usage{
refine.linf.csts.iter(
  M,
  refined,
  cells.to.refine = NULL,
  n0 = 50,
  refinement.factor = 5,
  sep = "__",
  low.freq.policy = c("pure", "absorb"),
  rare.label = "RARE_DOMINANT",
  verbose = TRUE,
  backend = c("auto", "dense", "sparse")
)
}
\arguments{
\item{M}{Numeric matrix used for refinement.}

\item{refined}{A \code{"linf.csts"} object with an existing hierarchy.}

\item{cells.to.refine}{Character vector of leaf dominance-lineage IDs to
refine, or NULL for auto-selection. The argument name is retained for
backward compatibility.}

\item{n0}{Integer >= 1. Minimum support required to retain a child lineage.}

\item{refinement.factor}{Numeric > 0. Auto-selection threshold multiplier.}

\item{sep}{Character scalar used to concatenate hierarchical labels.}

\item{low.freq.policy}{Character. One of \code{"pure"} or \code{"absorb"}.
Default: \code{"pure"}. The legacy value \code{"rare"} is still accepted as
a deprecated alias for \code{"pure"}.}

\item{rare.label}{Character scalar for rare buckets when \code{low.freq.policy = "pure"}.}

\item{verbose}{Logical. If TRUE, print progress information.}

\item{backend}{Character. Matrix backend to use: \code{"auto"},
\code{"dense"}, or \code{"sparse"}. The default \code{"auto"} inherits the
backend from \code{M} or from \code{refined} when available.}
}
\value{
Updated \code{"linf.csts"} object with \code{cst.depth} increased by one.
}
\description{
Appends one refinement level to an existing dCST hierarchy produced by
\code{\link{refine.linf.csts}} or \code{\link{refine.linf.csts.iter}}.
Dominance-lineages to refine can be provided explicitly or selected
automatically based on \code{refinement.factor * n0}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_asv.R
\name{filter.asv}
\alias{filter.asv}
\title{Filter ASV count matrix by library size (samples) and prevalence (features)}
\usage{
filter.asv(
  S.counts,
  min.lib = 1000,
  prev.prop = 0.05,
  min.count = 2,
  min.rel = NULL,
  min.feat.total = NULL,
  verbose = TRUE
)
}
\arguments{
\item{S.counts}{Numeric matrix (samples x features) of nonnegative counts.}

\item{min.lib}{Integer. Minimum library size (row sum of counts) to keep a sample. Default: 1000.}

\item{prev.prop}{Numeric in (0,1]. Minimum fraction of samples where a feature must be present.
Default: 0.05.}

\item{min.count}{Integer (>=1) or NULL. Reads to call "present" (ignored if \code{min.rel} set). Default: 2.}

\item{min.rel}{Numeric in (0,1) or NULL. Relative abundance to call "present" (overrides \code{min.count}). Default: NULL.}

\item{min.feat.total}{Integer (>=0) or NULL. Optional minimum total reads across all samples per feature. Default: NULL.}

\item{verbose}{Logical. Print keep/drop summaries. Default: TRUE.}
}
\value{
A list with elements:
\describe{
\item{counts}{Filtered count matrix (samples x features).}
\item{rel}{Row-normalized relative-abundance matrix.}
\item{kept.sample.idx}{Kept sample indices (original order).}
\item{kept.feature.idx}{Kept feature indices (original order).}
\item{prevalence}{Per-feature prevalence counts for kept features.}
\item{thresholds}{List of thresholds actually used.}
}
}
\description{
Filters an ASV \strong{count} matrix by:
(1) removing samples with library size below \code{min.lib};
(2) keeping features (taxa) that are "present" in at least
\code{ceiling(prev.prop * nrow(S.counts))} samples, where presence is
defined by either \code{min.count} (counts) or \code{min.rel} (relative).
After feature filtering, zero-total samples are dropped and a row-normalized
relative-abundance matrix is returned alongside the filtered counts.
}
\details{
Sample filtering is applied on raw counts first. Prevalence is computed on the
post-sample-filter matrix using either a count or relative rule. A feature is
retained if prevalence \eqn{\ge \lceil \text{prev.prop} \times n_{\text{samples}} \rceil}.
After feature filtering, samples with zero remaining counts are dropped and
the relative matrix \code{rel} is row-normalized.
}
\section{Conventions}{

Dot-delimited names; rows are samples, columns are features. Supply raw counts.
}

\examples{
set.seed(1)
S <- matrix(rpois(100 * 20, lambda = 5), nrow = 100, ncol = 20)
res <- filter.asv(S, min.lib = 50, prev.prop = 0.1, min.count = 2)
dim(res$counts); dim(res$rel)
range(rowSums(res$rel))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linf.R
\name{expand.rare}
\alias{expand.rare}
\title{Switch a dCST hierarchy to the "rare" view (explicit rare buckets)}
\usage{
expand.rare(csts)
}
\arguments{
\item{csts}{A \code{"linf.csts"} object produced by \code{\link{linf.csts}}
and optionally refined by \code{\link{refine.linf.csts}} /
\code{\link{refine.linf.csts.iter}}.}
}
\value{
A \code{"linf.csts"} object using the "rare" view.
}
\description{
Returns a copy of a \code{"linf.csts"} object with \code{cell.label} (and, if
present, \code{cst.levels}) replaced by the precomputed "rare" labeling.
This does not recompute dCSTs; it only switches between labelings already
stored in the object. This explicit rare-bucket view is the active labeling
used by \code{low.freq.policy = "pure"}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/label_format.R
\name{linf.feature.labels}
\alias{linf.feature.labels}
\title{Format feature labels for dominant-feature assignments and dCSTs}
\usage{
linf.feature.labels(
  feature.ids,
  taxonomy,
  abbreviations = NULL,
  aliases = NULL,
  duplicate.index = c("global", "within_taxon", "none"),
  fallback.to.id = TRUE
)
}
\arguments{
\item{feature.ids}{Character vector of stable feature identifiers.}

\item{taxonomy}{Character vector of taxonomy strings, same length as
\code{feature.ids}.}

\item{abbreviations}{Optional named character vector mapping genus tokens to
display abbreviations, for example \code{c(Lactobacillus = "L.")}.}

\item{aliases}{Optional named character vector mapping full taxonomy strings
to alternate labels, for example \code{c("Ca. Lachnocurva vaginae" = "BVAB1")}.}

\item{duplicate.index}{One of \code{"global"}, \code{"within_taxon"}, or
\code{"none"}. Controls how duplicate display taxa are disambiguated.}

\item{fallback.to.id}{Logical. If \code{TRUE}, missing taxonomy values fall
back to the feature ID.}
}
\value{
Character vector of unique display labels.
}
\description{
Builds unique display labels from stable feature IDs and taxonomy strings.
This is useful when dCST computation should operate on stable internal feature
identifiers (for example \code{asv_4}) while reports and figures should use
human-readable labels such as \verb{L. iners 4}.

Underscores in taxonomy strings are converted to spaces before abbreviation
and aliasing. If multiple features share the same display taxon, an index can
be appended either from the global feature ID (for example \code{asv_4 -> 4}) or
by within-taxon order.
}
\examples{
ids <- c("asv_1", "asv_4", "asv_5", "asv_6")
tax <- c(
  "Lactobacillus iners",
  "Lactobacillus iners",
  "Megasphaera lornae",
  "Ca_Lachnocurva_vaginae"
)
abbr <- c(
  Lactobacillus = "L.",
  Gardnerella = "Gard.",
  Megasphaera = "Mega."
)
aliases <- c(
  "Ca. Lachnocurva vaginae" = "BVAB1",
  "Ca_Lachnocurva_vaginae" = "BVAB1"
)
linf.feature.labels(ids, tax, abbreviations = abbr, aliases = aliases)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linf.R
\name{linf.csts}
\alias{linf.csts}
\title{Truncated dominant community state types with configurable low-support handling}
\usage{
linf.csts(
  S,
  feature.ids = NULL,
  feature.labels = NULL,
  n0 = 50,
  low.freq.policy = c("pure", "absorb"),
  rare.label = "RARE_DOMINANT",
  tie.method = c("first", "random", "error"),
  return.diagnostics = FALSE,
  return.landmarks = FALSE,
  landmark.types = c("endpoint.max", "endpoint.min"),
  landmark.view = c("active", "rare", "absorb"),
  backend = c("auto", "dense", "sparse")
)
}
\arguments{
\item{S}{Numeric matrix (samples x features), typically L-infinity relatives.}

\item{feature.ids}{Optional character vector of stable feature identifiers,
length \code{ncol(S)}.}

\item{feature.labels}{Optional character vector of display labels, length
\code{ncol(S)}.}

\item{n0}{Integer >= 1. Minimum support required to retain a dominance
sample set.}

\item{low.freq.policy}{Character. One of \code{"pure"} or \code{"absorb"}.
Default: \code{"pure"}. The legacy value \code{"rare"} is still accepted as
a deprecated alias for \code{"pure"}.}

\item{rare.label}{Character scalar used when \code{low.freq.policy = "pure"}.
Default: \code{"RARE_DOMINANT"}.}

\item{tie.method}{Character. Tie handling passed to \code{linf.cells()} and used
during absorb reassignment ("first", "random", "error").}

\item{return.diagnostics}{Logical. If TRUE, return reassignment diagnostics.}

\item{return.landmarks}{Logical. If TRUE, attach a depth-1 landmark summary
computed by \code{\link{linf.landmarks}}.}

\item{landmark.types}{Character vector of landmark types passed to
\code{\link{linf.landmarks}} when \code{return.landmarks = TRUE}.}

\item{landmark.view}{Character. Landmark view passed to
\code{\link{linf.landmarks}} when \code{return.landmarks = TRUE}.}

\item{backend}{Character. Matrix backend to use: \code{"auto"},
\code{"dense"}, or \code{"sparse"}. The default \code{"auto"} preserves
sparse input and otherwise uses the dense path.}
}
\value{
List with:
\itemize{
\item \code{cell.index}, \code{cell.id}, \code{cell.label}: active labeling per \code{low.freq.policy}
\item \code{cell.index.rare}, \code{cell.id.rare}, \code{cell.label.rare}
\item \code{cell.index.absorb}, \code{cell.id.absorb}, \code{cell.label.absorb}
\item \code{kept.cells.idx}, \code{kept.cells.id}, \code{kept.cells.lbl}
\item \code{raw.index}, \code{raw.id}, \code{raw.label}
\item \code{feature.ids}, \code{feature.labels}
\item \code{size.table}, \code{size.table.id}
\item \code{n0}, \code{low.freq.policy}, \code{rare.label}
\item \code{diagnostics} (if \code{return.diagnostics = TRUE})
\item \code{landmarks} (if \code{return.landmarks = TRUE})
}
}
\description{
Forms provisional depth-1 dominance sample sets from the dominant feature
of each sample and then applies the minimum support threshold \code{n0}.
Sets with fewer than \code{n0} samples are handled according to
\code{low.freq.policy}:
\itemize{
\item \code{"pure"}: retain only sets with support >= \code{n0} as named
dCSTs and collapse all low-support sets into \code{rare.label}.
\item \code{"absorb"}: reassign each low-support sample to the retained
state with the largest value among the retained features (ties handled
by \code{tie.method}).
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linf.R
\name{latex.linf.csts}
\alias{latex.linf.csts}
\title{Render dCSTs as a LaTeX table}
\usage{
latex.linf.csts(
  csts,
  depth = NULL,
  caption = NULL,
  label = NULL,
  digits = 1,
  include.percent = TRUE
)
}
\arguments{
\item{csts}{A dCST object produced by \code{linf.csts()} and optionally refined
by \code{refine.linf.csts()} or \code{refine.linf.csts.iter()}.}

\item{depth}{Integer. dCST depth to render. Default is the leaf level
(\code{csts$cst.depth}). If \code{cst.levels} is missing, depth is ignored.}

\item{caption}{Character or NULL. LaTeX table caption.}

\item{label}{Character or NULL. LaTeX label for referencing the table.}

\item{digits}{Integer. Digits to display for percentages. Default: 1.}

\item{include.percent}{Logical. If TRUE, include percent of total samples.}
}
\value{
Character vector containing LaTeX table code.
}
\description{
Generates LaTeX code summarizing dCSTs at a specified hierarchy depth.
Uses the explicit dCST hierarchy stored in \code{cst.levels} when available.
}
\details{
This function is hierarchy-aware. It does not infer dCST depth from label
strings. If the dCST object does not contain \code{cst.levels}, the function
falls back to using \code{cell.label} as a single-level dCST.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{valencia_linf_hypercube_1k}
\alias{valencia_linf_hypercube_1k}
\title{Valencia 1k four-component hypercube embedding example}
\format{
A list with four components:
\describe{
\item{rel4}{Numeric matrix (1000 x 4). Rows are samples and columns are
\code{Li}, \code{Lc}, \code{Gv}, and \code{Bv}. Each row sums to 1 after
restricting the original Valencia profile to the four mapped taxa.}
\item{meta}{Data frame (1000 rows) with columns:
\code{sample_id} (anonymized bundled-data identifier),
\code{source_row} (row number in the filtered Valencia 13k source),
\code{Val_CST}, \code{Val_subCST}, \code{selected_mass} (original
relative-abundance mass carried by the four selected taxa), and
\code{dominant_component} (largest of \code{Li}, \code{Lc},
\code{Gv}, \code{Bv} after renormalization).}
\item{component_map}{Named character vector mapping \code{Li},
\code{Lc}, \code{Gv}, and \code{Bv} to the original Valencia taxon names:
\code{Lactobacillus_iners}, \code{Lactobacillus_crispatus},
\code{Gardnerella_vaginalis}, and \code{BVAB1}.}
\item{source}{Character string documenting provenance.}
}
}
\source{
Generated from the VALENCIA training data at
\url{https://github.com/ravel-lab/VALENCIA}. See
\code{data-raw/build_valencia_linf_hypercube_1k.R} and the installed
\code{DATA_PROVENANCE.md} file.
}
\usage{
valencia_linf_hypercube_1k
}
\description{
A stratified 1,000-sample subset of the Valencia 13k vaginal microbiome
training set, reduced to four selected phylotype coordinates and
L1-normalized over those coordinates. The object is bundled as a lightweight
example for the zero-aware homogeneous-coordinate hypercube embedding in
\code{\link{linf.hypercube.embedding}}.
}
\details{
Rows with zero total mass in the four selected taxa are removed before
renormalization. Sampling is stratified by the dominant selected component,
using \code{set.seed(20261604)}. The object is not intended to replace the
full Valencia matrix; it is a compact reproducible example for visualizing
compositional projective-space coordinate charts.
}
\examples{
data(valencia_linf_hypercube_1k)
dim(valencia_linf_hypercube_1k$rel4)
table(valencia_linf_hypercube_1k$meta$dominant_component)
emb <- linf.hypercube.embedding(
  valencia_linf_hypercube_1k$rel4,
  reference = "Li"
)
dim(emb)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/linf.R
\name{normalize.linf}
\alias{normalize.linf}
\title{L-infinity normalization (row-wise)}
\usage{
normalize.linf(X, tol = 0, backend = c("auto", "dense", "sparse"))
}
\arguments{
\item{X}{Numeric matrix (samples x features).}

\item{tol}{Numeric >= 0. Values with row max <= tol are treated as zero rows.
Default: 0 (exact zero only).}

\item{backend}{Character. Matrix backend to use: \code{"auto"},
\code{"dense"}, or \code{"sparse"}. The default \code{"auto"} preserves
sparse input and otherwise uses the dense path.}
}
\value{
Numeric matrix of same dimensions as X, L-infinity normalized.
}
\description{
Scales each row of a numeric matrix by its L-infinity norm (row maximum).
Rows whose maximum is zero (or below tolerance) are left unchanged
and remain all-zero.
}
\details{
Zero rows have undefined L-infinity direction. By convention, they are
preserved as all-zero rows and will yield NA labels in downstream
dominant-feature or dCST assignment.
}
